(function() {
    var kTrackingTestRegExp = /\?jiveRnd=.*?jive/;

    function generateUuid() {
        var seed = new Date().getTime();

        return "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace( /[xy]/g, function( c ) {
            var r = ( seed + Math.random() * 16 ) % 16 | 0;

            seed  = Math.floor( seed / 16 );

            return ( c === "x" ? r : ( r & 0x7 | 0x8 ) ).toString( 16 );
        });
    }

    function convertQueryToJson( query ) {
        var rows = query.split( "&" ),
            i, len, row, keyValuePair,
            result = {};

        for ( i = 0, len = rows.length; i < len; i++ ) {
            row = rows[ i ];

            keyValuePair = row.split( "=" );

            result[ keyValuePair[ 0 ] ] = keyValuePair[ 1 ];
        }

        var stringifiedResult = JSON.stringify(result);
        console.log("Converted query \"" + query + "\" to json \"" + stringifiedResult + "\"");

        return result;
    }

    function serialize( frm ) {
        var jsonArr = [],
            i, len;

        for ( i = 0, len = frm.elements.length; i < len; i++ ) {
            jsonArr.push({
                name: frm.elements[ i ].name,
                value: frm.elements[ i ].value,
                type: frm.elements[ i ].type
            });
        }

        var serializedForm = JSON.stringify( jsonArr );
        // Can't log json because that might leak a password to the log.

        return serializedForm;
    }

    function generateSuffix() {
        return "?jiveRnd=" + generateUuid() + "jive";
    }

    function addTrackingTokenToUrl( url ) {
        var result = url;

        if ( !kTrackingTestRegExp.test( url ) ) {
            result += generateSuffix();
        } else {
            result = result.replace( kTrackingTestRegExp, generateSuffix() );
        }

        console.log("Added tracking token to url \"" + url + "\" yielding \"" + result +"\"");

        return result;
    }

    function addTrackingTokenToFormAction( frm ) {
        console.log("Adding tracking token to form action");
        frm.action = addTrackingTokenToUrl( frm.action );
        console.log("Done adding tracking token to form action");
    }

    function notifyFormSubmissionWithForm( frm, formActionAugmenter ) {
        var action, json, method, enctype;

        if (formActionAugmenter) {
            formActionAugmenter( frm );
        }

        action = frm.action;
        json = serialize( frm );
        method = frm.method || "GET";
        enctype = frm.enctype || "application/x-www-form-urlencoded";


        // can't log json because that might leak a password into the log
        console.log("Notifying Java about form submission. action: \"" + action + "\", method: \"" + method + "\", enctype: \"" + enctype + "\"");
        window.jiveForAndroid.notifyFormSubmission( action, json, method, enctype );
        console.log("Done notifying Java");
    }

    window.addEventListener( "submit", function( evt ) {
        var frm = evt ? evt.target : this;

        console.log("Notifying in response to a submit event");
        notifyFormSubmissionWithForm( frm, addTrackingTokenToFormAction );
    }, false );

    if ( window.HTMLFormElement ) {
        var cachedSubmit = window.HTMLFormElement.prototype.submit;

        window.HTMLFormElement.prototype.submit = function() {
            console.log("Entering overridden HTMLFormElement.prototype.submit");
            var frm = this;

            console.log("Notifying by overriding HTMLFormElement.prototype.submit");
            notifyFormSubmissionWithForm( frm, addTrackingTokenToFormAction );

            console.log("Calling original HTMLFormElement.prototype.submit");
            return cachedSubmit.apply( frm, arguments );
        };
    } else {
        console.log("!!! window.HTMLFormElement doesn't exist !!!");
    }


    if ( window.XMLHttpRequest ) {
        var cachedSend = XMLHttpRequest.prototype.send,
            cachedOpen = XMLHttpRequest.prototype.open;

        window.XMLHttpRequest.prototype.open = function( method, url, isAsync ) {
            console.log("Entering overridden XMLHttpRequest.prototype.open");

            this.method = method;
            if ( method.toLowerCase() === "post" ) {
                console.log("Adding tracking token to an XMLHttpRequest post");
                this.url = addTrackingTokenToUrl( url );
            } else {
                console.log("Leaving XMLHttpRequest url unmodified");
                this.url = url;
            }

            console.log("Calling original XMLHttpRequest.prototype.open");
            return cachedOpen.call(
                this,
                this.method,
                this.url,
                isAsync
            );
        };

        window.XMLHttpRequest.prototype.send = function( data ) {
            console.log("Entering overridden XMLHttpRequest.prototype.send");
            if ( this.method.toLowerCase() === "post" ) {
                var formData = convertQueryToJson( data ),
                    result = {};

                result.action = this.url;
                result.method = this.method;
                result.elements = [];

                for ( var key in formData ) {
                    if ( formData.hasOwnProperty( key ) ) {
                        result.elements.push({
                            name: key,
                            value: formData[ key ],
                            type: "hidden"
                        });
                    }
                }

                console.log("Notifying by overriding XMLHttpRequest.prototype.send");
                notifyFormSubmissionWithForm( result );
            } else {
                console.log("Not notifying of XMLHttpRequest. url: \"" + this.url + "\", method: \"" + this.method + "\"");
            }

            console.log("Calling original XMLHttpRequest.prototype.send");
            return cachedSend.apply(this, arguments);
        };
    } else {
        console.log("!!! window.XMLHttpRequest doesn't exist !!!");
    }

    if ( window.EventTarget ) {
        var cachedAddEventListener = window.EventTarget.prototype.addEventListener;

        window.EventTarget.prototype.addEventListener = function( name, delegate, shouldUseCapture ) {
            return cachedAddEventListener.call(this, name, function ( evt ) {
                console.log("Entering overridden listener via overridden EventTarget.prototype.addEventListener name:\"" + name + "\"");

                var cachedStopPropagation = evt.stopPropagation,
                    cachedPreventDefault = evt.preventDefault,
                    frm = evt.target,
                    result;

                if ( evt.type === "click" ) {
                    evt.stopPropagation = function() {
                        console.log("stopPropagation called on click event: " + this);
                        return cachedStopPropagation.apply( this, arguments );
                    };

                    evt.preventDefault = function() {
                        console.log("preventDefault called on click event: " + this);
                        return cachedPreventDefault.apply( this, arguments );
                    };
                }

                if ( evt.type === "submit" ) {
                    evt.stopPropagation = function() {
                        console.log("stopPropagation called on submit event: " + this);
                        evt.stopPropagation.called = true;

                        return cachedStopPropagation.apply( this, arguments );
                    };

                    evt.preventDefault = function() {
                        console.log("preventDefault called on submit event: " + this);
                        evt.preventDefault.called = true;

                        return cachedPreventDefault.apply( this, arguments );
                    };
                }

                result = delegate.call( this, evt );

                if ( evt.type === "submit" && evt.stopPropagation.called && !evt.preventDefault.called ) {
                    console.log("Notifying by overridden listener via overridden EventTarget.prototype.addEventListener");
                    notifyFormSubmissionWithForm( frm, addTrackingTokenToFormAction );
                }

                console.log("Exiting overridden listener via overridden EventTarget.prototype.addEventListener name:\"" + name + "\"");

                return result;
            }, shouldUseCapture );
        };
    } else {
        console.log("!!! window.EventTarget doesn't exist !!!");
    }
}());
