(function () {/**
 * @license almond 0.2.9 Copyright (c) 2011-2014, The Dojo Foundation All Rights Reserved.
 * Available via the MIT or new BSD license.
 * see: http://github.com/jrburke/almond for details
 */
//Going sloppy to avoid 'use strict' string cost, but strict practices should
//be followed.
/*jslint sloppy: true */
/*global setTimeout: false */

var requirejs, require, define;
(function (undef) {
    var main, req, makeMap, handlers,
        defined = {},
        waiting = {},
        config = {},
        defining = {},
        hasOwn = Object.prototype.hasOwnProperty,
        aps = [].slice,
        jsSuffixRegExp = /\.js$/;

    function hasProp(obj, prop) {
        return hasOwn.call(obj, prop);
    }

    /**
     * Given a relative module name, like ./something, normalize it to
     * a real name that can be mapped to a path.
     * @param {String} name the relative name
     * @param {String} baseName a real name that the name arg is relative
     * to.
     * @returns {String} normalized name
     */
    function normalize(name, baseName) {
        var nameParts, nameSegment, mapValue, foundMap, lastIndex,
            foundI, foundStarMap, starI, i, j, part,
            baseParts = baseName && baseName.split("/"),
            map = config.map,
            starMap = (map && map['*']) || {};

        //Adjust any relative paths.
        if (name && name.charAt(0) === ".") {
            //If have a base name, try to normalize against it,
            //otherwise, assume it is a top-level require that will
            //be relative to baseUrl in the end.
            if (baseName) {
                //Convert baseName to array, and lop off the last part,
                //so that . matches that "directory" and not name of the baseName's
                //module. For instance, baseName of "one/two/three", maps to
                //"one/two/three.js", but we want the directory, "one/two" for
                //this normalization.
                baseParts = baseParts.slice(0, baseParts.length - 1);
                name = name.split('/');
                lastIndex = name.length - 1;

                // Node .js allowance:
                if (config.nodeIdCompat && jsSuffixRegExp.test(name[lastIndex])) {
                    name[lastIndex] = name[lastIndex].replace(jsSuffixRegExp, '');
                }

                name = baseParts.concat(name);

                //start trimDots
                for (i = 0; i < name.length; i += 1) {
                    part = name[i];
                    if (part === ".") {
                        name.splice(i, 1);
                        i -= 1;
                    } else if (part === "..") {
                        if (i === 1 && (name[2] === '..' || name[0] === '..')) {
                            //End of the line. Keep at least one non-dot
                            //path segment at the front so it can be mapped
                            //correctly to disk. Otherwise, there is likely
                            //no path mapping for a path starting with '..'.
                            //This can still fail, but catches the most reasonable
                            //uses of ..
                            break;
                        } else if (i > 0) {
                            name.splice(i - 1, 2);
                            i -= 2;
                        }
                    }
                }
                //end trimDots

                name = name.join("/");
            } else if (name.indexOf('./') === 0) {
                // No baseName, so this is ID is resolved relative
                // to baseUrl, pull off the leading dot.
                name = name.substring(2);
            }
        }

        //Apply map config if available.
        if ((baseParts || starMap) && map) {
            nameParts = name.split('/');

            for (i = nameParts.length; i > 0; i -= 1) {
                nameSegment = nameParts.slice(0, i).join("/");

                if (baseParts) {
                    //Find the longest baseName segment match in the config.
                    //So, do joins on the biggest to smallest lengths of baseParts.
                    for (j = baseParts.length; j > 0; j -= 1) {
                        mapValue = map[baseParts.slice(0, j).join('/')];

                        //baseName segment has  config, find if it has one for
                        //this name.
                        if (mapValue) {
                            mapValue = mapValue[nameSegment];
                            if (mapValue) {
                                //Match, update name to the new value.
                                foundMap = mapValue;
                                foundI = i;
                                break;
                            }
                        }
                    }
                }

                if (foundMap) {
                    break;
                }

                //Check for a star map match, but just hold on to it,
                //if there is a shorter segment match later in a matching
                //config, then favor over this star map.
                if (!foundStarMap && starMap && starMap[nameSegment]) {
                    foundStarMap = starMap[nameSegment];
                    starI = i;
                }
            }

            if (!foundMap && foundStarMap) {
                foundMap = foundStarMap;
                foundI = starI;
            }

            if (foundMap) {
                nameParts.splice(0, foundI, foundMap);
                name = nameParts.join('/');
            }
        }

        return name;
    }

    function makeRequire(relName, forceSync) {
        return function () {
            //A version of a require function that passes a moduleName
            //value for items that may need to
            //look up paths relative to the moduleName
            return req.apply(undef, aps.call(arguments, 0).concat([relName, forceSync]));
        };
    }

    function makeNormalize(relName) {
        return function (name) {
            return normalize(name, relName);
        };
    }

    function makeLoad(depName) {
        return function (value) {
            defined[depName] = value;
        };
    }

    function callDep(name) {
        if (hasProp(waiting, name)) {
            var args = waiting[name];
            delete waiting[name];
            defining[name] = true;
            main.apply(undef, args);
        }

        if (!hasProp(defined, name) && !hasProp(defining, name)) {
            throw new Error('No ' + name);
        }
        return defined[name];
    }

    //Turns a plugin!resource to [plugin, resource]
    //with the plugin being undefined if the name
    //did not have a plugin prefix.
    function splitPrefix(name) {
        var prefix,
            index = name ? name.indexOf('!') : -1;
        if (index > -1) {
            prefix = name.substring(0, index);
            name = name.substring(index + 1, name.length);
        }
        return [prefix, name];
    }

    /**
     * Makes a name map, normalizing the name, and using a plugin
     * for normalization if necessary. Grabs a ref to plugin
     * too, as an optimization.
     */
    makeMap = function (name, relName) {
        var plugin,
            parts = splitPrefix(name),
            prefix = parts[0];

        name = parts[1];

        if (prefix) {
            prefix = normalize(prefix, relName);
            plugin = callDep(prefix);
        }

        //Normalize according
        if (prefix) {
            if (plugin && plugin.normalize) {
                name = plugin.normalize(name, makeNormalize(relName));
            } else {
                name = normalize(name, relName);
            }
        } else {
            name = normalize(name, relName);
            parts = splitPrefix(name);
            prefix = parts[0];
            name = parts[1];
            if (prefix) {
                plugin = callDep(prefix);
            }
        }

        //Using ridiculous property names for space reasons
        return {
            f: prefix ? prefix + '!' + name : name, //fullName
            n: name,
            pr: prefix,
            p: plugin
        };
    };

    function makeConfig(name) {
        return function () {
            return (config && config.config && config.config[name]) || {};
        };
    }

    handlers = {
        require: function (name) {
            return makeRequire(name);
        },
        exports: function (name) {
            var e = defined[name];
            if (typeof e !== 'undefined') {
                return e;
            } else {
                return (defined[name] = {});
            }
        },
        module: function (name) {
            return {
                id: name,
                uri: '',
                exports: defined[name],
                config: makeConfig(name)
            };
        }
    };

    main = function (name, deps, callback, relName) {
        var cjsModule, depName, ret, map, i,
            args = [],
            callbackType = typeof callback,
            usingExports;

        //Use name if no relName
        relName = relName || name;

        //Call the callback to define the module, if necessary.
        if (callbackType === 'undefined' || callbackType === 'function') {
            //Pull out the defined dependencies and pass the ordered
            //values to the callback.
            //Default to [require, exports, module] if no deps
            deps = !deps.length && callback.length ? ['require', 'exports', 'module'] : deps;
            for (i = 0; i < deps.length; i += 1) {
                map = makeMap(deps[i], relName);
                depName = map.f;

                //Fast path CommonJS standard dependencies.
                if (depName === "require") {
                    args[i] = handlers.require(name);
                } else if (depName === "exports") {
                    //CommonJS module spec 1.1
                    args[i] = handlers.exports(name);
                    usingExports = true;
                } else if (depName === "module") {
                    //CommonJS module spec 1.1
                    cjsModule = args[i] = handlers.module(name);
                } else if (hasProp(defined, depName) ||
                           hasProp(waiting, depName) ||
                           hasProp(defining, depName)) {
                    args[i] = callDep(depName);
                } else if (map.p) {
                    map.p.load(map.n, makeRequire(relName, true), makeLoad(depName), {});
                    args[i] = defined[depName];
                } else {
                    throw new Error(name + ' missing ' + depName);
                }
            }

            ret = callback ? callback.apply(defined[name], args) : undefined;

            if (name) {
                //If setting exports via "module" is in play,
                //favor that over return value and exports. After that,
                //favor a non-undefined return value over exports use.
                if (cjsModule && cjsModule.exports !== undef &&
                        cjsModule.exports !== defined[name]) {
                    defined[name] = cjsModule.exports;
                } else if (ret !== undef || !usingExports) {
                    //Use the return value from the function.
                    defined[name] = ret;
                }
            }
        } else if (name) {
            //May just be an object definition for the module. Only
            //worry about defining if have a module name.
            defined[name] = callback;
        }
    };

    requirejs = require = req = function (deps, callback, relName, forceSync, alt) {
        if (typeof deps === "string") {
            if (handlers[deps]) {
                //callback in this case is really relName
                return handlers[deps](callback);
            }
            //Just return the module wanted. In this scenario, the
            //deps arg is the module name, and second arg (if passed)
            //is just the relName.
            //Normalize module name, if it contains . or ..
            return callDep(makeMap(deps, callback).f);
        } else if (!deps.splice) {
            //deps is a config object, not an array.
            config = deps;
            if (config.deps) {
                req(config.deps, config.callback);
            }
            if (!callback) {
                return;
            }

            if (callback.splice) {
                //callback is an array, which means it is a dependency list.
                //Adjust args if there are dependencies
                deps = callback;
                callback = relName;
                relName = null;
            } else {
                deps = undef;
            }
        }

        //Support require(['a'])
        callback = callback || function () {};

        //If relName is a function, it is an errback handler,
        //so remove it.
        if (typeof relName === 'function') {
            relName = forceSync;
            forceSync = alt;
        }

        //Simulate async callback;
        if (forceSync) {
            main(undef, deps, callback, relName);
        } else {
            //Using a non-zero value because of concern for what old browsers
            //do, and latest browsers "upgrade" to 4 if lower value is used:
            //http://www.whatwg.org/specs/web-apps/current-work/multipage/timers.html#dom-windowtimers-settimeout:
            //If want a value immediately, use require('id') instead -- something
            //that works in almond on the global level, but not guaranteed and
            //unlikely to work in other AMD implementations.
            setTimeout(function () {
                main(undef, deps, callback, relName);
            }, 4);
        }

        return req;
    };

    /**
     * Just drops the config on the floor, but returns req in case
     * the config return value is used.
     */
    req.config = function (cfg) {
        return req(cfg);
    };

    /**
     * Expose module registry for debugging and tooling
     */
    requirejs._defined = defined;

    define = function (name, deps, callback) {

        //This module may not have dependencies
        if (!deps.splice) {
            //deps is not an array, so probably means
            //an object literal or factory function for
            //the value. Adjust args.
            callback = deps;
            deps = [];
        }

        if (!hasProp(defined, name) && !hasProp(waiting, name)) {
            waiting[name] = [name, deps, callback];
        }
    };

    define.amd = {
        jQuery: true
    };
}());

define("../../../../bower_components/almond/almond", function(){});

define( 'config/rte/mixin/app/partials/constants',[],function() {
    

    return {
        EMPTY_HTML_LENGTH: 12,

        /**
         * The browser in Android webview requires at least 300 milliseconds to send
         * the correct state change when `document.queryCommandState`
         * or `document.queryCommand` is called.
         *
         * We've not found a direct link that documents this issue, and it might be
         * related to the [300ms tap delay][tap-delay] in older webkit browsers.
         *
         * [tap-delay]: https://developers.google.com/mobile/articles/fast_buttons
         */
        MIN_STATE_NOTIFY_WAIT_INTERVAL: 500
    };
});

define( 'config/rte/mixin/app/partials/platformtype',[],function() {
    

    return {
        ANDROID: "android",
        IOS: "ios"
    };
});

define( 'config/rte/mixin/app/partials/separator',[],function() {
    

    return {
        COMMAND: "<%@$jive$@%>",
        SERIALIZED_NODE: ":",
        SERIALIZED_BEGIN_COMMENT: "<!--",
        SERIALIZED_END_COMMENT: "-->",
        SERIALIZED_BEGIN_INFO: "<",
        SERIALIZED_END_INFO: ">",
        SERIALIZED_END_INFO_TAG: "</>"
    };
});

define( 'config/rte/mixin/app/partials/content',[],function() {
    

    return {
        TAB_SIZE: 30,
        TAB_MULTIPLIER: 10,
        INDENT_SIZE: 10,
        MAX_INDENT_LIMIT: 10,

        IMAGE_CORNER_PADDING: 7,
        IMAGE_WIDTH: "620px",

        COMPOSITION_MARKER_ID: "JiveMarker",
        DELETION_MARKER_ID: "JiveDeletionMarker"
    };
});

define( 'config/rte/mixin/app/partials/pipe',[],function() {
    

    return {
        TOO_MANY_QUEUED_COMMANDS_THRESHOLD: 20,
        MULTIPLEX_THRESHOLD: 10,
        BATCH_SIZE_DECREASE_THRESHOLD: 3,
        BATCH_MULTIPLEX_POWER_FACTOR: 2
    };
});

define( 'config/rte/mixin/app/partials/scroll',[],function() {
    

    return {
        VERTICAL_TOLERANCE: 0.9,
        HORIZONTAL_TOLERANCE: 0.9
    };
});

define('config/rte/mixin/app/main',[
    "./partials/constants",
    "./partials/platformtype",
    "./partials/separator",
    "./partials/content",
    "./partials/pipe",
    "./partials/scroll"
], function( constants, platformType, separator, content, pipe, scroll) {
    

    return {
        constants: constants,
        platformType: platformType,
        separator: separator,
        content: content,
        pipe: pipe,
        scroll: scroll
    };
});

define( 'config/rte/mixin/dom/partials/html',[],function() {
    

    return {
        GREATER_THAN: "&gt;",
        LESS_THAN: "&lt;",
        NEW_LINE: "<br>",
        PARAGRAPH: "<p></p>",
        MENTION_START: "@",
        SPACE: "\xa0",
        BLANK: "&nbsp;",
        SIZER: (
            "<div class='image-sizer'>" +
            "<div class='handle left top'></div>" +
            "<div class='handle middle top'></div>" +
            "<div class='handle right top'></div>" +
            "<div class='handle left middle'></div>" +
            "<div class='handle right middle'></div>" +
            "<div class='handle left bottom'></div>" +
            "<div class='handle middle bottom'></div>" +
            "<div class='handle right bottom'></div>" +
            "</div>"
            ),
        CONTENT_EDITABLE: "<div contenteditable='true' style='position:absolute;left:-100000px;'></div>"
    };
});

define( 'config/rte/mixin/dom/partials/attribute',[],function() {
    

    return {
        CONTEXT: "__context"
    };
});

define( 'config/rte/mixin/dom/partials/classname',[],function() {
    

    return {
        JIVE_PLACEHOLDER: "jive-placeholder",
        SINGLE_LINE: "single-line",
        RESIZE_DISABLED: "resize-disabled",
        IMAGE_SIZER: "image-sizer"
    };
});

define( 'config/rte/mixin/dom/partials/nodetype',[],function() {
    

    return {
        DOCUMENT_NODE: 9,
        TEXT_NODE: 3,
        COMMENT_NODE: 8
    };
});

define( 'config/rte/mixin/dom/partials/selector',[],function() {
    

    return {

        /**
         * Main editable content.
         */
        RTE: "#jvrtecontent",

        RENDERED_CONTENT: ".jive-rendered-content",
        PLACEHOLDER: "[contenteditable][placeholder]",
        JIVE_PLACEHOLDER: ".jive-placeholder",
        FAKE_IMAGE_SRC: "[src*='webkit-fake-url://']",
        HEADING: "h1,h2,h3,h4,h5,h6",
        LIST: "ul,ol",
        LIST_ITEM: "li",
        PARAGRAPH: "p",
        LINK: "a",
        IMAGE: "img",
        NEW_LINE: "br",
        IMAGE_SIZER: ".image-sizer",
        IMAGE_SIZER_HANDLE: "div.handle",
        SPAN: "span"
    };
});

define('config/rte/mixin/dom/main',[
    "./partials/html",
    "./partials/attribute",
    "./partials/classname",
    "./partials/nodetype",
    "./partials/selector"
], function( html, attribute, className, nodeType, selector ) {
    

    return {
        html: html,
        attribute: attribute,
        className: className,
        nodeType: nodeType,
        selector: selector
    };
});

define( 'config/rte/mixin/key/partials/code',[],function() {
    

    return {
        ENTER: 13,
        BACKSPACE: 8,
        DELETE: 46,
        AT: 64
    };
});

define( 'config/rte/mixin/key/partials/identifier',[],function() {
    

    return {
        AT: "U+0040"
    };
});

define('config/rte/mixin/key/main',[
    "./partials/code",
    "./partials/identifier"
], function( keyCode, keyIdentifier ) {
    

    return {
        keyCode: keyCode,
        keyIdentifier: keyIdentifier
    };
});

define('config/rte/mixin/range/partials/attribute',[],function() {
    

    return {
        CONTAINER: "Container",
        OFFSET: "Offset"
    };
});

define( 'config/rte/mixin/range/partials/prefix',[],function() {
    

    return {
        END: "end",
        BEGIN: "start"
    };
});

define( 'config/rte/mixin/range/partials/properties',[],function() {
    

    return [
        "startContainer",
        "startOffset",
        "endContainer",
        "endOffset",
        "collapsed",
        "commonAncestorContainer",
        "START_TO_START",
        "START_TO_END",
        "END_TO_START",
        "END_TO_END"
    ];
});

define('config/rte/mixin/range/main',[
    "./partials/attribute",
    "./partials/prefix",
    "./partials/properties"
], function( rangeAttribute, rangePrefix, rangeProperties ) {
    

    return {
        rangeAttribute: rangeAttribute,
        rangePrefix: rangePrefix,
        rangeProperties: rangeProperties
    };
});

define('config/rte/mixin/state/main',[],function() {
    

    return {
        state: {
            SCROLL_RECTANGLE: "scrollRectangle",
            FORMAT_BLOCK: "formatBlock",
            INSERTION_RANGE: "insertionRange"
        }
    };
});

define( 'config/rte/mixin/regexp/partials/regexp',[],function() {
    

    return {
        HEADING: /h\d/,
        MENTION: /(^|\W)@(\W|$)/,
        WHITESPACE: /^\s+$/,
        GREATER_THAN: />/g,
        LESS_THAN: /</g,

        PUNCTUATION: /[\.,-\/#!$%\^&\*;:{}=\-_`~()]/,
        TRIM: /^\s+|\s+$/g,
        EVERY_SPACE: / /g,
        EVERY_NEW_LNE: /\n/g,
        ENDS_WITH_WHITESPACE: /[\u00A0\s]$/
    };
});

define('config/rte/mixin/regexp/main',[
    "./partials/regexp"
], function( regExp ) {
    

    return {
        regExp: regExp
    };
});

define( 'config/rte/mixin/topic/partials/native/main',[],function() {
    

    /**
     * From Native, to RTE
     *
     * Topics that are sent from the native side, to the RTE.
     *
     * Sometimes RTE takes the responsibility of publishing these topics, too,
     * by bypassing the native layer completely.
     */

    return {
        RECEIVE_CONTENT: "receiveContent",
        MENTION_START: "mentionStart",
        MENTION_FINISH: "mentionFinish",
        STATE_CHANGE: "stateChange",
        CONTENT_LENGTH: "contentLength",
        CONTENT_CHANGE: "contentChange",
        IMAGE_SELECT: "imageSelect",
        IMAGE_DESELECT: "imageDeselect",
        UPDATE_LEGACY_SELECTION: "updateLegacySelection",
        UPDATE_LEGACY_CLIPBOARD: "updateLegacyClipboard",
        INSERT_NEW_LINE: "insertNewLine",
        READY_STATUS_UPDATE: "readyStatusUpdate",
        REQUEST_FOCUS: "requestFocus",
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/mode',[],function() {
    

    /**
     * @module config.rte.partials.topic.partials.rte.mixin
     * @submodule config.rte.partials.topic.partials.rte.mixin.mode
     */

    /**
     * @class config.rte.partials.topic.partials.rte.mixin.mode
     * @static
     */

    return {

        /**
         * @property SET_SINGLE_LINE
         * @final
         */
        SET_SINGLE_LINE: "setSingleLine"
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/content',[],function() {
    

    return {
        SEND_CONTENT: "sendContent",
        REPLACE_CONTENT: "replaceContent",

        UPDATE_ATTACHED_IMAGE_COUNT: "updateAttachedImageCount"
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/block',[],function() {
    

    return {
        DEFAULT_PARAGRAPH_SEPARATOR: "defaultParagraphSeparator",
        FORMAT_BLOCK: "formatBlock"
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/format',[],function() {
    

    return {

        // TODO: get rid of these and use standard names, like "bold", "italic" etc.
        BOLD: "bold",
        ITALIC: "italic",
        UNDERLINE: "underline",

        SET_TEXT_COLOR: "setTextColor"
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/insertion',[],function() {
    

    return {
        INSERT_NEW_LINE: "insertNewLine",
        INSERT_BACKSPACE: "insertBackspace",
        INSERT_IMAGE: "insertImage",
        INSERT_HTML: "insertHtml"
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/legacy',[],function() {
    

    return {
        SET_SELECTABLEVIEW_END_POSITION: "setSelectableViewEndPosition",
        SET_SELECTABLEVIEW_START_POSITION: "setSelectableViewStartPosition",
        SET_SELECTABLEVIEW_SELECTION: "setSelectableViewSelection",
        CLEAR_SELECTABLEVIEW_SELECTION: "clearSelectableViewSelection",
        START_SELECTABLEVIEW_LONG_TOUCH: "startSelectableViewLongTouch",
        START_SELECTABLEVIEW_TOUCH: "startSelectableViewTouch",
        START_SELECTABLEVIEW_CUT: "startSelectableViewCut",
        START_SELECTABLEVIEW_COPY: "startSelectableViewCopy",
        START_SELECTABLEVIEW_PASTE: "startSelectableViewPaste"
    };
});
define('config/rte/mixin/topic/partials/rte/mixin/list',[],function() {
    

    return {
        INDENT: "indent",
        OUTDENT: "outdent",
        INSERT_ORDERED_LIST: "insertOrderedList",
        INSERT_UNORDERED_LIST: "insertUnorderedList",
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/mention',[],function() {
    

    return {
        INSERT_AT_AND_START_MENTION: "insertAtAndStartMention",
        MENTION: "mention",
        INSERT_MENTION: "insertMention"
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/sanitization',[],function() {
    

    return {
        ENSURE_CONTENT_EMPTY: "ensureContentEmpty"
    };
});

define('config/rte/mixin/topic/partials/rte/mixin/text',[],function() {
    

    return {
        DELETE_SURROUNDING_TEXT: "deleteSurroundingText",
        SET_COMPOSING_TEXT:  "setComposingText",
        SET_COMPOSING_REGION: "setComposingRegion",
        FINISH_COMPOSING_TEXT: "finishComposingText",
        COMMIT_TEXT: "commitText"
    };
});

define('util/object',[],function() {
    

    var exports = {};

    exports.extend = function( base, extension ) {
        var key;

        for ( key in extension ) {
            if ( extension.hasOwnProperty(key) ) {
                base[ key ] = extension[ key ];
            }
        }
    };

    return exports;
});

define('config/rte/mixin/topic/partials/rte/main',[
    "./mixin/mode",
    "./mixin/content",
    "./mixin/block",
    "./mixin/format",
    "./mixin/insertion",
    "./mixin/legacy",
    "./mixin/list",
    "./mixin/mention",
    "./mixin/sanitization",
    "./mixin/text",
    "../../../../../../util/object"
], function( mode, content, block, format, insertion, legacy, list, mention, sanitization, text, objectUtil ) {
    

    /**
     * From RTE to Native
     *
     * Topics that are sent from RTE to the native side.
     */

    var result = {},

        extend = objectUtil.extend;

    extend( result, mode );
    extend( result, content );
    extend( result, block );
    extend( result, format );
    extend( result, insertion );
    extend( result, legacy );
    extend( result, list );
    extend( result, mention );
    extend( result, sanitization );
    extend( result, text );

    return result;
});

define( 'config/rte/mixin/topic/main',[
    "./partials/native/main",
    "./partials/rte/main"
], function( jiveNative, rte ) {
    

    return {
        topic: {

            // "native" is a reserved word that Android 4.0.4 JavaScript engine does not like.
            "native": jiveNative,
            rte: rte
        }
    };
});

define( 'config/rte/main',[
    "./mixin/app/main",
    "./mixin/dom/main",
    "./mixin/key/main",
    "./mixin/range/main",
    "./mixin/state/main",
    "./mixin/regexp/main",
    "./mixin/topic/main",
    "../../util/object"
], function( app, dom, key, range, state, regExp, topic, objectUtil ) {
    

    var result = {},
        extend = objectUtil.extend;

    extend( result, app );
    extend( result, dom );
    extend( result, key );
    extend( result, range );
    extend( result, state );
    extend( result, regExp );
    extend( result, topic );

    return result;
});

define( 'config/native/main',[],function() {
    

    var exports = {};

    exports.NATIVE_NAME = "JiveNative";
    exports.RTE_NAME = "JiveRte";

    /**
     * Defaults to android.
     * Overridden at build time.
     */
    exports.PLATFORM = "android";

    return exports;
});

define( 'util/timer',[
    "require"
], function( require ) {
    

    var exports = {};

    function memoizedBridge() {
        /*jshint validthis: true*/

        if ( !this.memo ) {
            this.memo = require( "../io/bridge" );
        }

        return this.memo;
    }

    function getBridge() {
        return memoizedBridge.bind(memoizedBridge)();
    }

    /**
     * Use this instead of `window.setTimeout`.
     */
    exports.setTimeout = function( delegate, time ) {
        return getBridge().setTimeout( delegate, time );
    };

    /**
     * Use this instead of `window.clearTimeout`.
     */
    exports.clearTimeout = function( id ) {
        getBridge().clearTimeout( id );
    };

    return exports;
});

define( 'native/android',[
    "../config/native/main",
    "../util/timer"
], function( nativeConfig, timer ) {
    

    nativeConfig.PLATFORM = "android";

    var exports = {},
        kNativeHook = nativeConfig.NATIVE_NAME;

    function getJson( json ) {
        return typeof json === "object" ? JSON.stringify( json ) : json.toString();
    }

    exports.isReady = function() {
        var droid = window[ kNativeHook ];

        return !!droid;
    };

    /**
     * Returns a semicolon delimited command string.
     */
    exports.getPendingCommands = function() {
        var droid = window[ kNativeHook ];

        if ( !droid ) {
            exports.log( "native/android/getPendingCommands(): Droid native is not ready!" );

            return "";
        }

        // This should be defined on Device.java with an identical name.
        //
        // Also note that calling droid.getPendingCommands a second time will
        // return an empty string (i.e. droid.getPendingCommands has a side
        // effect/feature of draining its command queue upon each call)
        return droid.getPendingCommands();
    };

    /**
     * Audit log to `JiveNative`s console`.
     *
     * This is a sync call to the native side, the native might decide to log it
     * async in a separate thread.
     */
    exports.log = function() {

        // For #ANDROID-353: Using console.log, since Java already intercepts console.
        window.console.log( Array.prototype.join.call( arguments, "," ) );
    };

    exports.notify = function( json ) {
        if ( !json ) {
            return;
        }

        var droid = window[ kNativeHook ];

        if ( !droid ) {
            exports.log( "native/android/notify(): Droid native is not ready!" );

            return;
        }

        // To keep the API identical cross-device, communication TO the native land should always be async.
        timer.setTimeout( function() {
            droid.notify( getJson( json ) );
        });
    };

    exports.isLegacy = function() {
        var droid = window[ kNativeHook ];

        if ( !droid ) {
            exports.log( "native/android/isLegacy(): Droid native is not ready!" );

            return false;
        }

        return droid.isLegacy();
    };

    // TODO: this is not in the standard contract, check where it is used.
    exports.completeCommand = function( uuid, result ) {
        var droid = window[ kNativeHook ];

        if ( !droid ) {
            exports.log( "native/android/isLegacy(): Droid native is not ready!" );

            return false;
        }

        timer.setTimeout(function() {
            droid.completeCommand( uuid, result );
        });
    };

    return exports;
});

define( 'native/factory',[
    // Default platform is android for tests and all.
    // That module id gets replaced at build time.
    "./android"
], function( platform ) {
    

    var exports = {};

    exports.createInstance = function() {
        return platform;
    };

    return exports;
});

define( 'util/log',[
    "../native/factory"
], function( nativeFactory ) {
    

    var exports = {},

        jiveNative = nativeFactory.createInstance();

    exports.logError = function( error, payload ) {
        var key, value;

        if ( !error ) {
            return;
        }

        jiveNative.log( "error: " + error.toString() );
        jiveNative.log( "stack: " + error.stack );
        jiveNative.log( "message: " + error.message );

        if ( payload ) {
            for ( key in payload ) {
                if ( payload.hasOwnProperty( key ) ) {
                    value = payload[ key ];

                    jiveNative.log( key + ": `" + value + "`." );

                    if ( key === "delegate" ) {
                        jiveNative.log( "delegate:" );
                        jiveNative.log( value.toString() );
                    }
                }
            }
        }
    };

    exports.log = function( stuff ) {
        jiveNative.log( stuff );
    };

    exports.dump = function( obj ) {
        var key;

        for ( key in obj ) {
            if ( obj.hasOwnProperty( key ) ) {
                jiveNative.log( "'" + key + ":'" + obj[key] + "'");
            }
        }
    };

    return exports;
});

define( 'io/hub',[
    "../util/log"
], function( log ) {
    

    var exports = {},
        listenersByTopics = {};

    function add( topic, listener ) {
        if ( !listenersByTopics[ topic ] ) {
            listenersByTopics[ topic ] = [];
        }

        var listeners = listenersByTopics[ topic ];

        listeners.push( listener );
    }

    exports.publish = function( topic, payload ) {
        if ( !topic ) {
            throw new Error( "PUBLISH: topic must be a non-empty string" );
        }

        if ( listenersByTopics[ topic ] ) {
            var listeners = listenersByTopics[ topic ],
                i, len;

            for ( i = 0, len = listeners.length; i < len; i++ ) {
                try {
                    listeners[ i ].apply( this, payload );
                } catch ( ex ) {
                    log.logError( ex );
                }
            }
        }
    };

    exports.subscribe = function( topic, listener ) {
        if ( !topic ) {
            throw new Error( "SUBSCRIBE: topic must be a non-empty string." );
        }

        if ( !listener ) {
            throw new Error( "listener cannot be undefined." );
        }

        if ( typeof listener !== "function" ) {
            throw new Error( "listener must be a function." );
        }

        add( topic, listener );
    };

    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_getListeners = function() {
        return listenersByTopics;
    };
    exports._test_add = function( topic, listener ) {
        add( topic, listener );
    };
    exports._test_reset = function() {
        listenersByTopics = {};
    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

define( 'io/meta',[
    "../util/log",
    "./hub",
    "../native/factory"
], function( log, hub, nativeFactory ) {
    

    var exports = {},

        jiveNative = nativeFactory.createInstance();

    function getMetaInfoFromQueueItem( item ) {
        if ( !item ) {
            return null;
        }

        var parsed;

        try {
            parsed = typeof item === "string" ? JSON.parse( item ) : item;
        } catch (ex) {
            /*--(log.log( "getMetaInforFromQueueItem: Problem parsing command: '" + item + "'" );)--*/

            return null;
        }

        // See: https://jira.jivesoftware.com/browse/ANDROID-777
        return { command: parsed.command, args: parsed.args, delegate: parsed.delegate, id: parsed.id };
    }

    exports.executeCommand = function( command ) {
        var meta = getMetaInfoFromQueueItem( command );

        if ( meta ) {

            // `proxy/command.js` subscribes to these commands.
            if ( meta.command ) {
                /*--(log.log( "/bridge/loop(): Publishing queued command..." + meta.command );)--*/

                hub.publish( meta.command, meta.args );
            }

            // Used for `setTimeout`s.
            if ( meta.delegate ) {
                /*--(log.log( "/bridge/loop(): Executing queued delegate... " + meta.delegate.toString().substring(0, 100).replace( /\n/, " " ) );)--*/

                meta.delegate();
            }

            // See: https://jira.jivesoftware.com/browse/ANDROID-777
            if ( meta.id ) {
                /*--(log.log( "/bridge/loop(): Notifying complete for ID: " + meta.id );)--*/

                jiveNative.completeCommand( String(meta.id), null );
            }
        }
    };

    return exports;
});

define( 'io/stream',[
    "../config/rte/main",
    "./meta"
], function( rteConfig, meta ) {
    

    // TODO: this module is too tightly-coupled with the bridge.js (for instance it takes commandQueue (which is private to bridge.js)  by ref, and modifies it)
    // This might suggest that we may need to merge it with bridge.js code (maybe creating a closure inside bridge.js so that this code is still isolated in
    // itself -- like an inner class.)

    var exports = {},
        hits = 0,
        misses = 0;

    function getNextCommand( commandQueue ) {
        return commandQueue.shift();
    }

    function delegateCommand( command ) {
        if ( !command ) {
            return false;
        }

        meta.executeCommand( command );

        return true;
    }

    function delegateNextCommand( commandQueue ) {
        return delegateCommand( getNextCommand( commandQueue ) );
    }

    function multiplex( commandQueue ) {
        var len = Math.pow( rteConfig.pipe.BATCH_MULTIPLEX_POWER_FACTOR, misses ),
            i;

        for ( i = 0; i < len; i++ ) {
            if ( !delegateNextCommand( commandQueue ) ) {
                break;
            }
        }
    }

    exports.multiplex = function( commandQueue ) {
        if ( commandQueue.length > rteConfig.pipe.MULTIPLEX_THRESHOLD ) {
            hits = 0;
            misses++;

            multiplex( commandQueue );

            return true;
        }

        return false;
    };

    exports.adjustHitCount = function() {
        if ( misses <= 0 ) {
            return;
        }

        hits++;

        if ( hits >= rteConfig.pipe.BATCH_SIZE_DECREASE_THRESHOLD ) {
            misses--;
            hits = 0;
        }
    };

    return exports;
});

define( 'io/bridge',[
    "../config/rte/main",
    "../io/hub",
    "../native/factory",
    "../util/log",
    "./meta",
    "./stream"
], function( rteConfig, hub, nativeFactory, log, meta, stream ) {
    

    var exports = {},

        /**
         * A Unified native interface.
         */
        jiveNative = nativeFactory.createInstance(),

        /**
         * RAF polyfill.
         */
        tick = window.requestAnimationFrame ||
        window.mozRequestAnimationFrame ||
        window.webkitRequestAnimationFrame ||
        window.msRequestAnimationFrame ||
        function( delegate ) {
            return window.setTimeout( delegate, 17 );
        },

        untick = window.cancelRequestAnimationFrame ||
        window.mozCancelRequestAnimationFrame ||
        window.webkitCancelRequestAnimationFrame ||
        window.msCancelRequestAnimationFrame ||
        window.clearTimeout,

        lastId,

        commandQueue = [],

        kCommandSeparator = rteConfig.separator.COMMAND;

    /**
     *
     */
    function queuePendingCommands() {

        // `native.getPendingCommands` returns a ";"-delimited string of "JSON"s.
        var pending = jiveNative.getPendingCommands(),
            i, len = 0, commands, command;

        if ( pending ) {
            commands = pending.split( kCommandSeparator );

            for ( i = 0, len = commands.length; i < len; i++ ) {
                command = commands[ i ];

                if ( command ) {
                    commandQueue.push( command );
                }
            }
        }

        return len;
    }

    /**
     * Main event pipe.
     */
    function loop() {
        lastId = tick( loop );

        if ( !jiveNative.isReady() ) {
            return;
        }

        var len;

        // If job queue is congested, we adaptively batch more than one jobs and run them together.
        if ( commandQueue.length && stream.multiplex( commandQueue ) ) {
            return;
        }

        stream.adjustHitCount();

        // Queue the commands that are pushed from the native land.
        len = queuePendingCommands();

        if ( len ) {
            /*--(log.log( "Native pushed commands. Will process them in the next cycle." );)--*/

            return;
        }

        // So far, so good.
        if ( commandQueue.length ) {
            meta.executeCommand( commandQueue.shift() );
        }
    }

    /**
     * This method is **protected** (*meaning: it is intended to be called by
     * only a certain set of modules*).
     *
     * ONLY `util/timer` module should call this.
     */
    exports.setTimeout = function( delegate, timeout ) {
        if ( !delegate ) {
            throw new Error( "/io/bridge/enqueue: Delegate method is not defined!" );
        }

        return window.setTimeout( function() {

            // type:fn is useful for debugging.:
            commandQueue.push({ delegate: delegate, type: "fn" });
        }, timeout || 0 );
    };

    /**
     * This method is **protected** (*meaning: it is intended to be called by
     * only a certain set of modules*).
     *
     * ONLY `util/timer` module should call this.
     */
    exports.clearTimeout = function( id ) {
        window.clearTimeout( id );
    };

    /**
     * Instead of calling document.execCommand from the client, call this.
     *
     * Calling this method will push the command to the `/proxy/command.js` pipeline
     * just as if it has been called from the native side.
     *
     * This is a shortcut that the rte uses, to bypass going back and forth the native bridge.
     * The command is still executed async.
     *
     * If for some reason, the native code needs to get notified first, then use
     * `native.notify` instead.
     *
     * Note that `native.notify` is also guaranteed to be async, so that it does not block any rendering.
     */
    exports.enqueueCommand = function( commandString ) {
        var command = typeof commandString === "string" ? commandString : JSON.stringify( commandString );

        /*--(log.log( "rte/bridge/enqueueCommand: `" + command + "`" );)--*/

        commandQueue.push( command );
    };

    exports.initialize = function( newNative ) {
        if ( newNative ) {
            jiveNative = newNative;
        }

        loop();

        // Don't let anyone initialize it again.
        this.initialize = function() {};
    };

    exports.getCommandQueue = function() {
        return commandQueue;
    };

    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_resetState = function() {
        untick( lastId );

        exports.initialize = function( newNative ) {
            if ( newNative ) {
                jiveNative = newNative;
            }

            loop();
        };
    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

/* Zepto v1.1.3 - zepto event ajax form ie - zeptojs.com/license */


var Zepto = (function() {
  var undefined, key, $, classList, emptyArray = [], slice = emptyArray.slice, filter = emptyArray.filter,
    document = window.document,
    elementDisplay = {}, classCache = {},
    cssNumber = { 'column-count': 1, 'columns': 1, 'font-weight': 1, 'line-height': 1,'opacity': 1, 'z-index': 1, 'zoom': 1 },
    fragmentRE = /^\s*<(\w+|!)[^>]*>/,
    singleTagRE = /^<(\w+)\s*\/?>(?:<\/\1>|)$/,
    tagExpanderRE = /<(?!area|br|col|embed|hr|img|input|link|meta|param)(([\w:]+)[^>]*)\/>/ig,
    rootNodeRE = /^(?:body|html)$/i,
    capitalRE = /([A-Z])/g,

    // special attributes that should be get/set via method calls
    methodAttributes = ['val', 'css', 'html', 'text', 'data', 'width', 'height', 'offset'],

    adjacencyOperators = [ 'after', 'prepend', 'before', 'append' ],
    table = document.createElement('table'),
    tableRow = document.createElement('tr'),
    containers = {
      'tr': document.createElement('tbody'),
      'tbody': table, 'thead': table, 'tfoot': table,
      'td': tableRow, 'th': tableRow,
      '*': document.createElement('div')
    },
    readyRE = /complete|loaded|interactive/,
    simpleSelectorRE = /^[\w-]*$/,
    class2type = {},
    toString = class2type.toString,
    zepto = {},
    camelize, uniq,
    tempParent = document.createElement('div'),
    propMap = {
      'tabindex': 'tabIndex',
      'readonly': 'readOnly',
      'for': 'htmlFor',
      'class': 'className',
      'maxlength': 'maxLength',
      'cellspacing': 'cellSpacing',
      'cellpadding': 'cellPadding',
      'rowspan': 'rowSpan',
      'colspan': 'colSpan',
      'usemap': 'useMap',
      'frameborder': 'frameBorder',
      'contenteditable': 'contentEditable'
    },
    isArray = Array.isArray ||
      function(object){ return object instanceof Array }

  zepto.matches = function(element, selector) {
    if (!selector || !element || element.nodeType !== 1) return false
    var matchesSelector = element.webkitMatchesSelector || element.mozMatchesSelector ||
                          element.oMatchesSelector || element.matchesSelector
    if (matchesSelector) return matchesSelector.call(element, selector)
    // fall back to performing a selector:
    var match, parent = element.parentNode, temp = !parent
    if (temp) (parent = tempParent).appendChild(element)
    match = ~zepto.qsa(parent, selector).indexOf(element)
    temp && tempParent.removeChild(element)
    return match
  }

  function type(obj) {
    return obj == null ? String(obj) :
      class2type[toString.call(obj)] || "object"
  }

  function isFunction(value) { return type(value) == "function" }
  function isWindow(obj)     { return obj != null && obj == obj.window }
  function isDocument(obj)   { return obj != null && obj.nodeType == obj.DOCUMENT_NODE }
  function isObject(obj)     { return type(obj) == "object" }
  function isPlainObject(obj) {
    return isObject(obj) && !isWindow(obj) && Object.getPrototypeOf(obj) == Object.prototype
  }
  function likeArray(obj) { return typeof obj.length == 'number' }

  function compact(array) { return filter.call(array, function(item){ return item != null }) }
  function flatten(array) { return array.length > 0 ? $.fn.concat.apply([], array) : array }
  camelize = function(str){ return str.replace(/-+(.)?/g, function(match, chr){ return chr ? chr.toUpperCase() : '' }) }
  function dasherize(str) {
    return str.replace(/::/g, '/')
           .replace(/([A-Z]+)([A-Z][a-z])/g, '$1_$2')
           .replace(/([a-z\d])([A-Z])/g, '$1_$2')
           .replace(/_/g, '-')
           .toLowerCase()
  }
  uniq = function(array){ return filter.call(array, function(item, idx){ return array.indexOf(item) == idx }) }

  function classRE(name) {
    return name in classCache ?
      classCache[name] : (classCache[name] = new RegExp('(^|\\s)' + name + '(\\s|$)'))
  }

  function maybeAddPx(name, value) {
    return (typeof value == "number" && !cssNumber[dasherize(name)]) ? value + "px" : value
  }

  function defaultDisplay(nodeName) {
    var element, display
    if (!elementDisplay[nodeName]) {
      element = document.createElement(nodeName)
      document.body.appendChild(element)
      display = getComputedStyle(element, '').getPropertyValue("display")
      element.parentNode.removeChild(element)
      display == "none" && (display = "block")
      elementDisplay[nodeName] = display
    }
    return elementDisplay[nodeName]
  }

  function children(element) {
    return 'children' in element ?
      slice.call(element.children) :
      $.map(element.childNodes, function(node){ if (node.nodeType == 1) return node })
  }

  // `$.zepto.fragment` takes a html string and an optional tag name
  // to generate DOM nodes nodes from the given html string.
  // The generated DOM nodes are returned as an array.
  // This function can be overriden in plugins for example to make
  // it compatible with browsers that don't support the DOM fully.
  zepto.fragment = function(html, name, properties) {
    var dom, nodes, container

    // A special case optimization for a single tag
    if (singleTagRE.test(html)) dom = $(document.createElement(RegExp.$1))

    if (!dom) {
      if (html.replace) html = html.replace(tagExpanderRE, "<$1></$2>")
      if (name === undefined) name = fragmentRE.test(html) && RegExp.$1
      if (!(name in containers)) name = '*'

      container = containers[name]
      container.innerHTML = '' + html
      dom = $.each(slice.call(container.childNodes), function(){
        container.removeChild(this)
      })
    }

    if (isPlainObject(properties)) {
      nodes = $(dom)
      $.each(properties, function(key, value) {
        if (methodAttributes.indexOf(key) > -1) nodes[key](value)
        else nodes.attr(key, value)
      })
    }

    return dom
  }

  // `$.zepto.Z` swaps out the prototype of the given `dom` array
  // of nodes with `$.fn` and thus supplying all the Zepto functions
  // to the array. Note that `__proto__` is not supported on Internet
  // Explorer. This method can be overriden in plugins.
  zepto.Z = function(dom, selector) {
    dom = dom || []
    dom.__proto__ = $.fn
    dom.selector = selector || ''
    return dom
  }

  // `$.zepto.isZ` should return `true` if the given object is a Zepto
  // collection. This method can be overriden in plugins.
  zepto.isZ = function(object) {
    return object instanceof zepto.Z
  }

  // `$.zepto.init` is Zepto's counterpart to jQuery's `$.fn.init` and
  // takes a CSS selector and an optional context (and handles various
  // special cases).
  // This method can be overriden in plugins.
  zepto.init = function(selector, context) {
    var dom
    // If nothing given, return an empty Zepto collection
    if (!selector) return zepto.Z()
    // Optimize for string selectors
    else if (typeof selector == 'string') {
      selector = selector.trim()
      // If it's a html fragment, create nodes from it
      // Note: In both Chrome 21 and Firefox 15, DOM error 12
      // is thrown if the fragment doesn't begin with <
      if (selector[0] == '<' && fragmentRE.test(selector))
        dom = zepto.fragment(selector, RegExp.$1, context), selector = null
      // If there's a context, create a collection on that context first, and select
      // nodes from there
      else if (context !== undefined) return $(context).find(selector)
      // If it's a CSS selector, use it to select nodes.
      else dom = zepto.qsa(document, selector)
    }
    // If a function is given, call it when the DOM is ready
    else if (isFunction(selector)) return $(document).ready(selector)
    // If a Zepto collection is given, just return it
    else if (zepto.isZ(selector)) return selector
    else {
      // normalize array if an array of nodes is given
      if (isArray(selector)) dom = compact(selector)
      // Wrap DOM nodes.
      else if (isObject(selector))
        dom = [selector], selector = null
      // If it's a html fragment, create nodes from it
      else if (fragmentRE.test(selector))
        dom = zepto.fragment(selector.trim(), RegExp.$1, context), selector = null
      // If there's a context, create a collection on that context first, and select
      // nodes from there
      else if (context !== undefined) return $(context).find(selector)
      // And last but no least, if it's a CSS selector, use it to select nodes.
      else dom = zepto.qsa(document, selector)
    }
    // create a new Zepto collection from the nodes found
    return zepto.Z(dom, selector)
  }

  // `$` will be the base `Zepto` object. When calling this
  // function just call `$.zepto.init, which makes the implementation
  // details of selecting nodes and creating Zepto collections
  // patchable in plugins.
  $ = function(selector, context){
    return zepto.init(selector, context)
  }

  function extend(target, source, deep) {
    for (key in source)
      if (deep && (isPlainObject(source[key]) || isArray(source[key]))) {
        if (isPlainObject(source[key]) && !isPlainObject(target[key]))
          target[key] = {}
        if (isArray(source[key]) && !isArray(target[key]))
          target[key] = []
        extend(target[key], source[key], deep)
      }
      else if (source[key] !== undefined) target[key] = source[key]
  }

  // Copy all but undefined properties from one or more
  // objects to the `target` object.
  $.extend = function(target){
    var deep, args = slice.call(arguments, 1)
    if (typeof target == 'boolean') {
      deep = target
      target = args.shift()
    }
    args.forEach(function(arg){ extend(target, arg, deep) })
    return target
  }

  // `$.zepto.qsa` is Zepto's CSS selector implementation which
  // uses `document.querySelectorAll` and optimizes for some special cases, like `#id`.
  // This method can be overriden in plugins.
  zepto.qsa = function(element, selector){
    var found,
        maybeID = selector[0] == '#',
        maybeClass = !maybeID && selector[0] == '.',
        nameOnly = maybeID || maybeClass ? selector.slice(1) : selector, // Ensure that a 1 char tag name still gets checked
        isSimple = simpleSelectorRE.test(nameOnly)
    return (isDocument(element) && isSimple && maybeID) ?
      ( (found = element.getElementById(nameOnly)) ? [found] : [] ) :
      (element.nodeType !== 1 && element.nodeType !== 9) ? [] :
      slice.call(
        isSimple && !maybeID ?
          maybeClass ? element.getElementsByClassName(nameOnly) : // If it's simple, it could be a class
          element.getElementsByTagName(selector) : // Or a tag
          element.querySelectorAll(selector) // Or it's not simple, and we need to query all
      )
  }

  function filtered(nodes, selector) {
    return selector == null ? $(nodes) : $(nodes).filter(selector)
  }

  $.contains = function(parent, node) {
    return parent !== node && parent.contains(node)
  }

  function funcArg(context, arg, idx, payload) {
    return isFunction(arg) ? arg.call(context, idx, payload) : arg
  }

  function setAttribute(node, name, value) {
    value == null ? node.removeAttribute(name) : node.setAttribute(name, value)
  }

  // access className property while respecting SVGAnimatedString
  function className(node, value){
    var klass = node.className,
        svg   = klass && klass.baseVal !== undefined

    if (value === undefined) return svg ? klass.baseVal : klass
    svg ? (klass.baseVal = value) : (node.className = value)
  }

  // "true"  => true
  // "false" => false
  // "null"  => null
  // "42"    => 42
  // "42.5"  => 42.5
  // "08"    => "08"
  // JSON    => parse if valid
  // String  => self
  function deserializeValue(value) {
    var num
    try {
      return value ?
        value == "true" ||
        ( value == "false" ? false :
          value == "null" ? null :
          !/^0/.test(value) && !isNaN(num = Number(value)) ? num :
          /^[\[\{]/.test(value) ? $.parseJSON(value) :
          value )
        : value
    } catch(e) {
      return value
    }
  }

  $.type = type
  $.isFunction = isFunction
  $.isWindow = isWindow
  $.isArray = isArray
  $.isPlainObject = isPlainObject

  $.isEmptyObject = function(obj) {
    var name
    for (name in obj) return false
    return true
  }

  $.inArray = function(elem, array, i){
    return emptyArray.indexOf.call(array, elem, i)
  }

  $.camelCase = camelize
  $.trim = function(str) {
    return str == null ? "" : String.prototype.trim.call(str)
  }

  // plugin compatibility
  $.uuid = 0
  $.support = { }
  $.expr = { }

  $.map = function(elements, callback){
    var value, values = [], i, key
    if (likeArray(elements))
      for (i = 0; i < elements.length; i++) {
        value = callback(elements[i], i)
        if (value != null) values.push(value)
      }
    else
      for (key in elements) {
        value = callback(elements[key], key)
        if (value != null) values.push(value)
      }
    return flatten(values)
  }

  $.each = function(elements, callback){
    var i, key
    if (likeArray(elements)) {
      for (i = 0; i < elements.length; i++)
        if (callback.call(elements[i], i, elements[i]) === false) return elements
    } else {
      for (key in elements)
        if (callback.call(elements[key], key, elements[key]) === false) return elements
    }

    return elements
  }

  $.grep = function(elements, callback){
    return filter.call(elements, callback)
  }

  if (window.JSON) $.parseJSON = JSON.parse

  // Populate the class2type map
  $.each("Boolean Number String Function Array Date RegExp Object Error".split(" "), function(i, name) {
    class2type[ "[object " + name + "]" ] = name.toLowerCase()
  })

  // Define methods that will be available on all
  // Zepto collections
  $.fn = {
    // Because a collection acts like an array
    // copy over these useful array functions.
    forEach: emptyArray.forEach,
    reduce: emptyArray.reduce,
    push: emptyArray.push,
    sort: emptyArray.sort,
    indexOf: emptyArray.indexOf,
    concat: emptyArray.concat,

    // `map` and `slice` in the jQuery API work differently
    // from their array counterparts
    map: function(fn){
      return $($.map(this, function(el, i){ return fn.call(el, i, el) }))
    },
    slice: function(){
      return $(slice.apply(this, arguments))
    },

    ready: function(callback){
      // need to check if document.body exists for IE as that browser reports
      // document ready when it hasn't yet created the body element
      if (readyRE.test(document.readyState) && document.body) callback($)
      else document.addEventListener('DOMContentLoaded', function(){ callback($) }, false)
      return this
    },
    get: function(idx){
      return idx === undefined ? slice.call(this) : this[idx >= 0 ? idx : idx + this.length]
    },
    toArray: function(){ return this.get() },
    size: function(){
      return this.length
    },
    remove: function(){
      return this.each(function(){
        if (this.parentNode != null)
          this.parentNode.removeChild(this)
      })
    },
    each: function(callback){
      emptyArray.every.call(this, function(el, idx){
        return callback.call(el, idx, el) !== false
      })
      return this
    },
    filter: function(selector){
      if (isFunction(selector)) return this.not(this.not(selector))
      return $(filter.call(this, function(element){
        return zepto.matches(element, selector)
      }))
    },
    add: function(selector,context){
      return $(uniq(this.concat($(selector,context))))
    },
    is: function(selector){
      return this.length > 0 && zepto.matches(this[0], selector)
    },
    not: function(selector){
      var nodes=[]
      if (isFunction(selector) && selector.call !== undefined)
        this.each(function(idx){
          if (!selector.call(this,idx)) nodes.push(this)
        })
      else {
        var excludes = typeof selector == 'string' ? this.filter(selector) :
          (likeArray(selector) && isFunction(selector.item)) ? slice.call(selector) : $(selector)
        this.forEach(function(el){
          if (excludes.indexOf(el) < 0) nodes.push(el)
        })
      }
      return $(nodes)
    },
    has: function(selector){
      return this.filter(function(){
        return isObject(selector) ?
          $.contains(this, selector) :
          $(this).find(selector).size()
      })
    },
    eq: function(idx){
      return idx === -1 ? this.slice(idx) : this.slice(idx, + idx + 1)
    },
    first: function(){
      var el = this[0]
      return el && !isObject(el) ? el : $(el)
    },
    last: function(){
      var el = this[this.length - 1]
      return el && !isObject(el) ? el : $(el)
    },
    find: function(selector){
      var result, $this = this
      if (typeof selector == 'object')
        result = $(selector).filter(function(){
          var node = this
          return emptyArray.some.call($this, function(parent){
            return $.contains(parent, node)
          })
        })
      else if (this.length == 1) result = $(zepto.qsa(this[0], selector))
      else result = this.map(function(){ return zepto.qsa(this, selector) })
      return result
    },
    closest: function(selector, context){
      var node = this[0], collection = false
      if (typeof selector == 'object') collection = $(selector)
      while (node && !(collection ? collection.indexOf(node) >= 0 : zepto.matches(node, selector)))
        node = node !== context && !isDocument(node) && node.parentNode
      return $(node)
    },
    parents: function(selector){
      var ancestors = [], nodes = this
      while (nodes.length > 0)
        nodes = $.map(nodes, function(node){
          if ((node = node.parentNode) && !isDocument(node) && ancestors.indexOf(node) < 0) {
            ancestors.push(node)
            return node
          }
        })
      return filtered(ancestors, selector)
    },
    parent: function(selector){
      return filtered(uniq(this.pluck('parentNode')), selector)
    },
    children: function(selector){
      return filtered(this.map(function(){ return children(this) }), selector)
    },
    contents: function() {
      return this.map(function() { return slice.call(this.childNodes) })
    },
    siblings: function(selector){
      return filtered(this.map(function(i, el){
        return filter.call(children(el.parentNode), function(child){ return child!==el })
      }), selector)
    },
    empty: function(){
      return this.each(function(){ this.innerHTML = '' })
    },
    // `pluck` is borrowed from Prototype.js
    pluck: function(property){
      return $.map(this, function(el){ return el[property] })
    },
    show: function(){
      return this.each(function(){
        this.style.display == "none" && (this.style.display = '')
        if (getComputedStyle(this, '').getPropertyValue("display") == "none")
          this.style.display = defaultDisplay(this.nodeName)
      })
    },
    replaceWith: function(newContent){
      return this.before(newContent).remove()
    },
    wrap: function(structure){
      var func = isFunction(structure)
      if (this[0] && !func)
        var dom   = $(structure).get(0),
            clone = dom.parentNode || this.length > 1

      return this.each(function(index){
        $(this).wrapAll(
          func ? structure.call(this, index) :
            clone ? dom.cloneNode(true) : dom
        )
      })
    },
    wrapAll: function(structure){
      if (this[0]) {
        $(this[0]).before(structure = $(structure))
        var children
        // drill down to the inmost element
        while ((children = structure.children()).length) structure = children.first()
        $(structure).append(this)
      }
      return this
    },
    wrapInner: function(structure){
      var func = isFunction(structure)
      return this.each(function(index){
        var self = $(this), contents = self.contents(),
            dom  = func ? structure.call(this, index) : structure
        contents.length ? contents.wrapAll(dom) : self.append(dom)
      })
    },
    unwrap: function(){
      this.parent().each(function(){
        $(this).replaceWith($(this).children())
      })
      return this
    },
    clone: function(){
      return this.map(function(){ return this.cloneNode(true) })
    },
    hide: function(){
      return this.css("display", "none")
    },
    toggle: function(setting){
      return this.each(function(){
        var el = $(this)
        ;(setting === undefined ? el.css("display") == "none" : setting) ? el.show() : el.hide()
      })
    },
    prev: function(selector){ return $(this.pluck('previousElementSibling')).filter(selector || '*') },
    next: function(selector){ return $(this.pluck('nextElementSibling')).filter(selector || '*') },
    html: function(html){
      return arguments.length === 0 ?
        (this.length > 0 ? this[0].innerHTML : null) :
        this.each(function(idx){
          var originHtml = this.innerHTML
          $(this).empty().append( funcArg(this, html, idx, originHtml) )
        })
    },
    text: function(text){
      return arguments.length === 0 ?
        (this.length > 0 ? this[0].textContent : null) :
        this.each(function(){ this.textContent = (text === undefined) ? '' : ''+text })
    },
    attr: function(name, value){
      var result
      return (typeof name == 'string' && value === undefined) ?
        (this.length == 0 || this[0].nodeType !== 1 ? undefined :
          (name == 'value' && this[0].nodeName == 'INPUT') ? this.val() :
          (!(result = this[0].getAttribute(name)) && name in this[0]) ? this[0][name] : result
        ) :
        this.each(function(idx){
          if (this.nodeType !== 1) return
          if (isObject(name)) for (key in name) setAttribute(this, key, name[key])
          else setAttribute(this, name, funcArg(this, value, idx, this.getAttribute(name)))
        })
    },
    removeAttr: function(name){
      return this.each(function(){ this.nodeType === 1 && setAttribute(this, name) })
    },
    prop: function(name, value){
      name = propMap[name] || name
      return (value === undefined) ?
        (this[0] && this[0][name]) :
        this.each(function(idx){
          this[name] = funcArg(this, value, idx, this[name])
        })
    },
    data: function(name, value){
      var data = this.attr('data-' + name.replace(capitalRE, '-$1').toLowerCase(), value)
      return data !== null ? deserializeValue(data) : undefined
    },
    val: function(value){
      return arguments.length === 0 ?
        (this[0] && (this[0].multiple ?
           $(this[0]).find('option').filter(function(){ return this.selected }).pluck('value') :
           this[0].value)
        ) :
        this.each(function(idx){
          this.value = funcArg(this, value, idx, this.value)
        })
    },
    offset: function(coordinates){
      if (coordinates) return this.each(function(index){
        var $this = $(this),
            coords = funcArg(this, coordinates, index, $this.offset()),
            parentOffset = $this.offsetParent().offset(),
            props = {
              top:  coords.top  - parentOffset.top,
              left: coords.left - parentOffset.left
            }

        if ($this.css('position') == 'static') props['position'] = 'relative'
        $this.css(props)
      })
      if (this.length==0) return null
      var obj = this[0].getBoundingClientRect()
      return {
        left: obj.left + window.pageXOffset,
        top: obj.top + window.pageYOffset,
        width: Math.round(obj.width),
        height: Math.round(obj.height)
      }
    },
    css: function(property, value){
      if (arguments.length < 2) {
        var element = this[0], computedStyle = getComputedStyle(element, '')
        if(!element) return
        if (typeof property == 'string')
          return element.style[camelize(property)] || computedStyle.getPropertyValue(property)
        else if (isArray(property)) {
          var props = {}
          $.each(isArray(property) ? property: [property], function(_, prop){
            props[prop] = (element.style[camelize(prop)] || computedStyle.getPropertyValue(prop))
          })
          return props
        }
      }

      var css = ''
      if (type(property) == 'string') {
        if (!value && value !== 0)
          this.each(function(){ this.style.removeProperty(dasherize(property)) })
        else
          css = dasherize(property) + ":" + maybeAddPx(property, value)
      } else {
        for (key in property)
          if (!property[key] && property[key] !== 0)
            this.each(function(){ this.style.removeProperty(dasherize(key)) })
          else
            css += dasherize(key) + ':' + maybeAddPx(key, property[key]) + ';'
      }

      return this.each(function(){ this.style.cssText += ';' + css })
    },
    index: function(element){
      return element ? this.indexOf($(element)[0]) : this.parent().children().indexOf(this[0])
    },
    hasClass: function(name){
      if (!name) return false
      return emptyArray.some.call(this, function(el){
        return this.test(className(el))
      }, classRE(name))
    },
    addClass: function(name){
      if (!name) return this
      return this.each(function(idx){
        classList = []
        var cls = className(this), newName = funcArg(this, name, idx, cls)
        newName.split(/\s+/g).forEach(function(klass){
          if (!$(this).hasClass(klass)) classList.push(klass)
        }, this)
        classList.length && className(this, cls + (cls ? " " : "") + classList.join(" "))
      })
    },
    removeClass: function(name){
      return this.each(function(idx){
        if (name === undefined) return className(this, '')
        classList = className(this)
        funcArg(this, name, idx, classList).split(/\s+/g).forEach(function(klass){
          classList = classList.replace(classRE(klass), " ")
        })
        className(this, classList.trim())
      })
    },
    toggleClass: function(name, when){
      if (!name) return this
      return this.each(function(idx){
        var $this = $(this), names = funcArg(this, name, idx, className(this))
        names.split(/\s+/g).forEach(function(klass){
          (when === undefined ? !$this.hasClass(klass) : when) ?
            $this.addClass(klass) : $this.removeClass(klass)
        })
      })
    },
    scrollTop: function(value){
      if (!this.length) return
      var hasScrollTop = 'scrollTop' in this[0]
      if (value === undefined) return hasScrollTop ? this[0].scrollTop : this[0].pageYOffset
      return this.each(hasScrollTop ?
        function(){ this.scrollTop = value } :
        function(){ this.scrollTo(this.scrollX, value) })
    },
    scrollLeft: function(value){
      if (!this.length) return
      var hasScrollLeft = 'scrollLeft' in this[0]
      if (value === undefined) return hasScrollLeft ? this[0].scrollLeft : this[0].pageXOffset
      return this.each(hasScrollLeft ?
        function(){ this.scrollLeft = value } :
        function(){ this.scrollTo(value, this.scrollY) })
    },
    position: function() {
      if (!this.length) return

      var elem = this[0],
        // Get *real* offsetParent
        offsetParent = this.offsetParent(),
        // Get correct offsets
        offset       = this.offset(),
        parentOffset = rootNodeRE.test(offsetParent[0].nodeName) ? { top: 0, left: 0 } : offsetParent.offset()

      // Subtract element margins
      // note: when an element has margin: auto the offsetLeft and marginLeft
      // are the same in Safari causing offset.left to incorrectly be 0
      offset.top  -= parseFloat( $(elem).css('margin-top') ) || 0
      offset.left -= parseFloat( $(elem).css('margin-left') ) || 0

      // Add offsetParent borders
      parentOffset.top  += parseFloat( $(offsetParent[0]).css('border-top-width') ) || 0
      parentOffset.left += parseFloat( $(offsetParent[0]).css('border-left-width') ) || 0

      // Subtract the two offsets
      return {
        top:  offset.top  - parentOffset.top,
        left: offset.left - parentOffset.left
      }
    },
    offsetParent: function() {
      return this.map(function(){
        var parent = this.offsetParent || document.body
        while (parent && !rootNodeRE.test(parent.nodeName) && $(parent).css("position") == "static")
          parent = parent.offsetParent
        return parent
      })
    }
  }

  // for now
  $.fn.detach = $.fn.remove

  // Generate the `width` and `height` functions
  ;['width', 'height'].forEach(function(dimension){
    var dimensionProperty =
      dimension.replace(/./, function(m){ return m[0].toUpperCase() })

    $.fn[dimension] = function(value){
      var offset, el = this[0]
      if (value === undefined) return isWindow(el) ? el['inner' + dimensionProperty] :
        isDocument(el) ? el.documentElement['scroll' + dimensionProperty] :
        (offset = this.offset()) && offset[dimension]
      else return this.each(function(idx){
        el = $(this)
        el.css(dimension, funcArg(this, value, idx, el[dimension]()))
      })
    }
  })

  function traverseNode(node, fun) {
    fun(node)
    for (var key in node.childNodes) traverseNode(node.childNodes[key], fun)
  }

  // Generate the `after`, `prepend`, `before`, `append`,
  // `insertAfter`, `insertBefore`, `appendTo`, and `prependTo` methods.
  adjacencyOperators.forEach(function(operator, operatorIndex) {
    var inside = operatorIndex % 2 //=> prepend, append

    $.fn[operator] = function(){
      // arguments can be nodes, arrays of nodes, Zepto objects and HTML strings
      var argType, nodes = $.map(arguments, function(arg) {
            argType = type(arg)
            return argType == "object" || argType == "array" || arg == null ?
              arg : zepto.fragment(arg)
          }),
          parent, copyByClone = this.length > 1
      if (nodes.length < 1) return this

      return this.each(function(_, target){
        parent = inside ? target : target.parentNode

        // convert all methods to a "before" operation
        target = operatorIndex == 0 ? target.nextSibling :
                 operatorIndex == 1 ? target.firstChild :
                 operatorIndex == 2 ? target :
                 null

        nodes.forEach(function(node){
          if (copyByClone) node = node.cloneNode(true)
          else if (!parent) return $(node).remove()

          traverseNode(parent.insertBefore(node, target), function(el){
            if (el.nodeName != null && el.nodeName.toUpperCase() === 'SCRIPT' &&
               (!el.type || el.type === 'text/javascript') && !el.src)
              window['eval'].call(window, el.innerHTML)
          })
        })
      })
    }

    // after    => insertAfter
    // prepend  => prependTo
    // before   => insertBefore
    // append   => appendTo
    $.fn[inside ? operator+'To' : 'insert'+(operatorIndex ? 'Before' : 'After')] = function(html){
      $(html)[operator](this)
      return this
    }
  })

  zepto.Z.prototype = $.fn

  // Export internal API functions in the `$.zepto` namespace
  zepto.uniq = uniq
  zepto.deserializeValue = deserializeValue
  $.zepto = zepto

  return $
})()

window.Zepto = Zepto
window.$ === undefined && (window.$ = Zepto)

;(function($){
  var _zid = 1, undefined,
      slice = Array.prototype.slice,
      isFunction = $.isFunction,
      isString = function(obj){ return typeof obj == 'string' },
      handlers = {},
      specialEvents={},
      focusinSupported = 'onfocusin' in window,
      focus = { focus: 'focusin', blur: 'focusout' },
      hover = { mouseenter: 'mouseover', mouseleave: 'mouseout' }

  specialEvents.click = specialEvents.mousedown = specialEvents.mouseup = specialEvents.mousemove = 'MouseEvents'

  function zid(element) {
    return element._zid || (element._zid = _zid++)
  }
  function findHandlers(element, event, fn, selector) {
    event = parse(event)
    if (event.ns) var matcher = matcherFor(event.ns)
    return (handlers[zid(element)] || []).filter(function(handler) {
      return handler
        && (!event.e  || handler.e == event.e)
        && (!event.ns || matcher.test(handler.ns))
        && (!fn       || zid(handler.fn) === zid(fn))
        && (!selector || handler.sel == selector)
    })
  }
  function parse(event) {
    var parts = ('' + event).split('.')
    return {e: parts[0], ns: parts.slice(1).sort().join(' ')}
  }
  function matcherFor(ns) {
    return new RegExp('(?:^| )' + ns.replace(' ', ' .* ?') + '(?: |$)')
  }

  function eventCapture(handler, captureSetting) {
    return handler.del &&
      (!focusinSupported && (handler.e in focus)) ||
      !!captureSetting
  }

  function realEvent(type) {
    return hover[type] || (focusinSupported && focus[type]) || type
  }

  function add(element, events, fn, data, selector, delegator, capture){
    var id = zid(element), set = (handlers[id] || (handlers[id] = []))
    events.split(/\s/).forEach(function(event){
      if (event == 'ready') return $(document).ready(fn)
      var handler   = parse(event)
      handler.fn    = fn
      handler.sel   = selector
      // emulate mouseenter, mouseleave
      if (handler.e in hover) fn = function(e){
        var related = e.relatedTarget
        if (!related || (related !== this && !$.contains(this, related)))
          return handler.fn.apply(this, arguments)
      }
      handler.del   = delegator
      var callback  = delegator || fn
      handler.proxy = function(e){
        e = compatible(e)
        if (e.isImmediatePropagationStopped()) return
        e.data = data
        var result = callback.apply(element, e._args == undefined ? [e] : [e].concat(e._args))
        if (result === false) e.preventDefault(), e.stopPropagation()
        return result
      }
      handler.i = set.length
      set.push(handler)
      if ('addEventListener' in element)
        element.addEventListener(realEvent(handler.e), handler.proxy, eventCapture(handler, capture))
    })
  }
  function remove(element, events, fn, selector, capture){
    var id = zid(element)
    ;(events || '').split(/\s/).forEach(function(event){
      findHandlers(element, event, fn, selector).forEach(function(handler){
        delete handlers[id][handler.i]
      if ('removeEventListener' in element)
        element.removeEventListener(realEvent(handler.e), handler.proxy, eventCapture(handler, capture))
      })
    })
  }

  $.event = { add: add, remove: remove }

  $.proxy = function(fn, context) {
    if (isFunction(fn)) {
      var proxyFn = function(){ return fn.apply(context, arguments) }
      proxyFn._zid = zid(fn)
      return proxyFn
    } else if (isString(context)) {
      return $.proxy(fn[context], fn)
    } else {
      throw new TypeError("expected function")
    }
  }

  $.fn.bind = function(event, data, callback){
    return this.on(event, data, callback)
  }
  $.fn.unbind = function(event, callback){
    return this.off(event, callback)
  }
  $.fn.one = function(event, selector, data, callback){
    return this.on(event, selector, data, callback, 1)
  }

  var returnTrue = function(){return true},
      returnFalse = function(){return false},
      ignoreProperties = /^([A-Z]|returnValue$|layer[XY]$)/,
      eventMethods = {
        preventDefault: 'isDefaultPrevented',
        stopImmediatePropagation: 'isImmediatePropagationStopped',
        stopPropagation: 'isPropagationStopped'
      }

  function compatible(event, source) {
    if (source || !event.isDefaultPrevented) {
      source || (source = event)

      $.each(eventMethods, function(name, predicate) {
        var sourceMethod = source[name]
        event[name] = function(){
          this[predicate] = returnTrue
          return sourceMethod && sourceMethod.apply(source, arguments)
        }
        event[predicate] = returnFalse
      })

      if (source.defaultPrevented !== undefined ? source.defaultPrevented :
          'returnValue' in source ? source.returnValue === false :
          source.getPreventDefault && source.getPreventDefault())
        event.isDefaultPrevented = returnTrue
    }
    return event
  }

  function createProxy(event) {
    var key, proxy = { originalEvent: event }
    for (key in event)
      if (!ignoreProperties.test(key) && event[key] !== undefined) proxy[key] = event[key]

    return compatible(proxy, event)
  }

  $.fn.delegate = function(selector, event, callback){
    return this.on(event, selector, callback)
  }
  $.fn.undelegate = function(selector, event, callback){
    return this.off(event, selector, callback)
  }

  $.fn.live = function(event, callback){
    $(document.body).delegate(this.selector, event, callback)
    return this
  }
  $.fn.die = function(event, callback){
    $(document.body).undelegate(this.selector, event, callback)
    return this
  }

  $.fn.on = function(event, selector, data, callback, one){
    var autoRemove, delegator, $this = this
    if (event && !isString(event)) {
      $.each(event, function(type, fn){
        $this.on(type, selector, data, fn, one)
      })
      return $this
    }

    if (!isString(selector) && !isFunction(callback) && callback !== false)
      callback = data, data = selector, selector = undefined
    if (isFunction(data) || data === false)
      callback = data, data = undefined

    if (callback === false) callback = returnFalse

    return $this.each(function(_, element){
      if (one) autoRemove = function(e){
        remove(element, e.type, callback)
        return callback.apply(this, arguments)
      }

      if (selector) delegator = function(e){
        var evt, match = $(e.target).closest(selector, element).get(0)
        if (match && match !== element) {
          evt = $.extend(createProxy(e), {currentTarget: match, liveFired: element})
          return (autoRemove || callback).apply(match, [evt].concat(slice.call(arguments, 1)))
        }
      }

      add(element, event, callback, data, selector, delegator || autoRemove)
    })
  }
  $.fn.off = function(event, selector, callback){
    var $this = this
    if (event && !isString(event)) {
      $.each(event, function(type, fn){
        $this.off(type, selector, fn)
      })
      return $this
    }

    if (!isString(selector) && !isFunction(callback) && callback !== false)
      callback = selector, selector = undefined

    if (callback === false) callback = returnFalse

    return $this.each(function(){
      remove(this, event, callback, selector)
    })
  }

  $.fn.trigger = function(event, args){
    event = (isString(event) || $.isPlainObject(event)) ? $.Event(event) : compatible(event)
    event._args = args
    return this.each(function(){
      // items in the collection might not be DOM elements
      if('dispatchEvent' in this) this.dispatchEvent(event)
      else $(this).triggerHandler(event, args)
    })
  }

  // triggers event handlers on current element just as if an event occurred,
  // doesn't trigger an actual event, doesn't bubble
  $.fn.triggerHandler = function(event, args){
    var e, result
    this.each(function(i, element){
      e = createProxy(isString(event) ? $.Event(event) : event)
      e._args = args
      e.target = element
      $.each(findHandlers(element, event.type || event), function(i, handler){
        result = handler.proxy(e)
        if (e.isImmediatePropagationStopped()) return false
      })
    })
    return result
  }

  // shortcut methods for `.bind(event, fn)` for each event type
  ;('focusin focusout load resize scroll unload click dblclick '+
  'mousedown mouseup mousemove mouseover mouseout mouseenter mouseleave '+
  'change select keydown keypress keyup error').split(' ').forEach(function(event) {
    $.fn[event] = function(callback) {
      return callback ?
        this.bind(event, callback) :
        this.trigger(event)
    }
  })

  ;['focus', 'blur'].forEach(function(name) {
    $.fn[name] = function(callback) {
      if (callback) this.bind(name, callback)
      else this.each(function(){
        try { this[name]() }
        catch(e) {}
      })
      return this
    }
  })

  $.Event = function(type, props) {
    if (!isString(type)) props = type, type = props.type
    var event = document.createEvent(specialEvents[type] || 'Events'), bubbles = true
    if (props) for (var name in props) (name == 'bubbles') ? (bubbles = !!props[name]) : (event[name] = props[name])
    event.initEvent(type, bubbles, true)
    return compatible(event)
  }

})(Zepto)

;(function($){
  var jsonpID = 0,
      document = window.document,
      key,
      name,
      rscript = /<script\b[^<]*(?:(?!<\/script>)<[^<]*)*<\/script>/gi,
      scriptTypeRE = /^(?:text|application)\/javascript/i,
      xmlTypeRE = /^(?:text|application)\/xml/i,
      jsonType = 'application/json',
      htmlType = 'text/html',
      blankRE = /^\s*$/

  // trigger a custom event and return false if it was cancelled
  function triggerAndReturn(context, eventName, data) {
    var event = $.Event(eventName)
    $(context).trigger(event, data)
    return !event.isDefaultPrevented()
  }

  // trigger an Ajax "global" event
  function triggerGlobal(settings, context, eventName, data) {
    if (settings.global) return triggerAndReturn(context || document, eventName, data)
  }

  // Number of active Ajax requests
  $.active = 0

  function ajaxStart(settings) {
    if (settings.global && $.active++ === 0) triggerGlobal(settings, null, 'ajaxStart')
  }
  function ajaxStop(settings) {
    if (settings.global && !(--$.active)) triggerGlobal(settings, null, 'ajaxStop')
  }

  // triggers an extra global event "ajaxBeforeSend" that's like "ajaxSend" but cancelable
  function ajaxBeforeSend(xhr, settings) {
    var context = settings.context
    if (settings.beforeSend.call(context, xhr, settings) === false ||
        triggerGlobal(settings, context, 'ajaxBeforeSend', [xhr, settings]) === false)
      return false

    triggerGlobal(settings, context, 'ajaxSend', [xhr, settings])
  }
  function ajaxSuccess(data, xhr, settings, deferred) {
    var context = settings.context, status = 'success'
    settings.success.call(context, data, status, xhr)
    if (deferred) deferred.resolveWith(context, [data, status, xhr])
    triggerGlobal(settings, context, 'ajaxSuccess', [xhr, settings, data])
    ajaxComplete(status, xhr, settings)
  }
  // type: "timeout", "error", "abort", "parsererror"
  function ajaxError(error, type, xhr, settings, deferred) {
    var context = settings.context
    settings.error.call(context, xhr, type, error)
    if (deferred) deferred.rejectWith(context, [xhr, type, error])
    triggerGlobal(settings, context, 'ajaxError', [xhr, settings, error || type])
    ajaxComplete(type, xhr, settings)
  }
  // status: "success", "notmodified", "error", "timeout", "abort", "parsererror"
  function ajaxComplete(status, xhr, settings) {
    var context = settings.context
    settings.complete.call(context, xhr, status)
    triggerGlobal(settings, context, 'ajaxComplete', [xhr, settings])
    ajaxStop(settings)
  }

  // Empty function, used as default callback
  function empty() {}

  $.ajaxJSONP = function(options, deferred){
    if (!('type' in options)) return $.ajax(options)

    var _callbackName = options.jsonpCallback,
      callbackName = ($.isFunction(_callbackName) ?
        _callbackName() : _callbackName) || ('jsonp' + (++jsonpID)),
      script = document.createElement('script'),
      originalCallback = window[callbackName],
      responseData,
      abort = function(errorType) {
        $(script).triggerHandler('error', errorType || 'abort')
      },
      xhr = { abort: abort }, abortTimeout

    if (deferred) deferred.promise(xhr)

    $(script).on('load error', function(e, errorType){
      clearTimeout(abortTimeout)
      $(script).off().remove()

      if (e.type == 'error' || !responseData) {
        ajaxError(null, errorType || 'error', xhr, options, deferred)
      } else {
        ajaxSuccess(responseData[0], xhr, options, deferred)
      }

      window[callbackName] = originalCallback
      if (responseData && $.isFunction(originalCallback))
        originalCallback(responseData[0])

      originalCallback = responseData = undefined
    })

    if (ajaxBeforeSend(xhr, options) === false) {
      abort('abort')
      return xhr
    }

    window[callbackName] = function(){
      responseData = arguments
    }

    script.src = options.url.replace(/\?(.+)=\?/, '?$1=' + callbackName)
    document.head.appendChild(script)

    if (options.timeout > 0) abortTimeout = setTimeout(function(){
      abort('timeout')
    }, options.timeout)

    return xhr
  }

  $.ajaxSettings = {
    // Default type of request
    type: 'GET',
    // Callback that is executed before request
    beforeSend: empty,
    // Callback that is executed if the request succeeds
    success: empty,
    // Callback that is executed the the server drops error
    error: empty,
    // Callback that is executed on request complete (both: error and success)
    complete: empty,
    // The context for the callbacks
    context: null,
    // Whether to trigger "global" Ajax events
    global: true,
    // Transport
    xhr: function () {
      return new window.XMLHttpRequest()
    },
    // MIME types mapping
    // IIS returns Javascript as "application/x-javascript"
    accepts: {
      script: 'text/javascript, application/javascript, application/x-javascript',
      json:   jsonType,
      xml:    'application/xml, text/xml',
      html:   htmlType,
      text:   'text/plain'
    },
    // Whether the request is to another domain
    crossDomain: false,
    // Default timeout
    timeout: 0,
    // Whether data should be serialized to string
    processData: true,
    // Whether the browser should be allowed to cache GET responses
    cache: true
  }

  function mimeToDataType(mime) {
    if (mime) mime = mime.split(';', 2)[0]
    return mime && ( mime == htmlType ? 'html' :
      mime == jsonType ? 'json' :
      scriptTypeRE.test(mime) ? 'script' :
      xmlTypeRE.test(mime) && 'xml' ) || 'text'
  }

  function appendQuery(url, query) {
    if (query == '') return url
    return (url + '&' + query).replace(/[&?]{1,2}/, '?')
  }

  // serialize payload and append it to the URL for GET requests
  function serializeData(options) {
    if (options.processData && options.data && $.type(options.data) != "string")
      options.data = $.param(options.data, options.traditional)
    if (options.data && (!options.type || options.type.toUpperCase() == 'GET'))
      options.url = appendQuery(options.url, options.data), options.data = undefined
  }

  $.ajax = function(options){
    var settings = $.extend({}, options || {}),
        deferred = $.Deferred && $.Deferred()
    for (key in $.ajaxSettings) if (settings[key] === undefined) settings[key] = $.ajaxSettings[key]

    ajaxStart(settings)

    if (!settings.crossDomain) settings.crossDomain = /^([\w-]+:)?\/\/([^\/]+)/.test(settings.url) &&
      RegExp.$2 != window.location.host

    if (!settings.url) settings.url = window.location.toString()
    serializeData(settings)
    if (settings.cache === false) settings.url = appendQuery(settings.url, '_=' + Date.now())

    var dataType = settings.dataType, hasPlaceholder = /\?.+=\?/.test(settings.url)
    if (dataType == 'jsonp' || hasPlaceholder) {
      if (!hasPlaceholder)
        settings.url = appendQuery(settings.url,
          settings.jsonp ? (settings.jsonp + '=?') : settings.jsonp === false ? '' : 'callback=?')
      return $.ajaxJSONP(settings, deferred)
    }

    var mime = settings.accepts[dataType],
        headers = { },
        setHeader = function(name, value) { headers[name.toLowerCase()] = [name, value] },
        protocol = /^([\w-]+:)\/\//.test(settings.url) ? RegExp.$1 : window.location.protocol,
        xhr = settings.xhr(),
        nativeSetHeader = xhr.setRequestHeader,
        abortTimeout

    if (deferred) deferred.promise(xhr)

    if (!settings.crossDomain) setHeader('X-Requested-With', 'XMLHttpRequest')
    setHeader('Accept', mime || '*/*')
    if (mime = settings.mimeType || mime) {
      if (mime.indexOf(',') > -1) mime = mime.split(',', 2)[0]
      xhr.overrideMimeType && xhr.overrideMimeType(mime)
    }
    if (settings.contentType || (settings.contentType !== false && settings.data && settings.type.toUpperCase() != 'GET'))
      setHeader('Content-Type', settings.contentType || 'application/x-www-form-urlencoded')

    if (settings.headers) for (name in settings.headers) setHeader(name, settings.headers[name])
    xhr.setRequestHeader = setHeader

    xhr.onreadystatechange = function(){
      if (xhr.readyState == 4) {
        xhr.onreadystatechange = empty
        clearTimeout(abortTimeout)
        var result, error = false
        if ((xhr.status >= 200 && xhr.status < 300) || xhr.status == 304 || (xhr.status == 0 && protocol == 'file:')) {
          dataType = dataType || mimeToDataType(settings.mimeType || xhr.getResponseHeader('content-type'))
          result = xhr.responseText

          try {
            // http://perfectionkills.com/global-eval-what-are-the-options/
            if (dataType == 'script')    (1,eval)(result)
            else if (dataType == 'xml')  result = xhr.responseXML
            else if (dataType == 'json') result = blankRE.test(result) ? null : $.parseJSON(result)
          } catch (e) { error = e }

          if (error) ajaxError(error, 'parsererror', xhr, settings, deferred)
          else ajaxSuccess(result, xhr, settings, deferred)
        } else {
          ajaxError(xhr.statusText || null, xhr.status ? 'error' : 'abort', xhr, settings, deferred)
        }
      }
    }

    if (ajaxBeforeSend(xhr, settings) === false) {
      xhr.abort()
      ajaxError(null, 'abort', xhr, settings, deferred)
      return xhr
    }

    if (settings.xhrFields) for (name in settings.xhrFields) xhr[name] = settings.xhrFields[name]

    var async = 'async' in settings ? settings.async : true
    xhr.open(settings.type, settings.url, async, settings.username, settings.password)

    for (name in headers) nativeSetHeader.apply(xhr, headers[name])

    if (settings.timeout > 0) abortTimeout = setTimeout(function(){
        xhr.onreadystatechange = empty
        xhr.abort()
        ajaxError(null, 'timeout', xhr, settings, deferred)
      }, settings.timeout)

    // avoid sending empty string (#319)
    xhr.send(settings.data ? settings.data : null)
    return xhr
  }

  // handle optional data/success arguments
  function parseArguments(url, data, success, dataType) {
    if ($.isFunction(data)) dataType = success, success = data, data = undefined
    if (!$.isFunction(success)) dataType = success, success = undefined
    return {
      url: url
    , data: data
    , success: success
    , dataType: dataType
    }
  }

  $.get = function(/* url, data, success, dataType */){
    return $.ajax(parseArguments.apply(null, arguments))
  }

  $.post = function(/* url, data, success, dataType */){
    var options = parseArguments.apply(null, arguments)
    options.type = 'POST'
    return $.ajax(options)
  }

  $.getJSON = function(/* url, data, success */){
    var options = parseArguments.apply(null, arguments)
    options.dataType = 'json'
    return $.ajax(options)
  }

  $.fn.load = function(url, data, success){
    if (!this.length) return this
    var self = this, parts = url.split(/\s/), selector,
        options = parseArguments(url, data, success),
        callback = options.success
    if (parts.length > 1) options.url = parts[0], selector = parts[1]
    options.success = function(response){
      self.html(selector ?
        $('<div>').html(response.replace(rscript, "")).find(selector)
        : response)
      callback && callback.apply(self, arguments)
    }
    $.ajax(options)
    return this
  }

  var escape = encodeURIComponent

  function serialize(params, obj, traditional, scope){
    var type, array = $.isArray(obj), hash = $.isPlainObject(obj)
    $.each(obj, function(key, value) {
      type = $.type(value)
      if (scope) key = traditional ? scope :
        scope + '[' + (hash || type == 'object' || type == 'array' ? key : '') + ']'
      // handle data in serializeArray() format
      if (!scope && array) params.add(value.name, value.value)
      // recurse into nested objects
      else if (type == "array" || (!traditional && type == "object"))
        serialize(params, value, traditional, key)
      else params.add(key, value)
    })
  }

  $.param = function(obj, traditional){
    var params = []
    params.add = function(k, v){ this.push(escape(k) + '=' + escape(v)) }
    serialize(params, obj, traditional)
    return params.join('&').replace(/%20/g, '+')
  }
})(Zepto)

;(function($){
  $.fn.serializeArray = function() {
    var result = [], el
    $([].slice.call(this.get(0).elements)).each(function(){
      el = $(this)
      var type = el.attr('type')
      if (this.nodeName.toLowerCase() != 'fieldset' &&
        !this.disabled && type != 'submit' && type != 'reset' && type != 'button' &&
        ((type != 'radio' && type != 'checkbox') || this.checked))
        result.push({
          name: el.attr('name'),
          value: el.val()
        })
    })
    return result
  }

  $.fn.serialize = function(){
    var result = []
    this.serializeArray().forEach(function(elm){
      result.push(encodeURIComponent(elm.name) + '=' + encodeURIComponent(elm.value))
    })
    return result.join('&')
  }

  $.fn.submit = function(callback) {
    if (callback) this.bind('submit', callback)
    else if (this.length) {
      var event = $.Event('submit')
      this.eq(0).trigger(event)
      if (!event.isDefaultPrevented()) this.get(0).submit()
    }
    return this
  }

})(Zepto)

;(function($){
  // __proto__ doesn't exist on IE<11, so redefine
  // the Z function to use object extension instead
  if (!('__proto__' in {})) {
    $.extend($.zepto, {
      Z: function(dom, selector){
        dom = dom || []
        $.extend(dom, $.fn)
        dom.selector = selector || ''
        dom.__Z = true
        return dom
      },
      // this is a kludge but works
      isZ: function(object){
        return $.type(object) === 'array' && '__Z' in object
      }
    })
  }

  // getComputedStyle shouldn't freak out when called
  // without a valid element as argument
  try {
    getComputedStyle(undefined)
  } catch(e) {
    var nativeGetComputedStyle = getComputedStyle;
    window.getComputedStyle = function(element){
      try {
        return nativeGetComputedStyle(element)
      } catch(e) {
        return null
      }
    }
  }
})(Zepto)
;
define( 'jquery', (function (global) {
    return function () {
        var ret, fn;
        return ret || global.window.$;
    };
}(this)));

define( 'state/content',[
    "../config/rte/main",
    "jquery"
], function( rteConfig, $ ) {
    

    var exports = {
        attachedImageCount: 0
    };

    exports.setContent = function( content ) {
        $( rteConfig.selector.RTE ).html(content);
    };

    exports.getContent = function() {
        return $( rteConfig.selector.RTE ).html();
    };

    exports.getContentLength = function() {
        return $( rteConfig.selector.RTE ).text().trim().length;
    };

    exports.getEmbeddedImageCount = function() {
        return $( rteConfig.selector.RTE ).find( rteConfig.selector.IMAGE ).length + this.attachedImageCount;
    };

    return exports;
});

define( 'state/rte',[],function() {
    

    var exports = {};

    // # UI State Commands

    exports.formatBlock = "";
    exports.justifyLeft = false;
    exports.justifyCenter = false;
    exports.justifyRight = false;
    exports.justifyFull = false;
    exports.superscript = false;
    exports.subscript = false;
    exports.bold = false;
    exports.italic = false;
    exports.underline = false;

    exports.insertUnorderedList = false;
    exports.insertOrderedList = false;

    // # Ranges

    exports.insertionRange = null;

    /**
     * A Serialized JSON String for the current caret point.
     */
    exports.scrollRectangle = null;

    // # Helper Methods

    exports.getValue = function( command ) {
        if ( this.hasOwnProperty( command ) ) {
            return this[ command ];
        }

        return null;
    };

    return exports;
});

define( 'native/validation',[
    "../config/native/main",
    "../config/rte/main"
], function( nativeConfig, appConfig ) {
    

    var exports = {};

    exports.isIos = function() {
        return nativeConfig.PLATFORM === appConfig.platformType.IOS;
    };

    exports.isAndroid = function() {
        return nativeConfig.PLATFORM === appConfig.platformType.ANDROID;
    };

    exports.hasContextAppUrl = function( context ) {
        return exports.isIos() && !!context.iosUrl;
    };

    return exports;
});

define( 'dom/validation/main',[
    "../../config/rte/main",
    "../../native/factory",
    "../../native/validation",
    "jquery"
], function( rteConfig, nativeFactory, validation, $ ) {
    

    var exports = {};

    /**
     * Don't use this method frequently.
     * Don't use this method in a loop.
     * Don't use this method for large content.
     */
    exports.isElementEmpty = function( $el ) {

        // TODO: move to config.
        var possibleEmpties = { "": true, "<br>": true, "<p><br></p>": true },
            $children = $el.children(),
            html = $el.html();

        return ( $children.length <= 1 && html.length < rteConfig.constants.EMPTY_HTML_LENGTH && possibleEmpties[ html ] !== undefined );
    };

    exports.isContentEmpty = function( $node ) {
        var $content = $node || $( rteConfig.selector.RTE );

        return exports.isElementEmpty( $content );
    };

    exports.isAncestor = function( ancestor, descendant, selfIsAncestor ) {
        var n = selfIsAncestor ? descendant : descendant.parentNode;

        while ( n ) {
            if ( n === ancestor ) {
                return true;
            } else {
                n = n.parentNode;
            }
        }

        return false;
    };

    exports.isTextNode = function( node ) {
        return node && node.nodeType && node.nodeType === rteConfig.nodeType.TEXT_NODE;
    };

    exports.isEmptyTextNode = function( node ) {
        return exports.isTextNode( node ) && node.nodeValue === "";
    };

    exports.isInsideListItem = function( node ) {
        if ( !node ) {
            return false;
        }

        var parent = node.parentNode;

        while ( parent ) {
            if ( parent.nodeName === "LI" ) {
                return true;
            }

            parent = parent.parentNode;
        }

        return false;
    };

    exports.hasParentOfType = function( nodeName, pivot ) {
        if ( !nodeName ) {
            return false;
        }

        var parent = pivot.parentNode;

        while (parent) {
            if (parent.nodeName === nodeName) {
                return true;
            }

            parent = parent.parentNode;
        }

        return false;
    };

    exports.isAtTheEndOfParent = function( pivot ) {
        var next = pivot.nextSibling;

        if ( !next ) {
            return true;
        }

        while ( next ) {
            if ( next.nodeType !== 3 || next.nodeValue ) {
                return false;
            }

            next = next.nextSibling;
        }

        return true;
    };

    return exports;
});

define('legacy/ui/range/state/main',[],function() {
    

    var exports = {};

    exports.selectionStartRange = null;
    exports.selectionEndRange = null;
    exports.lastTouchPoint = null;
    exports.serializedRange = "";
    exports.contentWidth = 0;
    exports.rangeText = "";
    exports.handleBounds = null;
    exports.menuBounds = null;

    return exports;
});

define( 'proxy/io/main',[
    "../../config/rte/main",
    "../../native/factory",
    "../../state/content",
    "../../state/rte",
    "../../dom/validation/main",
    "../../native/validation",
    "../../legacy/ui/range/state/main",
    "../../util/log",
    "../../util/timer"
], function( rteConfig, nativeFactory, contentState, rteState, validation, nativeValidation, legacyRangeState, log, timer ) {
    

    /**
     * Proxy from the Rte to Native.
     *
     * All `native.notify()` calls should be made in a method defined here.
     *
     * `native.notify` act similar to  `window.send()`
     *
     * This module **should not** deal with DOM,
     * it should delegate DOM manipulation tasks to other modules.
     *
     * If you feel like you need to import "jquery" here, then you are probably
     * adding your logic to the wrong module.
     *
     * This module mostly calls `native.notify` methods to route the flow to
     * the correct delegate. It should remain lean and simple.
     */

    var exports = {},

        jiveNative = nativeFactory.createInstance(),

        lastContentLength = 0,

        lastEmbeddedImageCount = 0,

        oldScrollRectangle = "",

        kFormatBlock = rteConfig.topic.rte.FORMAT_BLOCK;

    function getItem( key, command, value ) {
        var item;

        // Fix browser bug for TABDEV-281 when queryCommandState is not accurate.
        if ( key === kFormatBlock ) {
            item = ( key === command ) ? value : document.queryCommandValue( key );
        } else {
            item = ( key === command ) ? value : document.queryCommandState( key );
        }

        return item;
    }

    function updateChangesCollection( key, command, value, changes ) {
        var item = getItem( key, command, value ),
            anyChanges = false;

        if ( item !== rteState[ key ] && typeof rteState[ key ] !== "function" && key !== "scrollRectangle" ) {
            changes[ key ] = rteState[ key ] = item;

            anyChanges = true;
        }

        return anyChanges;
    }

    function updateScrollRectangle( key, command, value, changes ) {
        var anyChanges = false;

        // Not using these right now; they exist just to retain signature parity with `updateChangesCollection`.
        void key;
        void command;
        void value;

        if ( oldScrollRectangle !== rteState.scrollRectangle ) {
            anyChanges = true;
        }

        oldScrollRectangle = rteState.scrollRectangle;

        try {
            changes.scrollRectangle = JSON.parse( rteState.scrollRectangle );
        } catch (ignore) {}

        // Do not send changes if the scrollRectangle is falsy.
        if ( !changes.scrollRectangle ) {
            anyChanges = false;
        }

        return anyChanges;
    }

    function sendScrollRectanglePositionUpdate( command, value ) {

        // FIX for #ANDROID-478: Send scroll rectangle state change events ASAP.
        // Only android needs scroll rectangle updates at every state change.
        if ( !nativeValidation.isAndroid() ) {
            return;
        }

        var changes = {},
            anyChanges = false,
            key = rteConfig.state.SCROLL_RECTANGLE;

        anyChanges = updateScrollRectangle( key, command, value, changes ) || anyChanges;

        if ( anyChanges ) {
            jiveNative.notify({ type: rteConfig.topic.native.STATE_CHANGE, state: changes });
        }
    }

    var updateTimerId,
        rectanglePositionTimerId;

    function sendRectanglePositionUpdate( command, value ) {
        clearTimeout(rectanglePositionTimerId);

        rectanglePositionTimerId = timer.setTimeout(function() {
            sendScrollRectanglePositionUpdate( command, value );

            var stateHasChanged = false,
                contentLength = contentState.getContentLength(),
                embeddedImageCount = contentState.getEmbeddedImageCount();

            if ( contentLength !== lastContentLength ) {
                stateHasChanged = true;
                lastContentLength = contentLength;
            }

            if (embeddedImageCount !== lastEmbeddedImageCount) {
                stateHasChanged = true;
                lastEmbeddedImageCount = embeddedImageCount;
            }

            if ( stateHasChanged === true ) {
                jiveNative.notify({ type: rteConfig.topic.native.CONTENT_CHANGE, textLength: contentLength, imageCount: embeddedImageCount });

                if ( contentLength === 0 ) {
                    jiveNative.notify({ type: rteConfig.topic.native.REQUEST_FOCUS });
                }
            }
        });
    }

    function handleUpdateStateChange(command, value) {
        /*--(log.log( "/proxy/io/updateStateChanged()#setTimeout" );)--*/

        var changes = {},
            anyChanges = false,
            key;

        for ( key in rteState ) {
            if ( rteState.hasOwnProperty( key ) ) {
                anyChanges = updateChangesCollection( key, command, value, changes ) || anyChanges;
            }
        }

        if ( anyChanges ) {

            // In certain devices|events, the formatBlock data was coming in the form {"0":"h2"}, hence the regExp check.
            if ( changes && changes.formatBlock && (/^(p|h\d|b|i|u)$/i).test(changes.formatBlock[0]) ) {
                changes.formatBlock = changes.formatBlock[0];
            }

            jiveNative.notify({ type: rteConfig.topic.native.STATE_CHANGE, state: changes });
        }
    }

    function sendStateUpdate( command, value ) {
        clearTimeout( updateTimerId );

        // wrt #ANDROID-2059 we are using window.setTimeout here, instead of
        // timer.setTimeout, because the `handleUpdateStateChange` callback does
        // not touch the DOM, and therefore does not need to be processed
        // in an animation frame.
        updateTimerId = window.setTimeout(
            handleUpdateStateChange.bind(null, command, value),
            rteConfig.constants.MIN_STATE_NOTIFY_WAIT_INTERVAL
        );
    }

    exports.updateStateChange = function( command, value ) {
        /*--(log.log( "/proxy/io/updateStateChanged()" );)--*/

        sendRectanglePositionUpdate( command, value );
        sendStateUpdate( command, value );
    };

    exports.sendContent = function( request ) {
        /*--(log.log( "/proxy/io/sendContent(): sending..." );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.RECEIVE_CONTENT, request: request, content: contentState.getContent() });

        /*--(log.log( "/proxy/io/sentContent(): sent.");)--*/
    };

    exports.updateImageSelect = function() {
        /*--(log.log( "/proxy/io/updateImageSelect()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.IMAGE_SELECT, state: { success: true } });
    };

    exports.updateMention = function( meta ) {
        /*--(log.log( "/proxy/io/updateMention()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.MENTION_START, state: { success: true }, payload: meta });
    };

    exports.updateMentionFinish = function() {
        /*--(log.log( "/proxy/io/updateMentionFinish()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.MENTION_FINISH, state: { success: true } });
    };

    exports.updateImageDeselect = function() {
        /*--(log.log( "/proxy/io/updateImageDeselect()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.IMAGE_DESELECT, state: { success: true } });
    };

    exports.updateContentEmpty = function() {
        /*--(log.log( "/proxy/io/updateContentEmpty()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.ENSURE_CONTENT_EMPTY, state: { success: validation.isContentEmpty() } });
    };

    exports.sendRteReadyStatusUpdate = function() {
        /*--(log.log( "/proxy/io/sendRteReadyStatusUpdate()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.READY_STATUS_UPDATE, state: { success: true } });
    };

    exports.updateInsertNewLine = function() {
        /*--(log.log( "/proxy/io/updateInsertNewLine()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.INSERT_NEW_LINE, state: { success: true } });
    };

    /* #region legacy selection */

    exports.updateLegacySelection = function() {
        /*--(log.log( "/proxy/io/updateLegacySelection()" );)--*/

        jiveNative.notify({
            type: rteConfig.topic.native.UPDATE_LEGACY_SELECTION,
            contentWidth: legacyRangeState.contentWidth,
            serializedRange: legacyRangeState.serializedRange,
            rangeText: legacyRangeState.rangeText,
            handleBounds: JSON.stringify( legacyRangeState.handleBounds ),
            menuBounds: JSON.stringify( legacyRangeState.menuBounds )
        });
    };

    exports.updateClipboard = function( data ) {
        /*--(log.log( "/proxy/io/updateClipboard()" );)--*/

        jiveNative.notify({ type: rteConfig.topic.native.UPDATE_LEGACY_CLIPBOARD, data: data });
    };

    /* #endregion */

    return exports;
});

define( 'state/sizer',[],function() {
    

    var exports = {
        directions: null,
        left: false,
        right: false,
        vertical: false,
        top: 0,
        bottom: 0,
        horizontal: false,
        rect: null,
        newRect: null,
        point: null,
        ratio: 0
    };

    return exports;
});

define( 'ui/browser',[
    "../util/timer"
], function( timer ) {
    

    var exports = {};

    function openUrl( url ) {
        window.location = url;
    }

    exports.launchInstallUrl = function( context ) {
        if ( context.iosInstallUrl ) {
            timer.setTimeout( function() {
                openUrl( context.iosInstallUrl );
            });
        }

        openUrl( context.iosUrl );
    };

    return exports;
});

define( 'proxy/bridge/main',[
    "../../config/rte/main",
    "../../io/bridge"
], function( rteConfig, bridge ) {
    

    /**
     * This is a Facade around `proxy/bridge`.
     */

    var exports = {};

    exports.setDefaultParagraphSeparator = function() {

        // Make "paragraph"s the default, this is what the web RTE uses.
        bridge.enqueueCommand({
            command: rteConfig.topic.rte.DEFAULT_PARAGRAPH_SEPARATOR,
            args: [ false, rteConfig.selector.PARAGRAPH ]
        });
    };

    exports.insertHtml = function( html ) {
        bridge.enqueueCommand({
            command: rteConfig.topic.INSERT_HTML,
            args: [ false, html ]
        });
    };

    exports.startMention = function( position ) {
        bridge.enqueueCommand({
            command: rteConfig.topic.rte.MENTION,
            args: [
                { started: true, position: position, text: "" }
            ]
        });
    };

    return exports;
});

define( 'dom/style/main',[
    "jquery"
], function( $ ) {
    

    var exports = {};

    /**
     * Abstracting style operations, since they behave differently on iOS and Android.
     */

    /**
     * Zepto.js works differently in Android(pre kitkat) versus iOS.
     * This method is a Facade that takes care of those differences.
     */
    exports.hasBackgroundStyle = function( $node ) {
        var style = $( $node ).attr( "style" );

        if ( !style ) {
            return false;
        }

        if ( window.CSSStyleDeclaration && typeof style === window.CSSStyleDeclaration ) {
            return style.cssText.cssText.indexOf( "background" ) > -1;
        }

        if ( typeof style === "string" ) {
            return style.indexOf( "background" ) > -1;
        }

        return false;
    };

    return exports;
});

define( 'dom/sanitization/main',[
    "../../config/rte/main",
    "../style/main",
    "../../util/log",
    "jquery"
], function( rteConfig, style, log, $ ) {
    

    var exports = {},

        kRenderedContent = rteConfig.className.RENDERED_CONTENT,

        kContent = rteConfig.selector.RTE,
        kFakeImageSrc = rteConfig.selector.FAKE_IMAGE_SRC;

    exports.removeStyles = function( $node ) {
        /*--(log.log( "dom/sanitization/removeStyles()" );)--*/

        var $container = $node || $( kContent );

        $container.find( "[style]" ).each( function() {
            var $this = $( this );

            // remove style, unless it is a background image for a link
            if ( style.hasBackgroundStyle( $this ) ) {
                $this.removeAttr( "style" );
            }
        });
    };

    /**
     * Removed `jive-rendered-content` classes from the document.
     */
    exports.removeRenderedContentClasses = function( $node ) {
        var $container = $node || $( kRenderedContent );

        $container.children().removeClass( kRenderedContent );
    };

    exports.removeFakeImages = function( $node ) {
        /*--(log.log( "dom/sanitization/removeFakeImages()" );)--*/

        var $container = $node || $( kContent );

        $container.find( kFakeImageSrc ).remove();
    };

    exports.removeFonts = function( $node ) {
        /*--(log.log( "dom/sanitization/removeFonts()" );)--*/

        var $container = $node || $( kContent );

        $container.find( "font" ).each( function() {
            var $this = $( this );

            $this.replaceWith( $this.contents() );
        });
    };

    exports.removeSpans = function( $node ) {
        /*--(log.log( "dom/sanitization/removeSpans()" );)--*/

        var $container = $node || $( kContent );

        $container.find( "span" ).each( function() {
            var $this = $( this );

            // don't remove spans with a class attribute
            if ( $this.attr( "class" ) || style.hasBackgroundStyle( $this ) ) {
                return;
            }

            $this.replaceWith( $this.contents() );
        });
    };

    return exports;
});

define( 'dom/main',[
    "jquery",
    "../config/rte/main"
], function( $, rteConfig ) {
    

    var exports = {};

    exports.getById = function( id ) {
        return document.getElementById( id );
    };

    exports.getDocument = function( node ) {
        if ( node.nodeType === rteConfig.nodeType.DOCUMENT_NODE ) {
            return node;
        } else if ( node.ownerDocument ) {
            return node.ownerDocument;
        } else if ( node.document ) {
            return node.document;
        } else if ( node.parentNode ) {
            return exports.getDocument( node.parentNode );
        } else {
            throw new Error( "/dom/main/getDocument(): no document found for node!" );
        }
    };

    exports.getNodeIndex = function( node ) {
        var i = 0;

        while ( ( node = node.previousSibling ) ) {
            i++;
        }

        return i;
    };

    /**
     * @returns {DomNode} the original node, if it has a ancestor, `null` otherwise.
     */
    // TODO: this method does not do what it says it does. rename it as getClosestIntermediateNode(node, ancestor, selfIsChild)
    exports.getClosestAncestor = function( node, ancestor, selfIsAncestor ) {
        var p, n = selfIsAncestor ? node : node.parentNode;

        while ( n ) {
            p = n.parentNode;

            if ( p === ancestor ) {
                return n;
            }

            n = p;
        }

        return null;
    };

    exports.getCommonAncestor = function( node1, node2 ) {
        var ancestors = [], n;

        for ( n = node1; n; n = n.parentNode ) {
            ancestors.push( n );
        }

        for ( n = node2; n; n = n.parentNode ) {
            if ( ancestors.indexOf( n ) > -1 ) {
                return n;
            }
        }

        return null;
    };

    exports.appendToEditorEnd = function( node ) {
        $( rteConfig.selector.RTE ).append( node );
    };

    return exports;
});

define( 'dom/range/serialization/wrapped/dom',[
    "../../../main"
], function( dom ) {
    

    var exports = {};

    exports.getRangeDocument = function( range ) {
        return dom.getDocument( range.startContainer );
    };

    exports.getDocument = function( node ) {
        return dom.getDocument( node );
    };

    return exports;
});

define( 'dom/range/serialization/wrapped/wrappedrange/main',[
    "../../../../../config/rte/main"
], function( rteConfig ) {
    

    /**
     * Only `WrappedSelection` should use this.
     */

    var rangeProperties = rteConfig.rangeProperties;

    function updateRangeProperties( range ) {
        var i = rangeProperties.length, prop;

        while ( i-- ) {
            prop = rangeProperties[i];
            range[prop] = range.nativeRange[prop];
        }
    }

    function WrappedRange( range ) {
        if ( !range ) {
            throw new Error( "Range must be specified" );
        }

        this.nativeRange = range;

        updateRangeProperties( this );
    }

    return WrappedRange;
});

define( 'dom/range/comparison',[
    "../main"
], function( dom ) {
    

    var exports = {};

    function doCompare( n, childA, childB ) {
        while ( n ) {
            if ( n === childA ) {
                return -1;
            } else if ( n === childB ) {
                return 1;
            }

            n = n.nextSibling;
        }

        throw new Error( "/range/serialization/wrapped/wrappedselection/validation/compare(): Should not be here!" );
    }

    function compareCase4( nodeA, nodeB ) {
        var root = dom.getCommonAncestor( nodeA, nodeB ),
            childA = ( nodeA === root ) ? root : dom.getClosestAncestor( nodeA, root, true ),
            childB = ( nodeB === root ) ? root : dom.getClosestAncestor( nodeB, root, true ),
            n;

        if ( childA === childB ) {
            throw new Error( "/range/serialization/wrapped/wrappedselection/validation/compare(): comparePoints got to case 4 and childA and childB are the same!" );
        } else {
            n = root.firstChild;

            return doCompare( n, childA, childB );
        }
    }

    function compareCase1( offsetA, offsetB ) {
        return offsetA === offsetB ? 0 : (offsetA < offsetB) ? -1 : 1;
    }

    function compareCase2( offsetA, nodeC ) {
        return offsetA <= dom.getNodeIndex( nodeC ) ? -1 : 1;
    }

    function compareCase3( nodeC, offsetB ) {
        return dom.getNodeIndex( nodeC ) < offsetB ? -1 : 1;
    }

    /**
     * Implements <http://www.w3.org/TR/DOM-Level-2-Traversal-Range/ranges.html#Level-2-Range-Comparing>
     */
    exports.compare = function( nodeA, offsetA, nodeB, offsetB ) {
        var nodeC;

        if ( nodeA === nodeB ) {

            // Case 1: nodes are the same
            return compareCase1( offsetA, offsetB );

        } else if ( ( nodeC = dom.getClosestAncestor( nodeB, nodeA, true ) ) ) {

            // Case 2: node C (container B or an ancestor) is a child node of A
            return compareCase2( offsetA, nodeC );
        } else if ( (nodeC = dom.getClosestAncestor( nodeA, nodeB, true ) ) ) {

            // Case 3: node C (container A or an ancestor) is a child node of B
            return compareCase3( nodeC, offsetB );
        } else {

            // Case 4: containers are siblings or descendants of siblings
            return compareCase4( nodeA, nodeB );
        }
    };

    return exports;
});

define( 'dom/range/serialization/wrapped/wrappedselection/validation',[
    "../../../comparison"
], function( range ) {
    

    var exports = {};

    exports.isBackwards = function( sel ) {
        var backwards = false;

        if ( sel.anchorNode ) {
            backwards = ( range.compare( sel.anchorNode, sel.anchorOffset, sel.focusNode, sel.focusOffset ) === 1 );
        }

        return backwards;
    };

    exports.isCollapsed = function selectionIsCollapsed( sel ) {
        return sel.anchorNode === sel.focusNode && sel.anchorOffset === sel.focusOffset;
    };

    return exports;
});

define( 'dom/range/serialization/wrapped/wrappedselection/main',[
    "../../../../../config/rte/main",
    "../wrappedrange/main",
    "./validation"
], function( rteConfig, WrappedRange, validation ) {
    

    /**
     * Only `ui.range.serialization.main` should use this.
     */

    function WrappedSelection( selection, win ) {
        this.nativeSelection = selection;
        this._ranges = [];
        this.win = win;
        this.anchorNode = null;
        this.focusNode = null;
        this.anchorOffset = 0;
        this.focusOffset = 0;
        this.rangeCount = 0;
        this.isCollapsed = true;

        this.refresh();
    }

    function updateSelectionFromRange( sel, range, backwards ) {
        var anchorPrefix = backwards ? rteConfig.rangePrefix.END : rteConfig.rangePrefix.BEGIN,
            focusPrefix = backwards ? rteConfig.rangePrefix.BEGIN : rteConfig.rangePrefix.END;

        sel.anchorNode = range[ anchorPrefix + rteConfig.rangeAttribute.CONTAINER ];
        sel.anchorOffset = range[ anchorPrefix + rteConfig.rangeAttribute.OFFSET ];
        sel.focusNode = range[ focusPrefix + rteConfig.rangeAttribute.CONTAINER ];
        sel.focusOffset = range[ focusPrefix + rteConfig.rangeAttribute.OFFSET ];
    }

    function updateEmptySelection( sel ) {
        sel.anchorNode = sel.focusNode = null;
        sel.anchorOffset = sel.focusOffset = 0;
        sel.rangeCount = 0;
        sel.isCollapsed = true;
        sel._ranges.length = 0;
    }

    function prepareSelectionRanges( sel ) {
        for ( var i = 0, len = sel.rangeCount; i < len; ++i ) {
            sel._ranges[ i ] = new WrappedRange( sel.nativeSelection.getRangeAt( i ) );
        }

        updateSelectionFromRange( sel, sel._ranges[ sel.rangeCount - 1 ], validation.isBackwards( sel.nativeSelection ) );

        sel.isCollapsed = validation.isCollapsed( sel );
    }

    function refreshSelection( sel ) {
        sel._ranges.length = sel.rangeCount = sel.nativeSelection.rangeCount;

        if ( sel.rangeCount ) {
            prepareSelectionRanges( sel );
        } else {
            updateEmptySelection( sel );
        }
    }

    WrappedSelection.prototype.refresh = function() {
        refreshSelection( this );
    };

    WrappedSelection.prototype.getAllRanges = function() {
        return this._ranges.slice( 0 );
    };

    return WrappedSelection;
});

define( 'dom/range/serialization/main',[
    "./wrapped/dom",
    "../../../dom/validation/main",
    "../../../dom/main",
    "./wrapped/wrappedselection/main"
], function( dom, validation, domCore, WrappedSelection ) {
    

    var exports = {},

        // TODO: to constants.
        windowPropertyName = "_jiveSelection";

    function serializePosition( node, offset, rootNode ) {
        var pathBits = [], n = node;

        rootNode = rootNode || domCore.getDocument( node ).documentElement;

        while ( n && n !== rootNode ) {
            pathBits.push( domCore.getNodeIndex( n ) );

            n = n.parentNode;
        }

        // TODO: to constants.
        return pathBits.join( "/" ) + ":" + offset;
    }

    function getWinSelection( winParam ) {
        return ( winParam || window ).getSelection();
    }

    var counter = 0;

    function serializeRange( range ) {
        var rootNode = dom.getRangeDocument( range ).documentElement,
            serialized;

        if ( !validation.isAncestor( rootNode, range.commonAncestorContainer, true ) ) {
            throw new Error( "serializeRange: range is not wholly contained within specified root node." );
        }

        // TODO: to consts.
        serialized = serializePosition( range.startContainer, range.startOffset, rootNode ) + "," +
            serializePosition( range.endContainer, range.endOffset, rootNode );
        serialized += "{" + (counter++) + "}";

        return serialized;
    }

    function getSelection( win ) {
        win = win || window;

        var sel = win[ windowPropertyName ],
            nativeSel = getWinSelection( win );

        if ( sel ) {
            sel.nativeSelection = nativeSel;
        } else {
            sel = new WrappedSelection( nativeSel, win );

            win[ windowPropertyName ] = sel;
        }

        return sel;
    }

    function serializeSelection() {
        var selection = getSelection(),
            ranges = selection.getAllRanges(),
            serializedRanges = [],
            i, len;

        for ( i = 0, len = ranges.length; i < len; ++i ) {
            serializedRanges[ i ] = serializeRange( ranges[ i ] );
        }

        return serializedRanges.join( "|" );
    }

    exports.getSerializedRange = function() {
        var serializedRangeSelected = serializeSelection(),
            beginningCurly;

        if ( serializedRangeSelected ) {

            // TODO: we won't need this curly check if we don't calculate checksums.
            beginningCurly = serializedRangeSelected.indexOf( "{" );

            serializedRangeSelected = serializedRangeSelected.substring( 0, beginningCurly );

            return serializedRangeSelected;
        }

        return "";
    };

    return exports;
});

define( 'dom/range/main',[
    "../../config/rte/main",
    "../../util/log",
    "../main",
    "./serialization/main",
    "jquery"
], function( rteConfig, log, dom, serialization, $ ) {
    

    /**
     * This module is protected. only ui/range/* and legacy/ui/range/*
     * modules SHOULD require it.
     *
     * This class ia basically a Façade around `getSelection` api,
     * to account for possible cross-device inconsistencies.
     */

    var exports = {};

    function removeNode( previous ) {
        var temp = previous.previousSibling;

        previous.parentNode.removeChild( previous );

        return temp;
    }

    function trimPrevious( previous, text ) {
        var nodeValue = previous.nodeValue,
            nodeValueEndTextStartIndex,
            nodeValueEndText;

        if ( nodeValue === text ) {
            previous.parentNode.removeChild( previous );
        }

        nodeValueEndTextStartIndex = nodeValue.length - text.length;
        nodeValueEndText = nodeValue.slice( nodeValueEndTextStartIndex );
        if ( nodeValueEndText === text ) {
            previous.nodeValue = nodeValue.substring( 0, nodeValueEndTextStartIndex );
        }
    }

    //TODO: get rid of this constant, it will never change.
    //kBlank = rteConfig.html.BLANK;

    exports.create = function() {
        return  document.createRange();
    };

    exports.getWindowSelection = function() {
        return window.getSelection();
    };

    exports.clearWindowSelection = function() {
        var sel = window.getSelection();

        sel.removeAllRanges();
    };

    exports.getSerializedRange = function() {
        return serialization.getSerializedRange();
    };

    exports.hasWindowSelection = function() {
        return window.getSelection().toString().length > 0;
    };

    exports.getSelectedRange = function() {
        var selection = exports.getWindowSelection();

        if (selection.rangeCount === 0) {
            return null;
        }

        return selection.getRangeAt(0);
    };

    exports.setSelectedRange = function( range ) {
        var selection = exports.getWindowSelection();

        selection.removeAllRanges();
        selection.addRange( range );
    };

    exports.insertNode = function( range, node ) {
        range.insertNode( node );
    };

    exports.getBoundingClientRect = function( range ) {
        return range.getBoundingClientRect();
    };

    /**
     * **node** should be a DOM node. You cannot insert a **documentFragment** using selection ranges.
     */
    exports.insertNodeAtCaret = function( node ) {
        /*--(log.log( "/dom/range/main/insetNodeAtCaret()" );)--*/

        var sel = exports.getWindowSelection(),
            activeRange, startNode, startOffset, boundaryRange;

        if ( sel.rangeCount <= 0 ) {
            /*--(log.log( "/dom/range/main/insertNodeAtCaret(): zero range count, appending to editor end." );)--*/

            dom.appendToEditorEnd( node );

            activeRange = exports.create();
            activeRange.setStartAfter( node );

            exports.setSelectedRange( activeRange );

            return;
        }

        activeRange = sel.getRangeAt( 0 );
        startNode = activeRange.startContainer;
        startOffset = activeRange.startOffset;

        boundaryRange = activeRange.cloneRange();

        boundaryRange.setStart( startNode, startOffset );
        boundaryRange.collapse( true );

        exports.insertNode( boundaryRange, node );

        // Reselect the original text
        activeRange.setStartAfter( node );

        exports.setSelectedRange( activeRange );

        /*--(log.log( "/dom/range/main/insetNodeAtCaret(): exited" );)--*/
    };

    exports.setCaretAfter = function( range, pivot ) {

        // Set the caret to the end of the text.
        range.selectNode( pivot );

        range.collapse();

        exports.setSelectedRange( range );
    };

    // TODO: this belongs to /dom/main instead.
    exports.deleteLastOccurrenceOfTextBefore = function( pivot, text ) {
        var previous = pivot.previousSibling,
            kTextNode = rteConfig.nodeType.TEXT_NODE,
            nodeValue;

        while ( previous ) {
            nodeValue = previous.nodeValue;

            if ( previous.nodeType === kTextNode ) {
                if ( !nodeValue ) {
                    previous = removeNode( previous );
                } else {
                    trimPrevious( previous, text );

                    break;
                }
            } else {
                break;
            }
        }
    };

    exports.replaceSelectionWithNodes = function( nodes ) {
        var activeRange = exports.getSelectedRange(),
            i, len;

        if ( !activeRange ) {
            return;
        }

        activeRange.deleteContents();

        for ( i = 0, len = nodes.length; i < len; i++ ) {
            exports.insertNode( activeRange, nodes[ i ] );
        }

        activeRange.collapse( false );

        exports.setSelectedRange( activeRange );
    };

    exports.formatEmptyLine = function( activeRange ) {
        if ( !activeRange.collapsed || $( activeRange.commonAncestorContainer ).text() ) {
            return false;
        }

        // TODO: rteConfig.selector.SPAN is redundant, remove it and replace its usages with "span"
        // because we don't aim to select a span, we are creating a span node, they are two differnet things.
        // and keeping "span" as a constant does not make sense, since we won't change the spans we use into
        // something else. -- same is true for rteConfig.selector.PARAGRAPH
        var span = document.createElement( "span" );

        span.innerHTML = "&nbsp;";

        exports.insertNode( activeRange, span );

        return true;
    };

    return exports;
});

define( 'dom/command/validation',[
], function() {
    

    var exports = {},

        // TODO: to constants.
        kBlockLevelRegExp = /^(p|h[1-6])$/i;

    exports.isBlockLevel = function( commandValue ) {
        return kBlockLevelRegExp.test( "" + commandValue );
    };

    return exports;
});

define('dom/command/main',[
    "../../state/rte",
    "../../util/log",
    "../range/main",
    "../validation/main",
    "./validation"
], function( rteState, log, range, domValidation, commandValidation ) {
    

    var exports = {};

    exports.getState = function( name ) {
        return ( name === "formatBlock" ) ? document.queryCommandValue( name ) : document.queryCommandState( name );
    };

    // TODO: too many parameters. simplify.
    exports.exec = function( command, showUi, value, activeRange, formatted ) {
        /*--(log.log( "/dom/command/main/exec: '" + command + "' '" + value + "'");)--*/

        // Caveat: do not remove this pivot, otherwise formatting gets funky.
        var pivot = document.createTextNode( "" );

        range.insertNodeAtCaret( pivot );

        if ( domValidation.isInsideListItem( pivot ) && commandValidation.isBlockLevel( value ) ) {
            /*--(log.log( "/dom/command/main/exec: cannot put a heading inside a list item. Exiting." );)--*/

            return;
        }

        // TODO: to consts.
        // FIX for #ANDROID-720: formatting headings twice was causing problems.
        if ( /h\d/i.test( value ) ) {
            document.execCommand( "formatBlock", null, "p", false);
        }

        document.execCommand( command, showUi, value !== undefined ? value : !rteState.getValue( command ) );

        // TODO: check if this is ever used.
        if ( formatted ) {
            activeRange.deleteContents();
        }

        return true;
    };

    return exports;
});

define('dom/pivot/main',[
    "../../config/rte/main"
], function( rteConfig ) {
    

    var exports = {},

        kText = rteConfig.nodeType.TEXT_NODE;

    function checkNodeState( node ) {
        if ( node.nodeType === kText ) {
            if ( node.nodeValue ) {
                return true;
            }
        } else {
            if ( node.nodeName === "BR" ) {
                return false;
            } else {
                return true;
            }
        }

        return null;
    }

    exports.isBeforePopulated = function( pivot ) {
        if ( !pivot ) {
            return false;
        }

        var node = pivot.previousSibling,
            value;

        while ( node ) {
            value = checkNodeState( node );

            if ( value !== null ) {
                return value;
            }

            node = node.previousSibling;
        }

        return false;
    };

    return exports;
});

define('dom/query/main',[],function() {
    

    var exports = {};

    exports.findClosestAncestor = function( node, nodeName ) {
        if ( !node ) {
            return null;
        }

        var parent = node.parentNode;

        while ( parent ) {
            if ( parent.nodeName === nodeName ) {
                return parent;
            }

            parent = parent.parentNode;
        }

        return null;
    };

    // TODO: to constants.
    var kHeadingRegExp = /^(h[1-6])$/i;

    exports.findClosestHeading = function( node ) {
        if ( !node ) {
            return null;
        }

        var parent = node.parentNode;

        while ( parent ) {
            if ( kHeadingRegExp.test( parent.nodeName ) ) {
                return parent;
            }

            parent = parent.parentNode;
        }

        return null;
    };

    exports.findClosestListItem = function( node ) {
        if ( !node ) {
            return null;
        }

        var parent = node.parentNode;

        while ( parent ) {
            if ( parent.nodeName === "LI" ) {
                return parent;
            }

            parent = parent.parentNode;
        }

        return null;
    };

    return exports;
});

define( 'ui/rendering/validation/main',[
    "../../../util/log",
    "../../../config/rte/main",
    "../../../native/validation",
    "../../../dom/range/main"
], function( log, rteConfig, validation, range ) {
    

    /**
     * Only ui/range/* modules SHOULD include this.
     */

    var exports = {},

        kMentionRegExp = rteConfig.regExp.MENTION,
        kTextNode = rteConfig.nodeType.TEXT_NODE,
        kLink = rteConfig.selector.LINK;

    exports.isSupported = function() {
        return typeof window.getSelection === "function";
    };

    exports.isSelectionAvailable = function() {
        var selection = range.getWindowSelection();

        return !!selection.rangeCount;
    };

    /**
     * Tests whether the @ just typed (i.e. just left of the cursor) is in a
     * good position to be an at-mention.
     *
     * @returns {boolean} `true` if the range is either at the beginning,
     * or at the end, or has a space before and after.
     */
    exports.isValidAt = function( range ) {
        /*--(log.log( "/range/validate/isValidAt()" );)--*/

        var node = range.startContainer;

        if ( !range.collapsed || node.nodeType !== kTextNode || ( node.parentNode && node.parentNode.nodeName.toLowerCase() === kLink ) ) {
            /*--(log.log( "/range/validate/isValidAt: range problems, exiting." );)--*/

            return false;
        }

        return kMentionRegExp.test( node.nodeValue.slice( range.startOffset - 2, range.startOffset + 1 ) );
    };

    return exports;
});

define( 'ui/rendering/state/main',[
    "../../../config/rte/main",
    "../../../state/rte",
    "../validation/main",
    "../../../dom/range/main",
    "jquery"
], function( rteConfig, rteState, validation, range, $ ) {
    

    var exports = {};

    // TODO: move to dom/range/main
    function add( activeRange ) {
        var selection = range.getWindowSelection();

        selection.removeAllRanges();
        selection.addRange( activeRange );
    }

    function createInsertion() {
        if ( !validation.isSelectionAvailable() ) {
            return null;
        }

        var activeRange = range.getWindowSelection().getRangeAt( 0 );

        if ( !activeRange ) {
            return null;
        }

        activeRange.setStart( activeRange.endContainer, activeRange.endOffset );

        return activeRange;
    }

    exports.restorePoint = function( cancelled ) {
        var activeRange = rteState.insertionRange;

        if ( !activeRange ) {
            return;
        }

        rteState.insertionRange = null;

        $( rteConfig.selector.RTE )[ 0 ].focus();

        if ( cancelled ) {
            activeRange.collapse( true );
        }

        add( activeRange );
    };

    exports.savePoint = function() {
        rteState.insertionRange = createInsertion();
    };

    exports.forgetRange = function() {
        rteState.insertionRange = null;
    };

    exports.getRange = function() {
        return rteState.insertionRange;
    };

    exports.setRange = function( activeRange ) {
        rteState.insertionRange = activeRange;
    };

    return exports;
});

define( 'ui/rendering/main',[
    "../../config/rte/main",
    "../../dom/command/main",
    "../../dom/main",
    "../../dom/pivot/main",
    "../../dom/query/main",
    "../../dom/range/main",
    "../../dom/validation/main",
    "../../state/rte",
    "../../util/log",
    "../../util/timer",
    "./state/main",
    "./validation/main",
    "jquery"
], function( rteConfig, domCommand, dom, domPivot, query, range, validation, rteState, log, timer, state, rangeValidation, $ ) {
    

    /**
     * Range utilities.
     * All `range` operations should be done via this module.
     */

    var exports = {},

        kHeadingRegExp = rteConfig.regExp.HEADING,

        getById = dom.getById;

    // TODO: this file has grown in size, split it.

    exports.createFromPoint = function( x, y ) {
        /*--(log.log( "/rendering/main/createFromPoint()" );)--*/

        return document.caretRangeFromPoint( x - window.scrollX, y - window.scrollY );
    };

    exports.addSelectedNode = function( node ) {
        /*--(log.log( "/rendering/main/addSelectedNode()" );)--*/

        var selection = range.getWindowSelection(),
            activeRange = range.create();

        selection.removeAllRanges();

        activeRange.selectNode( node );

        selection.addRange( activeRange );
    };

    exports.addFromTarget = function( target ) {
        /*--(log.log( "/rendering/main/addFromTarget()" );)--*/

        var selection = range.getWindowSelection(),
            activeRange;

        selection.removeAllRanges();

        activeRange = range.create();

        activeRange.setStartBefore( target );
        activeRange.collapse();

        selection.addRange( activeRange );
    };

    exports.addFromRectangle = function( rect, afterRangeCreationDelegate ) {
        /*--(log.log( "/rendering/main/addFromRectangle()" );)--*/

        var selection,
            cornerPad = rteConfig.rte.IMAGE_CORNER_PADDING,
            //TODO: to dom/range/main
            activeRange = document.caretRangeFromPoint(
                    Math.max( rect.left - cornerPad, 0 ) - window.scrollX,
                    Math.max( rect.top - cornerPad, 0 ) - window.scrollY
                );

        afterRangeCreationDelegate();

        selection = range.getWindowSelection();

        selection.removeAllRanges();
        selection.addRange( activeRange );
    };

    exports.addFromPosition = function( pos ) {
        /*--(log.log( "/rendering/main/addFromPosition()" );)--*/

        var activeRange = range.create();

        activeRange.setStart( pos.startContainer, pos.startOffset );
        activeRange.setEnd( pos.endContainer, pos.endOffset );

        range.getWindowSelection().addRange( activeRange );
    };

    /**
     * This fixes #ANDROID-1058 and #ANDROID-1059
     */
    function createParagraphAfterNode( heading ) {
        var p = document.createElement( "p" ),
            zwsp = document.createTextNode( "\u200b" );

        p.appendChild( zwsp );

        // TODO: make this a dom utility method.
        if ( heading.nextSibling ) {

            // TODO: make this a dom utility method
            heading.parentNode.insertBefore( p, heading.nextSibling );
        } else {
            heading.parentNode.appendChild( p );
        }

        // TODO: refactor this to range.setCaretAfter( zwsp )
        range.setCaretAfter( range.create(), zwsp );
    }

    // TODO: this is almost identical to the function above, be DRY.
    function createListItemAfterNode( listItem ) {
        var li = document.createElement( "li" ),
            zwsp = document.createTextNode( "\u200b" );

        li.appendChild( zwsp );

        if ( listItem.nextSibling ) {
            listItem.parentNode.insertBefore( li, listItem.nextSibling );
        } else {
            listItem.parentNode.appendChild( li );
        }

        range.setCaretAfter( range.create(), zwsp );
    }

    exports.insertNewLineAtCaret = function() {
        /*--(log.log( "/rendering/main/insertNewLineAtCaret()" );)--*/

        var pivot = document.createTextNode( "" );

        range.insertNodeAtCaret( pivot );

        var heading = query.findClosestHeading( pivot );

        if ( heading ) {
            createParagraphAfterNode( heading );

            return true;
        }

        var listItem = query.findClosestListItem( pivot );

        if ( listItem ) {
            if ( listItem.innerHTML === "" || listItem.innerHTML === "\u200b" ) {
                var parent = listItem.parentNode;

                parent.removeChild( listItem );

                createParagraphAfterNode( parent );

                return true;
            }

            createListItemAfterNode( listItem );

            return true;
        }

        var node = document.createElement( "br" );

        range.insertNodeAtCaret( node );

        // In "some" JellyBeans simply inserting BR does not move the caret
        // to the next line.
        // Inserting <span><br></span> solves this, but it creates styling
        // problems such as #ANDROID-1058.
        // As a better solution, we insert a new line at the caret position
        // and then we insert a zero-width spacepoint.

        // TODO: to consts.
        // Zero-width space point.
        var zwsp = document.createTextNode( "\u200b" );

        if ( node.nextSibling ) {
            node.parentNode.insertBefore( zwsp, node.nextSibling );
        } else {
            node.parentNode.appendChild( zwsp );
        }

        var activeRange = range.create();

        range.setCaretAfter( activeRange, zwsp );

        return true;
    };

    function detachNodes( garbage ) {
        var node;

        while ( garbage.length ) {
            node = garbage.pop();

            node.parentNode.removeChild(node);
        }
    }

    function deleteCharacterAtCaretUsingWindowSelectionApi( pivot ) {
        var selection = range.getWindowSelection(),
            activeRange = selection.getRangeAt( 0 ),
            clonedRange = activeRange.cloneRange();

        clonedRange.setStart( activeRange.startContainer, activeRange.startOffset - 1 );
        clonedRange.setEnd( activeRange.endContainer, activeRange.startOffset );
        clonedRange.deleteContents();

        selection.removeAllRanges();

        selection.addRange( clonedRange );

        pivot.parentNode.removeChild( pivot );
    }

    function detachPreviousGarbageNodes( pivot ) {
        var garbage = [],
            pivotPreviousSibling = pivot.previousSibling;

        while ( pivotPreviousSibling && pivotPreviousSibling.nodeType === rteConfig.nodeType.TEXT_NODE && !pivotPreviousSibling.nodeValue ) {
            garbage.push( pivotPreviousSibling );

            pivotPreviousSibling = pivotPreviousSibling.previousSibling;
        }

        detachNodes( garbage );
    }

    function transferNodesBackwards( pivotParent, previousParent ) {
        while ( pivotParent.hasChildNodes() ) {
            if ( previousParent.nextSibling ) {
                previousParent.parentNode.insertBefore( pivotParent.lastChild, previousParent.nextSibling );
            } else {
                previousParent.parentNode.appendChild( pivotParent.firstChild );
            }
        }
    }

    function transferNodesInside( pivotParent, previousParent ) {
        while ( pivotParent.hasChildNodes() ) {
            previousParent.appendChild( pivotParent.firstChild );
        }
    }

    function discardLastLineBreak( children ) {
        var garbage = [];

        for ( var i = children.length - 1; i >= 0; i-- ) {
            var child = children[ i ];

            if ( child.nodeName === "BR" ) {
                garbage.push( child );

                break;
            }
        }

        detachNodes( garbage );
    }

    // TODO: to a module.
    var processing = {
        processEmptyTextNode: function( previousParent, pivotParent, garbage ) {
            garbage.push( previousParent );

            return false;
        },
        processNonEmptyTextNode: function( previousParent, pivotParent, garbage ) {
            detachNodes( garbage );

            /*--(log.log( "/ui/rendering/main/processPivot: Meaningful text node found." );)--*/

            transferNodesBackwards( pivotParent, previousParent );

            return true;
        },
        processLineBreak: function( previousParent, pivotParent, garbage ) {
            garbage.push( previousParent );

            /*--(log.log("/ui/rendering/main/processPivot: Removing line break.");)--*/

            detachNodes( garbage );

            return true;
        },
        processListItem: function( previousParent, pivotParent, garbage ) {
            detachNodes( garbage );

            transferNodesInside( pivotParent, previousParent );

            pivotParent.parentNode.removeChild( pivotParent );

            return true;
        },
        processDefault: function( previousParent, pivotParent, garbage ) {
            detachNodes( garbage );

            var children = previousParent.childNodes;

            discardLastLineBreak( children );

            if ( previousParent.innerHTML ) {
                /*--(log.log( "/ui/rendering/main/processPivot: Transferring nodes inside." );)--*/

                transferNodesInside( pivotParent, previousParent );
            } else {
                /*--(log.log( "/ui/rendering/main/processPivot: Removing previous parent." );)--*/

                previousParent.parentNode.removeChild( previousParent );
            }

            if ( pivotParent.nodeName === "LI" ) {
                /*--(log.log( "pivot parent is a list item!" );)--*/

                pivotParent.parentNode.removeChild( pivotParent );
            }

            return true;
        }
    };

    function processPivot( previousParent, pivotParent, garbage ) {
        var kText = rteConfig.nodeType.TEXT_NODE;

        // Empty text node.
        if ( previousParent.nodeType === kText && !previousParent.nodeValue ) {
            return processing.processEmptyTextNode( previousParent, pivotParent, garbage );
        }

        if ( previousParent.nodeType === kText ) {
            return processing.processNonEmptyTextNode( previousParent, pivotParent, garbage );
        }

        if ( previousParent.nodeName === "BR" ) {
            return processing.processLineBreak( previousParent, pivotParent, garbage );
        }

        if ( previousParent.nodeName === "LI" ) {
            return processing.processListItem( previousParent, pivotParent, garbage );
        }

        return processing.processDefault( previousParent, pivotParent, garbage );
    }

    function isBlankNode( node ) {
        var next = node.firstChild;

        while ( next ) {
            if ( next.nodeType !== rteConfig.nodeType.TEXT_NODE ) {
                return false;
            }

            if ( next.nodeType === rteConfig.nodeType.TEXT_NODE && next.nodeValue ) {
                return false;
            }

            next = next.nextSibling;
        }

        return true;
    }

    function shiftNodes( pivotParent, pivotParentParentParent ) {
        while ( pivotParent.hasChildNodes() ) {
            pivotParentParentParent.parentNode.insertBefore( pivotParent.firstChild, pivotParentParentParent );
        }
    }

    function postShiftProcess( pivotParent, pivotParentParent ) {
        if ( isBlankNode( pivotParentParent ) ) {
            pivotParentParent.parentNode.removeChild( pivotParentParent );
        } else if ( isBlankNode( pivotParent ) ) {
            pivotParent.parentNode.removeChild( pivotParent );
        }
    }

    function shiftPivotParentChildrenOneLevelUp( pivotParent, garbage ) {
        /*--(log.log( "/ui/rendering/main/injectNextSiblingNearPreviousParent: Did not break." );)--*/

        detachNodes( garbage );

        var pivotParentParent = pivotParent.parentNode;

        // ...<div><div>pvt...</div>... becomes
        // ...<div>pvt...</div><div>...
        if ( pivotParentParent && pivotParentParent.parentNode ) {
            /*--(log.log( "Shifting the pivot parent one level up." );)--*/

            var pivotParentParentParent = pivotParentParent.parentNode;

            if (pivotParent.nodeName === "LI" && ( pivotParentParentParent.nodeName !== "UL" || pivotParentParentParent.nodeName !== "OL" )) {
                /*--(log.log( "in a list item..." );)--*/

                shiftNodes( pivotParent, pivotParentParentParent );
                postShiftProcess( pivotParent, pivotParentParent );
            } else {
                /*--(log.log( "not in a list item." );)--*/

                pivotParentParentParent.insertBefore( pivotParent, pivotParentParent );
            }
        }
    }

    function injectNextSiblingNearPreviousParent( pivot ) {
        /*--(log.log( "injectNextSiblingNearPreviousParent" );)--*/

        var pivotParent = pivot.parentNode,
            previousParent = pivotParent.previousSibling,

            shouldBreak = false,

            garbage = [],

            didBreak = false,

            lastNonNullPrev;

        /*--(log.log( "before while" );)--*/

        while ( previousParent ) {
            shouldBreak = processPivot( previousParent, pivotParent, garbage );

            if ( shouldBreak ) {
                didBreak = true;

                break;
            }

            lastNonNullPrev = previousParent;
            previousParent = previousParent.previousSibling;
        }

        /*--(log.log( "after while" );)--*/

        if ( !didBreak ) {
            shiftPivotParentChildrenOneLevelUp( pivotParent, garbage );
        }

        /*--(log.log( "exit injectNextSiblingNearPreviousParent" );)--*/
    }

    function processLastLineBreak( last ) {
        if ( last.nodeType === rteConfig.nodeType.TEXT_NODE && !last.nodeValue ) {
            last.parentNode.removeChild( last );

            return false;
        }

        if ( last.nodeValue ) {
            return true;
        }

        if ( last.nodeName === "BR" ) {
            last.parentNode.removeChild( last );

            return true;
        }

        return true;
    }

    function removeLastLineBreak( node ) {
        var last = node.lastSibling;

        if ( !last ) {
            return;
        }

        var shouldBreak = false;

        while ( last ) {
            shouldBreak = processLastLineBreak( last );

            if ( shouldBreak ) {
                break;
            }

            last = node.lastSibling;
        }
    }

    function processNonEmptyPreviousSibling( pivot ) {
        removeLastLineBreak( pivot.previousSibling );

        var pivotPreviousSibling = pivot.previousSibling,
            nodes = [],
            node = pivot;

        while ( node ) {
            nodes.push( node );

            node = node.nextSibling;
        }

        while ( nodes.length ) {
            pivotPreviousSibling.appendChild( nodes.shift() );
        }
    }

    function tryStandardBackspace( pivot ) {

        // For Android, JellyBean and below, range api does not work as expected at times.
        // Especially if the caret is the first thing on a new line.
        //
        // If it is not and there's stuff before, yield it to range API, it works well.
        //
        // So we use the more standard selection API when we can, and use a fallback
        // mechanism otherwise.
        if ( domPivot.isBeforePopulated( pivot ) ) {
            try {
                deleteCharacterAtCaretUsingWindowSelectionApi( pivot );

                return true;
            } catch ( exception ) {
                /*--(log.log( "Failed to delete using regular api... Will try fallback." );)--*/
            }
        }

        return false;
    }

    function cleanupInitialGarbage( pivot ) {
        /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: fallback" );)--*/

        if ( pivot.previousSibling ) {
            /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: pivot has previous sibling." );)--*/

            detachPreviousGarbageNodes( pivot );
        }
    }

    function fixMissingPivot( pivot ) {
        if ( !pivot.parentNode ) {
            range.insertNodeAtCaret( pivot );
        }
    }

    function processPivotWithPreviousSibling( pivot ) {
        /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: pivot still has previous sibling." );)--*/

        if ( pivot.previousSibling ) {
            if ( pivot.previousSibling.nodeType ===  rteConfig.nodeType.TEXT_NODE ) {
                if ( !pivot.previousSibling.nodeValue ) {
                    pivot.previousSibling.parentNode.removeChild( pivot.previousSibling );

                    processPivotWithPreviousSibling( pivot );

                    return;
                }

                pivot.previousSibling.nodeValue = pivot.previousSibling.nodeValue.substring( 0, pivot.previousSibling.nodeValue.length - 1 );
            } else {
                if ( pivot.previousSibling.innerHTML ) {
                    processNonEmptyPreviousSibling( pivot );
                } else {
                    pivot.previousSibling.parentNode.removeChild( pivot.previousSibling );
                }
            }
        }

        exports.ensureCaret( pivot );
    }

    function processDefaultPivotFallback( pivot ) {
        /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: pivot has no previous sibling." );)--*/

        injectNextSiblingNearPreviousParent( pivot );

        exports.ensureCaret( pivot );
    }

    function mapCommandToNodeName( command ) {
        switch (command) {
            case "bold":
                return "B";
            case "italic":
                return "I";
            case "underline":
                return "U";
            default:
                return "";
        }
    }

    exports.deleteCharacterAtCaret = function() {
        /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: Begin." );)--*/

        var pivot = document.createTextNode( "" );

        range.insertNodeAtCaret( pivot );

        if ( tryStandardBackspace( pivot ) ) {
            /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: Exit 1." );)--*/

            return;
        }

        // #region Fallback

        fixMissingPivot( pivot );

        cleanupInitialGarbage( pivot );

        if ( pivot.previousSibling ) {
            processPivotWithPreviousSibling( pivot );

            /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: Exit 2." );)--*/

            return;
        }

        processDefaultPivotFallback( pivot );

        /*--(log.log( "/rendering/main/deleteCharacterAtCaret() :: End." );)--*/

        // #endregion
    };

    exports.removeSelected = function() {
        /*--(log.log( "/rendering/main/removeSelected()" );)--*/

        var selection = range.getWindowSelection(),
            activeRange, startContainer, clonedRange;

        if ( selection.rangeCount ) {
            activeRange = selection.getRangeAt( 0 );

            startContainer = activeRange.startContainer;

            activeRange.deleteContents();

            clonedRange = activeRange.cloneRange();

            clonedRange.setStartAfter( startContainer );
            clonedRange.collapse( true );

            selection.removeAllRanges();
            selection.addRange( clonedRange );
        }
    };

    exports.getText = function() {
        /*--(log.log( "/rendering/main/getText()" );)--*/

        var selection = range.getWindowSelection(),
            data = "",
            activeRange, contents, children, i, len;

        if ( selection.rangeCount ) {
            activeRange = selection.getRangeAt( 0 );
            contents = activeRange.cloneContents();

            children = contents.childNodes;

            for ( i = 0, len = children.length; i < len; i++ ) {
                data += ( children[ i ].innerText || children[ i ].data );
            }
        }

        return data;
    };

    exports.insertImageAtCaret = function( src ) {
        /*--(log.log( "/rendering/main/insertImageAtCaret()" );)--*/

        var node = document.createElement( "img" ),
            br1 = document.createElement( "br" ),
            br2 = document.createElement( "br" ),
            span = document.createElement( "span" );

        node.className = "jive-image";
        node.src = src;

        span.appendChild( br1 );
        span.appendChild( node );
        span.appendChild( br2 );

        range.insertNodeAtCaret( span );

        /*--(log.log( "inserted image at caret" );)--*/

        return true;
    };

     // TODO: rename -- this does not insert text, it inserts a wrapped span; its name is misleading.
    exports.insertTextAtCaret = function( text ) {

        // TODO: to constants
        var data = text.replace( /\r\n|\n/g, "<br>" ),
            span = document.createElement( rteConfig.selector.SPAN );

        span.innerHTML = data;

        range.insertNodeAtCaret( span );
    };

    // TODO: check whether this is used at all anymore.
    exports.cacheCurrentInsertion = function() {
        state.savePoint();
    };

    function isArrayLike( obj ) {
        return obj && typeof obj === "object" && obj.length !== undefined;
    }

    function adjustCaretAfterCommand( command, pivot ) {
        if (domCommand.getState(command)) {
            return;
        }

        range.insertNodeAtCaret( pivot );

        if ( validation.isAtTheEndOfParent( pivot ) && validation.hasParentOfType( mapCommandToNodeName( command ), pivot ) ) {
            if ( pivot.parentNode.nextSibling ) {
                pivot.parentNode.parentNode.insertBefore( pivot, pivot.parentNode.nextSibling );
            } else {
                pivot.parentNode.parentNode.appendChild( pivot );
            }

            range.setCaretAfter( range.create(), pivot );
        }
    }

    // TODO: where does formatEmpty come from? it looks like it's never used, if so remove it.
    exports.formatWithCommand = function( showUi, command, value, formatEmpty ) {
        /*--(log.log( "/rendering/main/formatWithCommand()" );)--*/

        var formatted = false,
            activeRange = range.getSelectedRange();

        // TODO: to consts.
        if ( getById( "JiveMarker" ) ) {
            /*--(log.log( "Skipping formatWithCommand() since there is a marker." );)--*/

            // Be defensive.
            // If there's a marker, then the formatting should be handled by alternative means.
            return true;
        }

        if ( !activeRange ) {
            /*--(log.log( "No active range... Skipping formatWithCommand()." );)--*/

            return true;
        }

        if ( formatEmpty ) {
            formatted = range.formatEmptyLine( activeRange );
        }

        var commandValue = isArrayLike( value ) ? value[ 0 ] : value,

            result = domCommand.exec( command, showUi, commandValue, activeRange, formatted ),

            pivot = document.createTextNode( "\u200B" );

        // FIX for #ANDROID-1435
        // Bolding a text, tapping space, and then unbolding it was not unbolding it, on KitKat.
        adjustCaretAfterCommand( command, pivot );

        return result;
    };

    exports.getScrollRectangle = function() {
        /*--(log.log( "/rendering/main/getScrollRectangle()" );)--*/

        if ( !rangeValidation.isSelectionAvailable() ) {
            return;
        }

        var selRange = range.getSelectedRange(),
            span = document.createElement( rteConfig.selector.SPAN ),
            rect = {},
            startNode, startOffset, boundaryRange;

        if ( !selRange ) {
            return;
        }

        span.innerHTML = rteConfig.html.BLANK;

        startNode = selRange.startContainer;
        startOffset = selRange.startOffset;
        boundaryRange = selRange.cloneRange();

        boundaryRange.setStart( startNode, startOffset );

        range.insertNode( boundaryRange, span );

        boundaryRange.collapse( true );

        // #ANDROID-478 using `getBoundingClientRect` retuns unreliable (sometimes negative)
        // values after a certain height aroung ~300px. We are using offests to workaround that.
        rect.height = span.offetHeight;
        rect.width = span.offsetWidth;
        rect.left = span.offsetLeft;
        rect.top = span.offsetTop;

        // to make the Java parser happy
        rect.right = rect.left + 100;
        rect.bottom = rect.top + 100;

        if ( span && span.parentNode ) {
            span.parentNode.removeChild( span );
        }

        // FIX for #ANDROID-516 do not re-apply boundary range; otherwise we get capitalization problems.
        // window.getSelection().addRange( boundaryRange );

        return rect;
    };

    exports.insertMacroWithType = function( url, name, type ) {
        /*--(log.log( "/ui/range/main/insertMacroWithType()" );)--*/

        if ( !rangeValidation.isSelectionAvailable() ) {
            /*--(log.log( "/ui/range/main/insertMacroWithType: not available, exiting." );)--*/

            return false;
        }

        var $link = $( "<a onclick=\"return false;\" class=\"jive_macro " + type + "\"></a>" );

        $link.attr( "href", url );
        $link.text( name );

            // Fixes https://jira.jivesoftware.com/browse/TABDEV-861. Put a space after an @mention to ensure the inserted link
            // is not copied if the user hits enter immediately after the inserted @mention link
        var space = document.createTextNode( rteConfig.html.SPACE),
            pivot = document.createTextNode( "" );

        range.replaceSelectionWithNodes( [ space, $link[ 0 ], pivot ] );

        // FIX #ANDROID-699 inserting @mention was not deleting the @ sign.
        range.deleteLastOccurrenceOfTextBefore( pivot, rteConfig.html.MENTION_START );

        /*--(log.log( "ui/range/main/insertMacroWithType: done." );)--*/

        return true;
    };

    exports.formatBlock = function( activeRange, value ) {
        /*--(log.log( "/rendering/main/formatBlock()" );)--*/

        if ( !kHeadingRegExp.test( value ) ) {
            return false;
        }

        if ( activeRange.startContainer.parentNode.nodeName !== "P" ) {
            return false;
        }

        var documentFragment = activeRange.extractContents(),

            childNodes = $( documentFragment ).children( "p" );

        if ( !childNodes.length ) {
            return false;
        }

        childNodes.each( function() {
            var $this = $( this );

            $this.replaceWith( "<" + value + ">" + $this.html() + "</" + value + ">" );
        });

        range.insertNode( activeRange, documentFragment );

        return true;
    };

    /**
     * Puts caret into marker, if marker has children.
     * Puts caret after marker otherwise.
     */
    exports.ensureCaret = function( marker ) {
        var textNode = marker.firstChild || marker,
            sel = range.getWindowSelection(),
            activeRange = range.create();

        // Set the caret to the end of the text.
        activeRange.selectNode( textNode );

        activeRange.collapse();

        // TODO: this is a common pattern, make it a method.
        sel.removeAllRanges();
        sel.addRange( activeRange );
    };

    exports.restoreContentCaret = function() {
        state.restorePoint( false );
    };

    exports.isAfterAt = function() {
        state.savePoint();

        var caret = state.getRange();

        if ( caret !== null && caret.collapsed ) {
            var range = caret.cloneRange();

            range.setStart( range.startContainer, 0 );

            // TODO: @ -> to constants.
            return range.toString().slice( -1 ) === "@";
        }

        return false;
    };

    exports.prepareMentionStart = function() {

        // TODO: rename: state.forget
        state.forgetRange();
    };

    exports.processPostMentionStart = function() {
        if ( !state.getRange() ) {

            // If insertionRange is not set (by startMention for instance) then save it
            state.savePoint();
        }
    };

    return exports;
});

define( 'ui/placeholder',[
    "../config/rte/main",
    "../dom/sanitization/main",
    "../util/timer",
    "../dom/validation/main",
    "../util/log",
    "jquery"
], function( rteConfig, sanitization, timer, validation, log, $ ) {
    

    var exports = {},

        $placeholder;

    // TODO: get rid of this method. We don't use placeholders anymore.
    exports.render = function() {
        /*--(log.log( "ui/placeholder/render()" );)--*/

    //    timer.setTimeout( function(/*c*/) {
    //        var $rte = $( rteConfig.selector.RTE );
//
 //           if ( validation.isElementEmpty( $rte ) ) {
//
 //               // this ensures that the first paragraph isn't a raw text node, which
  //              // generally causes formatting commands to barf (later paragraphs get
   //             // wrapped in <p> tags automatically)
    //            $rte.html( rteConfig.html.PARAGRAPH );
//
 //               $placeholder.show();
  //          } else {
   //             $placeholder.hide();
    //        }
     //   });
    };

    exports.initialize = function() {
        /*--(log.log( "/ui/placeholder/initialize()" );)--*/

        $( rteConfig.selector.PLACEHOLDER ).each( function() {
            sanitization.removeFonts();

            var $this = $( this ),
                offset = $this.offset(),
                placeholder = $this.clone();

            // TODO: to consts.
            placeholder.text( $this.attr( "placeholder" ) )
                .removeAttr( "contenteditable" )
                .removeAttr( "id" )
                .removeAttr( "placeholder" )
                .addClass( rteConfig.className.JIVE_PLACEHOLDER );

            // TODO: assign a class instead of inlining css.
            placeholder.appendTo( "body" ).css({
                position: "absolute",
                left: offset.left + ( parseInt( $this.css( "padding-left" ), 10 ) || 0 ),
                top: offset.top + ( parseInt( $this.css( "padding-top" ), 10 ) || 0 ),
                zIndex: -1,
                margin: 0,
                color: "darkGray",
                height: "auto"
            }).hide();
        });

        $placeholder = $( rteConfig.selector.JIVE_PLACEHOLDER );

        exports.initialize = function() {};
    };

    return exports;
});

define( 'ui/scroll',[
    "./rendering/main",
    "../util/timer",
    "../native/validation",
    "../state/rte"
], function( rendering, timer, validation, rteState) {
    

    var exports = {};

    function scrollToRectangleAsync( rect ) {
        if ( !rect ) {
            return;
        }

        // TODO: to constants.
        var kScrollOverflowThreshold = 30;

        if ( rect.top >= window.innerHeight - kScrollOverflowThreshold ) {
            timer.setTimeout( function() {
                window.scrollTo(
                    0,
                    rect.top + document.body.scrollTop - Math.floor( window.innerHeight / 2 )
                );
            });
        }
    }

    /**
     * Scrolls down when the cursor position changes; it should stay on-screen.
     */
    exports.set = function() {
        if ( validation.isIos() ) {

            // iOS can adjust the scroll through JavaScript/HTML DOM/Viewport.
            timer.setTimeout( function() {
                scrollToRectangleAsync( rendering.getScrollRectangle() );
            });
        } else {

            // The case is different for Android:
            //
            // FIX for #ANDROID-748: Let the Java-side decide how to scroll.
            rteState.scrollRectangle = JSON.stringify( rendering.getScrollRectangle() );
        }
    };

    return exports;
});

define( 'ui/rendering/mention/main',[
    "../../../config/rte/main",
    "../state/main",
    "../../../dom/range/main",
    "../../../util/log",
    "../validation/main"
], function( rteConfig, state, range, log, validation ) {
    

    var exports = {},

        kTextNode = rteConfig.nodeType.TEXT;

    function computeNodeOffset( offset, node ) {
        while ( node.nextSibling && node.nextSibling.nodeType === kTextNode ) {
            offset -= node.nodeValue.length;

            node = node.nextSibling;
        }

        return {
            node: node,
            offset: offset
        };
    }

    function removeNode( node ) {
        if ( node && node.parentNode ) {
            node.parentNode.removeChild( node );
        }
    }

    function computeNodeOffsetBackwards( prev, offset, node ) {
        while ( prev && prev.nodeType === kTextNode ) {
            offset += prev.nodeValue.length;

            prev.nodeValue += node.nodeValue;

            removeNode( node );

            node = prev;
            prev = node.previousSibling;
        }

        return {
            node: node,
            offset: offset
        };
    }

    function normalizeTextNodes( activeRange ) {
        /*--(log.log( "/range/main/normalizeTextNodes()" );)--*/

        if ( !activeRange.collapsed || activeRange.startContainer.nodeType !== kTextNode ) {
            return activeRange;
        }

        var selection = range.getWindowSelection(),
            node = activeRange.startContainer,
            offset = activeRange.startOffset,
            prev, meta;

        meta = computeNodeOffset( offset, node );

        node = meta.node;
        offset = meta.offset;

        prev = node.previousSibling;

        meta = computeNodeOffsetBackwards( prev, offset, node );

        node = meta.node;
        offset = meta.offset;

        activeRange.setStart( node, offset );
        activeRange.setEnd( node, offset );

        selection.removeAllRanges();

        selection.addRange( activeRange );
    }

    function prepareForMention( activeRange ) {
        /*--(log.log( "/range/main/prepareForMention()" );)--*/

        activeRange.setStart( activeRange.startContainer, activeRange.startOffset - 1 );
    }

    exports.start = function() {
        /*--(log.log( "/range/main/startMention()" );)--*/

        if ( !validation.isSelectionAvailable() ) {
            /*--(log.log( "/range/main/startMention: range not available -- returning." );)--*/

            return { started: false };
        }

        var activeRange = range.getSelectedRange(),
            posRange, pos;

        normalizeTextNodes( activeRange );

        if ( !validation.isValidAt( activeRange ) ) {
            /*--(log.log( "/range/main/startMention: not a valid at, exiting. " );)--*/

            return { started: false };
        }

        activeRange = range.getSelectedRange();
        posRange = activeRange.cloneRange();

        prepareForMention( posRange );

        pos = range.getBoundingClientRect( posRange );

        state.setRange( activeRange );

        /*--(log.log( "/range/main/startMention: started at mention successfully." );)--*/

        return { started: true, position: pos };
    };

    return exports;
});

define( 'ui/mention',[
    "../config/rte/main",
    "../proxy/bridge/main",
    "../proxy/io/main",
    "../state/rte",
    "./rendering/main",
    "./rendering/mention/main",
    "../util/log"
], function( rteConfig, bridge, io, rteState, rendering, mention, log ) {
    

    var exports = {};

    function enqueueMention( position ) {
        bridge.startMention( position );
    }

    /**
     * Starts an @mention wherever the cursor is.
     */
    function start() {
        /*--(log.log( "starting at mention..." );)--*/

        var meta = mention.start();

        if ( !meta.started ) {
            /*--(log.log( "meta not started, exiting." );)--*/

            return;
        }

        /*--(log.log( "started at mention." );)--*/

        enqueueMention( meta.position );

        /*--(log.log( "enqueued command." );)--*/
    }

    exports.check = function() {
        /*--(log.log( "/ui/mention/check(): started");)--*/

        // At this point we don't care if the @ is valid; we just want to know it's an @
        if ( rendering.isAfterAt() ) {
            rendering.prepareMentionStart();

            start();

            rendering.processPostMentionStart();

            io.updateStateChange();
        }

        /*--(log.log( "/ui/mention/check(): done." );)--*/

        return true;
    };

    exports.update = function( meta ) {
        io.updateMention( meta );

        return true;
    };

    exports.insert = function( url, name, type ) {
        return rendering.insertMacroWithType( url, name, type );
    };

    return exports;
});

define( 'ui/content',[
    "../config/rte/main",
    "../util/log",
    "../proxy/io/main",
    "../proxy/bridge/main",
    "../dom/sanitization/main",
    "../state/content",
    "../util/timer",
    "./rendering/main",
    "./placeholder",
    "./scroll",
    "./mention",
    "jquery"
], function( rteConfig, log, io, bridge, sanitization, contentState, timer, rendering, placeholder, scroll, mention, $ ) {
    

    var exports = {};

    exports.initialize = function() {
        /*--(log.log( "ui/content/initialize(): begin." );)--*/

        bridge.setDefaultParagraphSeparator();

        // Provide placeholder text for contenteditable.
        var $rteContent = $( rteConfig.selector.RTE ),
            $renderedContent = $rteContent.find( rteConfig.selector.RENDERED_CONTENT );

        // TABDEV-2534: unwrap the `jive-rendered-content`.
        if ( !$renderedContent.length ) {

            // TODO: this pattern is repeated in two files, move it into a method.
            if ( $( window ).height() ) {
                $rteContent.css( "height", $( window ).height() );
                $rteContent[ 0 ].focus();
            }

            return;
        }

        while ( $renderedContent.parent()[ 0 ] !== $renderedContent[ 0 ] ) {
            $renderedContent.unwrap();
        }

        // Unwrap the `jive-rendered-content` children.
        $renderedContent.children().unwrap();

        // TODO: this pattern is repeated in two files, move it into a method.
        if ( $( window ).height() ) {
            $rteContent.css( "height", $( window ).height() );
            $rteContent[ 0 ].focus();
        }

        /*--(log.log( "ui/content/initialize(): end." );)--*/

        this.initialize = function() {};
    };

    exports.restoreCaret = function() {
        /*--(log.log( "/ui/content/restoreCaret()" );)--*/

        rendering.restoreContentCaret();
    };

    exports.updateStateChange = function() {
        /*--(log.log( "/ui/content/updateStateChange()" );)--*/

        io.updateStateChange();
    };

    exports.replaceContent = function( content ) {
        /*--(log.log( "/ui/content/replaceContent()" );)--*/
        contentState.setContent( content );
    };

    exports.sendContent = function( request ) {
        /*--(log.log( "/ui/content/sendContent(): begin..." );)--*/

        io.sendContent( request );

        /*--(log.log( "/ui/content/sendContent(): end." );)--*/
    };

    exports.setSingleLine = function( singleLine ) {
        /*--(log.log( "/ui/content/setSingleLine()" );)--*/

        var $rteContent = $( rteConfig.selector.RTE );

        $rteContent.toggleClass( rteConfig.className.SINGLE_LINE, singleLine );
    };

    exports.setTextColor = function( color ) {
        var $rteContent = $( rteConfig.selector.RTE );
        $rteContent.css( "color", color );
    };

    exports.updateViewStateChange = function( command, value ) {
        /*--(log.log( "/ui/content/updateViewStateChange()" );)--*/

        io.updateStateChange( command, value );
    };

    exports.cacheCurrentInsertion = function() {
        /*--(log.log( "/ui/content/cacheCurrentInsertion()" );)--*/

        rendering.cacheCurrentInsertion();
    };

    exports.updateEmpty = function() {
        /*--(log.log( "/ui/content/updateEmpty()" );)--*/

        var $content = $( rteConfig.selector.RTE );

        sanitization.removeFonts( $content );

        io.updateContentEmpty();

        return true;
    };

    exports.insertNewLineAtCaret = function() {
        /*--(log.log( "/ui/content/insertNewLineAtCaret()" );)--*/

        var handled = rendering.insertNewLineAtCaret();

        scroll.set();

        io.updateInsertNewLine();

        return handled;
    };

    exports.deleteCharacterAtCaret = function() {
        /*--(log.log( "/ui/content/deleteCharacterAtCaret()" );)--*/

        // TODO: suppressed until we fix the jumpy caret issue.
        var handled = true;//rendering.deleteCharacterAtCaret();

        scroll.set();

        return handled;
    };

    exports.insertAtAndStartMention = function() {
        /*--(log.log( "/ui/content/insertAtAndStartMention()" );)--*/

        var handled = rendering.insertTextAtCaret("@");

        mention.check();

        return handled;
    };

    exports.insertImage = function( src ) {
        /*--(log.log( "/ui/content/insertImage()" );)--*/

        return rendering.insertImageAtCaret( src );
    };

    exports.format = function( showUi, command, value ) {
        /*--(log.log( "/ui/content/format()" );)--*/

        return rendering.formatWithCommand( showUi, command, value );
    };

    exports.enableRte = function() {
        /*--(log.log( "/ui/content/enableRte()" );)--*/

        var $rteContent = $( rteConfig.selector.RTE );

        $rteContent.prop("contenteditable", true );
    };

    exports.disableRte = function() {
        /*--(log.log( "/ui/content/disableRte()" );)--*/

        var $rteContent = $( rteConfig.selector.RTE );

        $rteContent.prop("contenteditable", false );
    };

    return exports;
});

define( 'ui/rendering/clipboard/main',[
    "../main",
    "../../../util/log"
], function( range, log ) {
    

    var exports = {};

    exports.cut = function() {
        /*--(log.log( "ui/range/clipboard/cut" );)--*/

        var data = range.getText();

        range.removeSelected();

        return data;
    };

    exports.copy = function() {
        /*--(log.log( "ui/range/clipboard/copy" );)--*/

        return range.getText();
    };

    exports.paste = function( text ) {
        range.insertTextAtCaret( text );
        range.removeSelected();
    };

    return exports;
});

define( 'ui/clipboard',[
    "../proxy/io/main",
    "./rendering/clipboard/main",
    "../util/log"
], function( io, clipboard, log ) {
    

    var exports = {};

    exports.cut = function() {
        /*--(log.log( "ui/clipboard/cut" );)--*/
        var clipboardData = clipboard.cut();

        io.updateClipboard( clipboardData );

        return true;
    };

    exports.copy = function() {
        /*--(log.log( "ui/clipboard/copy" );)--*/
        var clipboardData = clipboard.copy();

        io.updateClipboard( clipboardData );

        return true;
    };

    exports.paste = function( text ) {
        clipboard.paste( text );

        return true;
    };

    return exports;
});

define( 'ui/sizer',[
    "../config/rte/main",
    "../proxy/io/main",
    "../state/sizer",
    "../util/timer",
    "./rendering/main",
    "jquery"
], function( rteConfig, io, sizerState, timer, rendering, $ ) {
    

    var exports = {},

        kImageWidth = rteConfig.content.IMAGE_WIDTH;

    function updateRectWidth( aNewRect, delta ) {
        if ( sizerState.left ) {
            aNewRect.width -= delta.left;
        } else if ( sizerState.right ) {
            aNewRect.width += delta.left;
        }
    }

    function updateRectHeight( aNewRect, delta ) {
        if ( sizerState.top ) {
            aNewRect.height -= delta.top;
        } else if ( sizerState.bottom ) {
            aNewRect.height += delta.top;
        }
    }

    function isHeightAdjustmentRequired( aNewRect ) {
        return !sizerState.vertical && ( sizerState.horizontal || aNewRect.height / aNewRect.width < sizerState.ratio );
    }

    function preserveRectRatio( aNewRect ) {

        // adjust the width/height to keep the same ratio
        if ( isHeightAdjustmentRequired( aNewRect ) ) {
            aNewRect.height = aNewRect.width * sizerState.ratio;
        } else {
            aNewRect.width = aNewRect.height / sizerState.ratio;
        }
    }

    function updateRectLeft( aNewRect ) {
        if ( sizerState.left ) {
            aNewRect.left += sizerState.rect.width - aNewRect.width;
        }
    }

    function updateRectTop( aNewRect ) {
        if ( sizerState.top ) {
            aNewRect.top += sizerState.rect.height - aNewRect.height;
        }
    }

    exports.$el = $( rteConfig.html.SIZER );

    exports.image = null;

    exports.select = function( target ) {
        $( "body" ).after( exports.$el );

        exports.image = $( target );

        rendering.addFromTarget( target );

        exports.apply();

        io.updateImageSelect();
    };

    exports.deselect = function() {
        if ( !exports.image ) {
            return;
        }

        exports.$el.remove();

        exports.image = undefined;

        io.updateImageDeselect();
    };

    exports.apply = function() {
        if ( !exports.image ) {
            return;
        }

        exports.$el.css( exports.image.offset() );
    };

    exports.show = function() {
        var size = exports.$el.offset();

        exports.ghost.remove();
        exports.ghost = undefined;

        exports.image.css({ visibility: "", width: size.width, height: size.height });
    };

    exports.applyAsync = function() {
        timer.setTimeout( exports.apply );
    };

    exports.replaceSrc = function( existingSrc, newSrc ) {
        $( "img[src='" + existingSrc + "']" ).attr( "src", newSrc );
    };

    exports.repaint = function( touch ) {
        var aNewRect = { left: sizerState.rect.left, top: sizerState.rect.top, width: sizerState.rect.width, height: sizerState.rect.height },
            delta = { left: touch.pageX - sizerState.point.left, top: touch.pageY - sizerState.point.top };

        updateRectWidth( aNewRect, delta );
        updateRectHeight( aNewRect, delta );
        preserveRectRatio( aNewRect );
        updateRectLeft( aNewRect );
        updateRectTop( aNewRect );

        exports.ghost.css({ width: aNewRect.width, height: aNewRect.height });

        exports.$el.css( aNewRect );
    };

    exports.prepare = function() {
        rendering.cacheCurrentInsertion();
    };

    exports.scaleToView = function( imgSrc ) {
        $( "img[src='" + imgSrc + "']" ).css({ width: kImageWidth, height: "auto" });
    };

    exports.disableResizing = function() {
        this.$el.addClass( rteConfig.className.RESIZE_DISABLED );
    };

    exports.moveTo = function( x, y ) {
        this.$el.css( "display", "none" );
        this.image.css( "display", "none" );

        var activeRange = rendering.createFromPoint( x, y );

        this.$el.css( "display", "" );
        this.image.css( "display", "" );

        if ( !activeRange ) {
            return;
        }

        rendering.insertNode( activeRange, this.image.get( 0 ) );
    };

    exports.initialize = function( touch, directions ) {
        sizerState.directions = directions;
        sizerState.left = ( sizerState.directions[ 0 ] === "left" );
        sizerState.right = ( sizerState.directions[ 0 ] === "right" );
        sizerState.vertical = ( sizerState.directions[ 0 ] === "middle" );
        sizerState.top = ( sizerState.directions[ 1 ] === "top" );
        sizerState.bottom = ( sizerState.directions[ 1 ] === "bottom" );
        sizerState.horizontal = ( sizerState.directions[ 1 ] === "middle" );
        sizerState.rect = exports.$el.offset();
        sizerState.point = { top: touch.pageY, left: touch.pageX };
        sizerState.ratio = sizerState.rect.height / sizerState.rect.width;

        exports.ghost = exports.image.clone().css({ width: sizerState.rect.width, height: sizerState.rect.height, margin: "0", boxSizing: "border-box" });

        exports.$el.prepend( exports.ghost );

        exports.image.css( "visibility", "hidden" );
    };

    return exports;
});

define( 'ui/rendering/list/main',[
    "../../../config/rte/main",
    "../../../dom/range/main",
    "../../../util/log",
    "jquery"
], function( rteConfig, range, log, $ ) {
    

    var exports = {},

        selector = rteConfig.selector,
        kList = selector.LIST,
        kHeading = selector.HEADING,
        kParagraph = selector.PARAGRAPH,

        content = rteConfig.content,
        kTabSize = content.TAB_SIZE,
        kTabMultiplier = content.TAB_MULTIPLIER;

    function isPastMaxIndent( activeRange ) {
        /*--(log.log( "/range/main/isPastMaxIndent()" );)--*/

        return $( activeRange.commonAncestorContainer ).parents( kList ).length > rteConfig.content.MAX_INDENT_LIMIT;
    }

    function hasListParents( activeRange ) {
        /*--(log.log( "/range/main/hasListParents()" );)--*/

        return !!$( activeRange.commonAncestorContainer ).parents( kList ).length;
    }

    exports.exit = function() {
        /*--(log.log( "/range/main/exitList()" );)--*/

        var activeRange = range.getSelectedRange(),
            $headerWrap = $( activeRange.startContainer ).parent( kList ).parent( kHeading ),
            $listItem, selection, $paragraph;

        if ( !$headerWrap.length ) {
            return false;
        }

        $listItem = $( activeRange.startContainer );

        $listItem.remove();

        selection = range.getWindowSelection();

        selection.removeAllRanges();

        // Workaround https://bugs.webkit.org/show_bug.cgi?id=15256
        $paragraph = $( rteConfig.html.PARAGRAPH );

        $headerWrap.after( $paragraph );

        activeRange = document.createRange();

        activeRange.setStart( $paragraph[0], 0 );
        activeRange.setEnd( $paragraph[0], 0 );

        selection.addRange( activeRange );

        return true;
    };

    /**
     * `event/key` uses this.
     */
    exports.exitItem = function() {
        /*--(log.log( "/range/main/exitListItem()" );)--*/

        var selection = range.getWindowSelection(),
            activeRange = selection.getRangeAt( 0 );

        if ( !activeRange ) {
            return false;
        }

        if ( activeRange.collapsed && activeRange.startContainer.nodeName.toLowerCase() === rteConfig.selector.LIST_ITEM ) {
            return exports.exit();
        }

        return false;
    };

    exports.indent = function() {
        /*--(log.log( "/range/main/indent()" );)--*/

        var activeRange = range.getSelectedRange(),
            result = {
                stopFurtherProcessing: false,
                handled: false
            },
            $ancestor, $closestParagraph, paddingLeft;

        if ( hasListParents( activeRange ) ) {
            if ( isPastMaxIndent( activeRange ) ) {
                result.stopFurtherProcessing = true;
            }

            return result;
        }

        $ancestor = $( activeRange.commonAncestorContainer );
        $closestParagraph = $ancestor.closest( kParagraph );

        if ( !$closestParagraph.length ) {
            $closestParagraph = $ancestor.wrap( kParagraph );
        }

        paddingLeft = parseInt( $closestParagraph.css( "padding-left" ), 10 ) + kTabSize;
        paddingLeft = Math.min( paddingLeft, kTabSize * kTabMultiplier );

        $closestParagraph.css( "padding-left", paddingLeft + "px" );

        result.handled = true;

        return result;
    };

    exports.outdent = function() {
        /*--(log.log( "/range/main/outdent()" );)--*/

        var activeRange = range.getSelectedRange(),
            result = {
                stopFurtherProcessing: false,
                handled: false
            },
            $ancestor, $closestParagraph, paddingLeft;

        if ( hasListParents( activeRange ) ) {
            result.handled = false;

            return result;
        }

        $ancestor = $( activeRange.commonAncestorContainer );
        $closestParagraph = $ancestor.closest( kParagraph );

        if ( !$closestParagraph.length ) {
            $closestParagraph = $ancestor.wrap( kParagraph );
        }

        paddingLeft = parseInt( $closestParagraph.css( "padding-left" ), 10 ) - kTabSize;
        paddingLeft = Math.min( paddingLeft, kTabSize * kTabMultiplier );
        $closestParagraph.css( "padding-left", paddingLeft + "px" );

        result.handled = true;

        return result;
    };

    return exports;
});

define( 'ui/list',[
    "./rendering/list/main"
], function( list ) {
    

    var exports = {};

    exports.indent = function() {
        return list.indent();
    };

    exports.outdent = function() {
        return list.outdent();
    };

    exports.exit = function() {
        return list.exit();
    };

    exports.exitItem = function() {
        return list.exitItem();
    };

    return exports;
});

define('dom/composition/main',[
    "../validation/main"
], function( validation ) {
    

    var exports = {},

        commandToNodeNameMapping = {
            bold: "B",
            italic: "I",
            underline: "U"
        };

    function getFirstNonEmptyTextNode( previous ) {
        if ( validation.isEmptyTextNode( previous ) ) {
            return exports.getPreviousImmediateNonEmptyTextNode( previous );
        }

        if ( validation.isTextNode( previous ) ) {
            return previous;
        }

        return null;
    }

    function getFirstNonEmptyTextNodeFromChildren( previous ) {
        var last = previous.lastChild;

        while ( last ) {
            if ( validation.isEmptyTextNode( last ) ) {
                last = last.previousSibling;
            } else if ( validation.isTextNode( last ) ) {
                return last;
            } else {
                return null;
            }
        }

        return null;
    }

    function checkParents( nodeName, marker ) {
        var node = marker.parentNode;

        while ( node ) {
            if ( node.nodeName === nodeName ) {
                return true;
            }

            node = node.parentNode;
        }

        return false;
    }

    function extractParametersFromCommand( command ) {
        var result = {},
            key;

        for ( key in command ) {
            if ( command.hasOwnProperty( key ) ) {
                result.type = key;
                result.shouldWrap = command[ key ];

                break;
            }
        }

        return result;
    }

    exports.getWrappings = function( marker ) {
        var result =  { bold: false, italic: false, underline: false };

        if ( !marker ) {
            return result;
        }

        result.bold = checkParents( commandToNodeNameMapping.bold, marker );
        result.italic = checkParents( commandToNodeNameMapping.italic, marker );
        result.underline = checkParents( commandToNodeNameMapping.underline, marker );

        return result;
    };

    exports.wrapNode = function( marker, command ) {
        var node = document.createElement( command );

        marker.parentNode.insertBefore( node, marker );

        node.appendChild( marker );

        return marker;
    };

    exports.unwrapNode = function( marker, command ) {
        var parent = marker.parentNode,
            frag;

        while ( parent ) {
            frag = document.createDocumentFragment();

            if ( parent.nodeName === command ) {
                while ( parent.firstChild ) {
                    frag.appendChild( parent.firstChild );
                }

                parent.parentNode.insertBefore( frag, parent );
                parent.parentNode.removeChild( parent );
            }

            parent = parent.parentNode;
        }

        return marker;
    };

    exports.getPreviousImmediateNonEmptyTextNode = function( root ) {
        var previous = root.previousSibling;

        if ( !previous ) {
            return null;
        }

        var children = previous.childNodes;

        if ( !children.length ) {
            return getFirstNonEmptyTextNode( previous );
        }

        var last = getFirstNonEmptyTextNodeFromChildren( previous );

        return last ? last : exports.getPreviousImmediateNonEmptyTextNode( previous );
    };

    exports.formatStyle = function( marker, command ) {
        var meta = extractParametersFromCommand( command ),
            type = meta.type,
            shouldWrap = meta.shouldWrap;

        if ( shouldWrap && exports.getWrappings( marker )[ type ] ) {
            return;
        }

        if ( shouldWrap && !exports.getWrappings( marker )[ type ] ) {
            exports.wrapNode( marker, commandToNodeNameMapping[type] );

            return;
        }

        if ( !shouldWrap && !exports.getWrappings( marker )[ type ] ) {
            return;
        }

        if ( !shouldWrap && exports.getWrappings( marker, type ) ) {
            exports.unwrapNode( marker, commandToNodeNameMapping[type] );

            return;
        }
    };

    return exports;
});

define( 'ui/format',[
    "../config/rte/main",
    "../dom/command/main",
    "../dom/composition/main",
    "../dom/main",
    "./rendering/main"
], function( rteConfig, command, composition, dom, rendering ) {
    

    var exports = {},

        kMarkerId = rteConfig.content.COMPOSITION_MARKER_ID;

    function toggle( state ) {
        var currentState = command.getState( state ),
            marker = dom.getById( kMarkerId ),
            options = {};

        options[ state ] = !currentState;

        if ( marker ) {
            composition.formatStyle( marker, options );

            rendering.ensureCaret( marker );

            return true;
        }

        return rendering.formatWithCommand( false, state, null );
    }

    exports.bold = function() {
        return toggle( "bold" );
    };

    exports.italic = function() {
        return toggle( "italic" );
    };

    exports.underline = function() {
        return toggle( "underline" );
    };

    return exports;
});

define( 'ui/composition',[
    "../util/log"
],function( log ) {
    

    var exports = {};

    // TODO: composition.* calls here are disabled until we fix #ANDROID-1015

    exports.setText = function( text ) {
        /*--(log.log( "/ui/composition/setText()" );)--*/

        void text;

        //composition.set( text );

        return true;
    };

    exports.commitText = function( text ) {
        /*--(log.log( "/ui/composition/commitText()" );)--*/

        void text;

        //composition.commit( text );

        return true;
    };

    exports.finishText = function() {
        /*--(log.log( "/ui/composition/finishText()" );)--*/

        //composition.finish();

        return true;
    };

    exports.deleteSurrounding = function( before, after ) {
        //composition.deleteSurrounding( before, after );

        // TODO: suppress until we fix the jumpy caret behavior.
        //rendering.deleteCharacterAtCaret();

        void before;
        void after;

        return true;
    };

    return exports;
});

define('util/functional',[],function() {
    

    var exports = {};

    exports.throttleWithDebounce = function(fn, delay) {
        var time = (new Date()).getTime(),
            fired = false,
            timerId = null;

        return function() {
            var now = (new Date().getTime()),
                remaining = time + delay - now;

            remaining = remaining > 0 ? remaining : 0;

            if ( fired ) {
                clearTimeout( timerId );
                timerId = setTimeout(function() {
                    time = now;
                    fn();
                }, remaining);

                return;
            }

            fired = true;
            fn();
        };
    };

    exports.throttle = function(fn, delay) {
        var time = (new Date()).getTime(),
            fired = false;

        return function() {
            var now = (new Date()).getTime();

            if (fired) {
                if (now - time < delay) {
                    return;
                }

                time = (new Date()).getTime();

                fn();

                return;
            }

            time = (new Date()).getTime();
            fired = true;

            fn();
        };
    };

    return exports;
});

define( 'legacy/ui/range/selection',[
    "jquery",
    "../../../dom/range/main",
    "./state/main"
], function( $, range, state ) {
    

    var exports = {};

    function updateState( serializedRange, text ) {
        state.serializedRange = serializedRange;
        state.rangeText = text;
        state.contentWidth = document.body.clientWidth;
    }

    /**
     * Tells the app to show the context menu.
     */
    function renderSelectionChange() {
        var sel = window.getSelection(),
            activeRange, selectionEnd, selectionStart, startRange, endRange,
            rect, text, rangyRange;

        if ( !sel ) {
            return;
        }

        activeRange = sel.getRangeAt( 0 );

        // Add spans to the selection to get page offsets
        // TODO: to constants.
        selectionEnd = $( "<span id=\"selectionEnd\"/>" );
        selectionStart = $( "<span id=\"selectionStart\"/>" );

        startRange = document.createRange();
        endRange = document.createRange();

        startRange.setStart( activeRange.startContainer, activeRange.startOffset );
        startRange.insertNode( selectionStart[0] );

        endRange.setStart( activeRange.endContainer, activeRange.endOffset );
        endRange.insertNode( selectionEnd[0] );

        state.handleBounds = {
            "left": selectionStart.offset().left,
            "top": selectionStart.offset().top,
            "right": selectionEnd.offset().left,
            "bottom": selectionEnd.offset().top
        };

        // Pull the spans
        selectionStart.remove();
        selectionEnd.remove();

        // Reset range
        sel.removeAllRanges();
        sel.addRange( activeRange );

        // Menu bounds
        rect = activeRange.getBoundingClientRect();

        state.menuBounds = {
            "left": rect.left,
            "top": rect.top,
            "right": rect.right,
            "bottom": rect.bottom
        };

        // TODO: rename, we don't use rangy ranges anymore.
        rangyRange = getRange();

        // Text to send to the selection
        text = window.getSelection().toString();

        updateState( rangyRange, text );
    }

    function getRange() {
        var serializedRangeSelected = range.getSerializedRange(),
            beginningCurly;

        // TODO: This section can be optimized.
        if ( serializedRangeSelected ) {
            beginningCurly = serializedRangeSelected.indexOf( "{" );

            serializedRangeSelected = serializedRangeSelected.substring( 0, beginningCurly );

            return serializedRangeSelected;
        }
    }

    function saveSelectionStart() {

        // Save the starting point of the selection
        // TODO: use dom/range/main
        var sel = window.getSelection(),
            activeRange = sel.getRangeAt( 0 ),
            saveRange = document.createRange();

        saveRange.setStart( activeRange.startContainer, activeRange.startOffset );

        state.selectionStartRange = saveRange;
    }

    function saveSelectionEnd() {

        // Save the end point of the selection
        // TODO: use dom/range/main
        var sel = window.getSelection(),
            activeRange = sel.getRangeAt( 0 ),
            saveRange = document.createRange();

        saveRange.setStart( activeRange.endContainer, activeRange.endOffset );

        state.selectionEndRange = saveRange;
    }

    function selectBetweenHandles() {
        var startCaret = state.selectionStartRange,
            endCaret = state.selectionEndRange,
            temp, activeRange, selection;

        // If we have two carets, update the selection
        if ( startCaret && endCaret ) {

            // If end caret comes before start caret, need to flip
            if ( startCaret.compareBoundaryPoints( window.Range.START_TO_END, endCaret ) > 0 ) {
                temp = startCaret;

                startCaret = endCaret;
                endCaret = temp;

                state.selectionStartRange = startCaret;
                state.selectionEndRange = endCaret;
            }

            activeRange = document.createRange();

            activeRange.setStart( startCaret.startContainer, startCaret.startOffset );
            activeRange.setEnd( endCaret.startContainer, endCaret.startOffset );

            range.clearWindowSelection();

            selection = window.getSelection();

            selection.addRange( activeRange );
        }

        renderSelectionChange();
    }

    exports.setEndPosition = function( x, y ) {

        // TODO: to dom/range/main
        state.selectionEndRange = document.caretRangeFromPoint( parseFloat( x + "" ), parseFloat( y + "" ) );

        selectBetweenHandles();

        return true;
    };

    exports.setStartPosition = function( x, y ) {

        // TODO: to dom/range/main
        state.selectionStartRange = document.caretRangeFromPoint( parseFloat( x + "" ), parseFloat( y + "" ) );

        selectBetweenHandles();

        return true;
    };

    exports.startLongTouch = function() {
        range.clearWindowSelection();

        // if current selection clear it.
        var sel = window.getSelection(),

        // TODO: to dom/range/main
            oneWordCaret = document.caretRangeFromPoint( state.lastTouchPoint.x, state.lastTouchPoint.y );

        oneWordCaret.expand( "word" );

        sel.addRange( oneWordCaret );

        saveSelectionStart();
        saveSelectionEnd();
        renderSelectionChange();

        return true;
    };

    exports.startTouch = function( x, y ) {
        state.lastTouchPoint = { x: parseFloat( x + "" ), y: parseFloat( y + "" ) };
    };

    exports.clear = function() {

        // TODO: use dom/range/main
        var sel = window.getSelection();

        sel.removeAllRanges();
    };

    return exports;
});

define( 'legacy/ui/selection',[
    "./range/selection",
    "../../proxy/io/main"
], function( range, io ) {
    

    /**
     * Range support for Android Ice Cream Sandwich, and other legacy Android devices.
     */

    var exports = {};

    exports.setEndPosition = function( x, y ) {
        range.setEndPosition( x, y );

        io.updateLegacySelection();

        return true;
    };

    exports.setStartPosition = function( x, y ) {
        range.setStartPosition( x, y );

        io.updateLegacySelection();

        return true;
    };

    exports.startLongTouch = function() {
        range.startLongTouch();

        io.updateLegacySelection();

        return true;
    };

    exports.startTouch = function( x, y ) {
        range.startTouch( x, y );

        return true;
    };

    exports.clear = function() {
        range.clear();

        return true;
    };

    return exports;
});

define( 'proxy/ui/main',[
    "../../config/rte/main",
    "../../native/factory",
    "../../state/content",
    "../../state/sizer",
    "../../ui/browser",
    "../../ui/content",
    "../../ui/clipboard",
    "../../ui/placeholder",
    "../../ui/mention",
    "../../ui/sizer",
    "../../ui/list",
    "../../ui/format",
    "../../ui/scroll",
    "../../ui/composition",
    "../../util/timer",
    "../../dom/validation/main",
    "../../util/functional",
    "../../util/log",
    "../../legacy/ui/selection",
    "jquery",
], function( rteConfig, nativeFactory, contentState, sizerState, browser, content, clipboard, placeholder, mention, sizer, list, format, scroll, composition, timer, validation, functional, log, legacySelection, $ ) {
    

    // TODO: stopFurtherProcessing SHOULD be false for all non-execDomCommand-able commands, fix them and test if everything works correctly.

    var exports = {},

        throttle = functional.throttle,

        kInsertHtml = rteConfig.topic.INSERT_HTML;

    /**
     * A proxy to work with the contentEditable div.
     *
     * proxy/command, and proxy/event modules should call proxy/ui to delegate
     * UI tasks, they should NOT call ui/* modules directly.
     *
     * Try NOT to do direct DOM manipulations here, instead yield those DOM
     * manipulations to appropriate ui/* modules.
     */

    function updateViewStateAsync( command, value ) {
        /*--(log.log( "proxy/command/updateViewStateAsync()" );)--*/

        placeholder.render();

        return content.updateViewStateChange( command, value );
    }

    exports.initialize = function() {
        /*--(log.log( "/proxy/ui/initialize()" );)--*/

        content.initialize();
        placeholder.initialize();
    };

    exports.launchInstallUrl = function( context ) {
        /*--(log.log( "/proxy/ui/launchInstallUrl()" );)--*/

        if ( !validation.hasContextAppUrl( context ) ) {
            return;
        }

        browser.launchInstallUrl( context );
    };

    exports.startSelectableViewCut = function() {
        /*--(log.log( "proxy/ui/startSelectableViewCut()" );)--*/

        var handled = clipboard.cut();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.startSelectableViewCopy = function() {
        /*--(log.log( "proxy/ui/starSelectableViewCopy()" );)--*/

        var handled = clipboard.copy();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.startSelectableViewPaste = function( pasteText ) {
        /*--(log.log( "/proxy/ui/startSelectableViewPaste()" );)--*/

        var handled = clipboard.paste( pasteText );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.updateSizerState = function( touch ) {
        /*--(log.log( "/proxy/ui/updateSizerState()" );)--*/

        sizerState.point = { top: touch.pageY, left: touch.pageX };

        sizerState.rect = sizer.$el.offset();
        sizerState.newRect = sizerState.rect;

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.applySizer = function() {
        /*--(log.log( "/proxy/ui/applySizer()" );)--*/

        sizer.show();
        sizer.apply();

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.renderSizer = function( target ) {
        /*--(log.log( "/proxy/ui/renderSizer()" );)--*/

        sizer.prepare();

        if ( target.nodeName.toLowerCase() === rteConfig.selector.IMAGE ) {
            sizer.select( target );
        } else if ( sizer.image ) {
            sizer.deselect();
        }

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.deselectSizer = function() {
        /*--(log.log( "/proxy/ui/deselectSizer()" );)--*/

        sizer.deselect();

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.initializeSizer = function( touch, directions ) {
        /*--(log.log( "/proxy/ui/initializeSizer()" );)--*/

        sizer.initialize( touch, directions );

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.repaintSizer = function( touch ) {
        /*--(log.log( "/proxy/ui/repaintSizer()" );)--*/

        sizer.repaint( touch );

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.updateStateChange = function() {
        /*--(log.log( "/proxy/ui/updateStateChange()" );)--*/

        content.updateStateChange();

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    // We throttle the resize update for 1 second, because when we change
    // heading styles certain devices such as Android 4.4.2 Nexus tablet
    // fire rezie events, and each style adjustment below fires yet another
    // resize event that turns out to be a never-ending loop of resizes.
    // Which "after a certain time elapses" causes the device to leak
    // memory, use excessive CPU, and extremely slow down.
    // The more heading styles we change the worse the impact is.
    //
    // Throttling resize solves this by executing the first style change and
    // discarding the second one that comes immediately after.
    exports.updateResize = throttle(function() {
        /*--(log.log( "/proxy/ui/updateResize(): Begin" );)--*/

        // FIX for #ANDROID-1287
        document.body.style.margin = "0";

        if ( $( window ).height() ) {
            var $rteContent = $( rteConfig.selector.RTE );

            $rteContent.css( "height", $( window ).height() );

            $rteContent[ 0 ].focus();
        }

        scroll.set();

        sizer.applyAsync();

        /*--(log.log( "/proxy/ui/updateResize(): End" );)--*/

    // TODO: to constants.
    }, 1000);

    exports.replaceContent = function( newContent ) {
        /*--(log.log("proxy/ui/replaceContent()");)--*/

        content.replaceContent(newContent);

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.sendContent = function( request ) {
        /*--(log.log( "proxy/ui/sendContent()" );)--*/

        content.sendContent( request );

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.setTextColor = function( color ) {
        /*--(log.log( "proxy/ui/setTextColor()" );)--*/

        content.setTextColor( color );

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.indent = function() {
        /*--(log.log( "proxy/ui/indent()" );)--*/

        var meta = list.indent();

        return {
            handled: meta.handled,
            stopFurtherProcessing: meta.stopFurtherProcessing
        };
    };

    exports.outdent = function() {
        /*--(log.log( "proxy/ui/outdent()" );)--*/

        var meta = list.outdent();

        return {
            handled: meta.handled,
            stopFurtherProcessing: meta.stopFurtherProcessing
        };
    };

    exports.setScroll = function() {
        /*--(log.log( "proxy/ui/setScroll()" );)--*/

        scroll.set();

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.insertOrderedList = function() {
        /*--(log.log( "proxy/ui/exitList()" );)--*/

        var handled = list.exit();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.insertOrderedList = function() {
        /*--(log.log( "proxy/ui/exitList()" );)--*/

        var handled = list.exit();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };

    };

    exports.exitListItem = function() {
        /*--(log.log( "proxy/ui/exitListItem()" );)--*/

        var handled = list.exitItem();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.formatBlock = function() {
        /*--(log.log( "proxy/ui/formatBlock()" );)--*/

        var handled = list.exit();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.bold = function() {
        /*--(log.log( "proxy/ui/bold()" );)--*/

        var handled = format.bold();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.italic = function() {
        /*--(log.log( "proxy/ui/italic()" );)--*/

        var handled = format.italic();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.underline = function() {
        /*--(log.log( "/proxy/ui/underline()" );)--*/

        var handled = format.underline();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.ensureContentEmpty = function() {
        /*--(log.log( "/proxy/ui/ensureContentEmpty()" );)--*/

        var handled = content.updateEmpty();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.insertNewLine = function() {
        /*--(log.log( "/proxy/ui/insertNewLine()" );)--*/

        var handled = content.insertNewLineAtCaret();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.insertBackspace = function() {
        /*--(log.log( "/proxy/ui/insertBackspace()" );)--*/

        // TODO: suppressed until we fix jumpy caret behavior.
        var handled = true; //content.deleteCharacterAtCaret();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.insertAtAndStartMention = function() {
        /*--(log.log( "ui/content/insertAtAndStartMention()" );)--*/

        var handled = content.insertAtAndStartMention();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.insertImage = function( src ) {
        /*--(log.log( "ui/content/insertImage()" );)--*/

        var handled = content.insertImage( src );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.insertMention = function( url, name, type ) {
        /*--(log.log( "ui/content/insertMention()" );)--*/

        var handled = mention.insert( url, name, type );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.checkMention = function() {
        var handled = mention.check();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.mention = function( meta ) {
        /*--(log.log( "ui/content/mention()" );)--*/

        var handled = mention.update( meta );

        /*--(log.log( "ui/content/mention: updated." );)--*/

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.setSelectableViewEndPosition = function( coordinate ) {
        /*--(log.log( "/proxy/ui/setSelectableViewEndPosition() " + arguments.length );)--*/
        /*--(log.log( JSON.stringify( arguments[0] ) );)--*/

        var endX = coordinate[ 0 ],
            endY = coordinate[ 1 ],
            handled = legacySelection.setEndPosition( endX, endY );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.setSelectableViewStartPosition = function( coordinate ) {
        /*--(log.log( "/proxy/ui/setSelectableViewStartPosition() " + arguments.length );)--*/

        var startX = coordinate[ 0 ],
            startY = coordinate[ 1 ],
            handled = legacySelection.setStartPosition( startX, startY );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.clearSelectableViewSelection = function() {
        /*--(log.log( "/proxy/ui/clearSelectableViewSelection() " + arguments.length );)--*/

        var handled = legacySelection.clear();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.startSelectableViewLongTouch = function() {
        /*--(log.log( "/proxy/ui/startSelectableViewLongTouch() " + arguments.length );)--*/

        var handled = legacySelection.startLongTouch();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.startSelectableViewTouch = function( coordinate ) {
        /*--(log.log( "/proxy/ui/startSelectableViewTouch() " + arguments.length );)--*/

        var x = coordinate[ 0 ],
            y = coordinate[ 1 ],
            handled = legacySelection.startTouch( x, y );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.execDomCommand = function( showUi, command, value ) {
        /*--(log.log( "/proxy/ui/execDomCommand( " + showUi + ", " + command + ", " + value + ")" );)--*/

        content.format( showUi, command, value );

        if ( command === kInsertHtml ) {
            placeholder.render();
        }

        return {
            handled: true,
            stopFurtherProcessing: true
        };
    };

    exports.restoreCaret = function() {
        /*--(log.log( "/proxy/ui/restoreCaret()" );)--*/

        content.restoreCaret();

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.determinePlaceholder = function() {
        /*--(log.log( "/proxy/ui/determinePlaceholder()" );)--*/

        placeholder.render();

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.updateViewState = function( command, value ) {
        /*--(log.log( "/proxy/ui/updateViewState()" );)--*/

        return updateViewStateAsync( command, value );
    };

    exports.updateAttachedImageCount = function( count ) {
        /*--(log.log( "/proxy/ui/updateAttachedImageCount()" );)--*/

        contentState.attachedImageCount = count;

        var handled = true;

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    // #region Text Mapping

    exports.deleteSurroundingText = function( before, after ) {
        /*--(log.log( "/proxy/ui/deleteSurroundingText()" );)--*/

        var handled = composition.deleteSurrounding( before, after );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.setComposingText = function( value ) {
        /*--(log.log( "/proxy/ui/setCompisingText()" );)--*/

        var handled = composition.setText( value );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    exports.commitText = function( value ) {
        /*--(log.log( "/proxy/ui/commitText()" );)--*/

        var handled = composition.commitText( value );

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    // TODO: this is never called, is this normal? -- if it won't be called, we should remove it from the code too.
    exports.setComposingRegion = function() {
        /*--(log.log( "/proxy/ui/setComposingRegion()" );)--*/

        return {
            handled: true,
            stopFurtherProcessing: false
        };
    };

    exports.finishComposingText = function() {
        /*--(log.log( "/proxy/ui/finishComposingText()" );)--*/

        var handled = composition.finishText();

        return {
            handled: handled,
            stopFurtherProcessing: false
        };
    };

    // #endregion

    return exports;
});

define('proxy/command/uimap',[],function() {
    

    var exports = {},
        map = {};

    /**
     * Only ./topic should call this method.
     */
    exports.add = function( topicName, methodName ) {
        map[ topicName ] = methodName;
    };

    exports.getAll = function() {
        return map;
    };

    return exports;
});

define('proxy/command/topic',[
    "../../native/factory",
    "../../config/rte/main",
    "./uimap"
], function( nativeFactory, rteConfig, uiMap ) {
    

    var jiveNative = nativeFactory.createInstance(),
        isLegacy = jiveNative.isLegacy && jiveNative.isLegacy(),
        exports = {},

        ctr = rteConfig.topic.rte,

        mapping = [
            ctr.SET_SINGLE_LINE,

            ctr.REPLACE_CONTENT,
            ctr.SEND_CONTENT,
            ctr.INDENT,
            ctr.OUTDENT,
            ctr.INSERT_ORDERED_LIST,
            ctr.INSERT_UNORDERED_LIST,
            ctr.FORMAT_BLOCK,

            ctr.BOLD,
            ctr.ITALIC,
            ctr.UNDERLINE,

            ctr.ENSURE_CONTENT_EMPTY,
            ctr.INSERT_NEW_LINE,
            ctr.INSERT_IMAGE,
            ctr.SET_TEXT_COLOR,
            ctr.MENTION,
            ctr.INSERT_MENTION,
            ctr.INSERT_AT_AND_START_MENTION,

            ctr.INSERT_BACKSPACE,
            ctr.DELETE_SURROUNDING_TEXT,

            ctr.UPDATE_ATTACHED_IMAGE_COUNT
        ],

        legacyMapping = [

            // Legacy cut/copy/paste
            ctr.START_SELECTABLEVIEW_CUT,
            ctr.START_SELECTABLEVIEW_COPY,
            ctr.START_SELECTABLEVIEW_PASTE,

            // Legacy selection for ICS.
            ctr.SET_SELECTABLEVIEW_END_POSITION,
            ctr.SET_SELECTABLEVIEW_START_POSITION,
            ctr.SET_SELECTABLEVIEW_SELECTION,
            ctr.START_SELECTABLEVIEW_LONG_TOUCH,
            ctr.START_SELECTABLEVIEW_TOUCH
        ],

        topics  = [];

    function wire( mapping ) {
        mapping.forEach(function(item) {
            topics.push(item);

            // TODO: reduce it to uiMap.add( item );
            uiMap.add( item, item );
        });
    }

    exports.getAll = function() {
        return topics;
    };

    exports.initialize = function() {
        wire( mapping );

        if ( isLegacy ) {
            wire( legacyMapping );
        }

        exports.initialize = function() {};
    };

    return exports;
});

define( 'proxy/command/main',[
    "../../config/rte/main",
    "../../io/hub",
    "../../util/log",
    "../ui/main",
    "./topic",
    "./uimap"
], function( rteConfig, hub, log, ui, topic, uiMap ) {
    

    /**
     * Proxy from Device to Editor.
     *
     * This module should not deal with DOM,
     * it should delegate DOM manipulation tasks to other modules.
     *
     * If you feel like you need to import "jquery" here, then you are probably
     * adding your logic to the wrong module.
     *
     * All Device-sent commands are initially handled here.
     */

    var exports = {},

        genericTopics,
        topicUiMap,

        mentionTopics = [
            rteConfig.topic.rte.INSERT_MENTION
        ],

        directive = {},

        bind;

    topic.initialize();

    genericTopics = topic.getAll();
    topicUiMap = uiMap.getAll();

    // Test runner was unable to find Function.prototype.bind.
    if ( Function.prototype.bind ) {
        bind = function( method, context ) {
            return method.bind( context );
        };
    } else {
        bind = function( method, context ) {
            return function() {
                method.apply( context, arguments );
            };
        };
    }

    // Initializes the `directive` collection.
    (function initializeCommand() {
        function createJsonFromMeta( meta ) {
            return {
                handled: meta.handled,
                stopFurtherProcessing: meta.stopFurtherProcessing
            };
        }

        function handleCommand( /*varargin*/ ) {
            /* jshint validthis:true */

            /*--(log.log( "/proxy/command/main/handleCommand: '" + this.key + "'.");)--*/

            var delegate = ui[ topicUiMap[ this.key ] ];

            if ( !delegate ) {
                /*--(log.log( "/proxy/command/main/handleCommand() :: Command '" + this.key + "' is not defined." );)--*/
            }

            return createJsonFromMeta( delegate.apply( null, arguments ) );
        }

        var key;

        for ( key in topicUiMap ) {
            if ( topicUiMap.hasOwnProperty( key ) ) {
                directive[ key ] = bind( handleCommand, { key: key } );
            }
        }
    }());

    function executeUnhandledCommand(handled, showUi, command, args) {
        if ( !handled ) {
            /*--(log.log( "execCommand: " + showUi + ", " + command + ", " + args );)--*/

            ui.execDomCommand( showUi, command, args );

            /*--(log.log( "done:execCommand" );)--*/
        }

        /*--(log.log( "before updateViewState" );)--*/

        ui.updateViewState( command,  /bold|italic|underline/.test(command) ? document.queryCommandState(command) : args );

        /*--(log.log( "after updateViewState" );)--*/
    }

    function handleCommand( command, showUi, args ) {
        /*--(log.log( "proxy/command/handleCommand() " + command );)--*/

        var handled = false,
            stopFurtherProcessing = false,
            meta,

            delegate = directive[ command ];

        if ( delegate ) {
            meta = delegate.apply( null, args );

            handled = meta.handled;
            stopFurtherProcessing = meta.stopFurtherProcessing;
        } else {
            /*--(log.log( "Unknown command: '" + command + ( stopFurtherProcessing ? "'. Skipped." : "'. Will try `document.execCommand()`." ) );)--*/
        }

        // TODO: find where this `sendContent` command is sent. We should set stopFurtherProcessing=true instead of checking it here.
        if ( stopFurtherProcessing || command === "sendContent" ) {
            return;
        }

        ui.deselectSizer();

        executeUnhandledCommand(handled, showUi, command, args);
    }

    function subscribeToGenericTopic( topic ) {
        /*--(log.log( "subscribing to... '" + topic + "'." );)--*/

        hub.subscribe( topic, function( /* value */ ) {
            /*--(log.log( "handleGenericTopic:topic: " + topic );)--*/

            handleCommand( topic, null, arguments );
        });
    }

    // TODO: we can do it with subscribeToGenericTopic, too.
    function subscribeToMentionTopic( topic ) {
        /*--(log.log( "subscribing to... '" + topic + "'." );)--*/

        hub.subscribe( topic, function( value, fullName, type ) {
            /*--(log.log( "handleMentionTopic:topic: " + topic );)--*/

            var payload = {
                url: value,
                name: fullName,
                type: type
            };

            handleCommand( topic, null, payload );
        });
    }

    exports.initialize = function() {
        /*--(log.log( "/proxy/command/initialize()" );)--*/

        // Since Android cannot directly call JavaScript method, we have to preserve the API parity
        // for other native codes, too. instead of `contentEmpty`, call ENSURE_CONTENT_EMPTY and
        // listen to the callback.

        genericTopics.map( function( topic ) {
            subscribeToGenericTopic( topic );
        });

        mentionTopics.map( function( topic ) {
            subscribeToMentionTopic( topic );
        });
    };

    return exports;
});

define( 'event/content',[
    "../config/rte/main",
    "../proxy/ui/main",
    "jquery"
], function( rteConfig, ui, $ ) {
    

    /**
     * Events specifically related to RTE main container.
     */

    var exports = {};

    function bindBlurEvents() {
        var $content = $( rteConfig.selector.RTE );

        $content.on( "blur", function() {
            $content.one( "focus", function() {
                ui.restoreCaret();
            });
        });
    }

    function bindKeyEvents() {
        var $content = $( rteConfig.selector.RTE );

        $content.on( "keydown", ui.determinePlaceholder );
    }

    exports.initialize = function() {
        bindBlurEvents();
        bindKeyEvents();

        this.initialize = function() {
        };
    };

    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_resetState = function() {
        exports.initialize = function() {
            bindBlurEvents();
            bindKeyEvents();

            this.initialize = function() {};
        };
    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

define( 'event/document',[
    "../proxy/ui/main",
    "../util/log"
], function( ui, log ) {
    

    var exports = {};

    // TODO: var bindFocusEvents = privates.bindFocusEvents where privates is a submodule.
    // do this fo all private methods in all modules.

    function bindFocusEvents() {
        window.addEventListener( "focus", ui.updateStateChange );
    }

    function bindErrorEvents() {
        window.onerror = function( error, file, line ) {
            log.logError( error, { file: file, line: line });
        };
    }

    function bindResizeEvents() {
        window.addEventListener( "resize", ui.updateResize );
    }

    function bindReadyStateChangeEvents() {
        document.addEventListener( "readystatechange", function() {

            // TODO: move this code to ui.someMethod.

            // Wait until document is loaded
            if ( document.readyState === "complete" ) {

                // Traverse the iframes that are used to embed media
                var iframes = Array.prototype.slice.call( document.querySelectorAll( "iframe[data-aspect-ratio]" ) );

                iframes.forEach(
                    function( iframe ) {
                        var aspectRatio = iframe.getAttribute( "data-aspect-ratio" );

                        if ( aspectRatio ) {

                            // Sets the height of the iframe using the current width in the client and the aspect ratio
                            iframe.setAttribute( "height", iframe.getBoundingClientRect().width / aspectRatio );
                        }
                    }
                );
            }
        });
    }

    exports.initialize = function() {
        bindReadyStateChangeEvents();
        bindFocusEvents();
        bindErrorEvents();
        bindResizeEvents();

        this.initialize = function() {};
    };

    // TODO: this looks ugly, use someHelper.resetState(obj) instead, which overrides this initialize function for test purposes.
    // do it for similar test stubs too.
    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_resetState = function() {
        exports.initialize = function() {
            bindReadyStateChangeEvents();
            bindFocusEvents();
            bindErrorEvents();
            bindResizeEvents();

            this.initialize = function() {};
        };
    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

define( 'event/mention',[
    "../util/log",
    "../native/validation",
    "../proxy/ui/main",
    "jquery"
], function( log, validation, ui, $ ) {
    

    var exports = {};

    function bindKeyEvents() {
        var $document = $( document );

        $document.on( "keyup", function( event ) {
            // Some devices send an additional event for the shift key; ignore it
            if ( !event.shiftKey ) {
                ui.checkMention();
            }
        } );
    }

    exports.initialize = function() {
        /*--(log.log( "event/mention/initialize" );)--*/

        bindKeyEvents();
    };

    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_resetState = function() {
        exports.initialize = function() {
            /*--(log.log( "event/mention/initialize" );)--*/

            bindKeyEvents();
        };
    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

define( 'event/gesture',[
    "../config/rte/main",
    "../proxy/ui/main",
    "jquery",
    "../native/validation"
], function( rteConfig, ui, $, validation ) {
    

    var exports = {},

        kContext = rteConfig.attribute.CONTEXT;

    /**
     * Registers general gesture events that are not handled in other modules.
     */

    function bindClickEvents() {
        document.addEventListener( "click", function( e ) {
            if ( !e.toElement || !e.toElement.hasAttribute( kContext ) ) {
                return;
            }

            e.preventDefault();

            ui.launchInstallUrl( JSON.parse( e.toElement.getAttribute( kContext ) ) );
        });
    }

    function bindTouchEvents() {

        // Temporary guard clause for merge.
        if ( validation.isAndroid() ) {
            document.addEventListener( "touchend", ui.updateStateChange );
        }
    }

    exports.initialize = function() {
        bindTouchEvents();
        bindClickEvents();

        this.initialize = function() {};
    };

    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_resetState = function() {
        exports.initialize = function() {
            bindTouchEvents();
            bindClickEvents();

            this.initialize = function() {
            };
        };

    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

define( 'event/key',[
    "../proxy/ui/main",
    "../native/validation",
    "jquery",
    "../util/log",
    "../util/timer",
    "../util/functional"
], function( ui, validation, $, log, timer, functional ) {
    

    var exports = {},

        throttle = functional.throttleWithDebounce,

        throttledStateChange = throttle(function() {
            ui.updateStateChange();

        // TODO: to constants.
        }, 307),

        throttledScrollChange = throttle(function() {
            ui.setScroll();

        // TODO: to constants. -- The intervals are intentionally prime numbers.
        }, 5003);

    var snifferTimerId,
        loopTimerId;

    function sendScrollChange() {
        throttledScrollChange();

        // Only need content length updates in Android
        if ( validation.isAndroid() ) {
            throttledStateChange();
        }
    }

    // FIX for #ANDROID-2040
    //
    // On "some" ICS devices keypress events
    // are not sent, therefore we wait for ~10 seconds after the editor launches
    // and if there are no key events, we start to send fabricated state change
    // events every 2 seconds.
    //
    // The first "keyup" event clears the timers, hence we no more send
    // fabricated events, and trigger state change on keyup only.
    function startKeypressSniffer() {
        snifferTimerId = timer.setTimeout(function() {
            function loop() {
                loopTimerId = timer.setTimeout(loop, 2000);

                sendScrollChange();
            }

            loop();
        }, 10000);
    }

    var clearKeypressSniffer = function() {
        timer.clearTimeout(snifferTimerId);
        timer.clearTimeout(loopTimerId);

        clearKeypressSniffer = function() {};
    };

    function bindKeyUpEvents() {
        $( document ).on( "keyup", function( event ) {
            /*--(log.log( "keyup:" + event.keyCode );)--*/

            clearKeypressSniffer();

            sendScrollChange();
        });
    }

    exports.initialize = function() {
        /*--(log.log( "event/key/initialize()" );)--*/

        bindKeyUpEvents();

        startKeypressSniffer();

        this.initialize = function() {};
    };

    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_resetState = function() {
        exports.initialize = function() {
            /*--(log.log( "event/key/initialize()" );)--*/

            bindKeyUpEvents();

            this.initialize = function() {};
        };
    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

define( 'event/sizer',[
    "../config/rte/main",
    "../native/validation",
    "../proxy/ui/main",
    "../util/timer",
    "jquery"
], function( rteConfig, validation, ui, timer, $ ) {
    

    var exports = {};

    /**
     * Events specific to the image sizer.
     */

    function bindTouchEvents() {
        var state = {
            resizeHandlesVisible: false,
            activeImage: null,
            isMoving: false,
            initialTouchLeft: 0,
            initialTouchTop: 0,
            boxLeft: 0,
            boxTop: 0,
            boxWidth: 0,
            boxHeight: 0,
            deltaLeft: 0,
            deltaTop: 0
        },
            $handles = $( rteConfig.html.SIZER ),
            clearMovingTimerId = null;

        function showHandles( img ) {
            document.body.appendChild( $handles[0] );

            $handles.css( $( img ).offset() );
        }

        function hideHandles() {
            if ( !$handles || !$handles.length || !$handles[ 0 ].parentNode ) {
                return;
            }

            $handles[ 0 ].parentNode.removeChild( $handles[ 0 ] );
        }

        function calculateWiderWidth( width, state ) {
            return Math.min( width, ( state.boxWidth + state.deltaLeft ) );
        }

        function calculateWiderHeight( height, state ) {
            return Math.min( height, ( state.boxHeight * ( state.boxWidth + state.deltaLeft ) / state.boxWidth ) );
        }

        function updateCssWider( state ) {
            var width = document.body.offsetWidth - 20,
                height = document.body.offsetHeight - 20,
                newWidth = calculateWiderWidth( width, state ),
                newHeight = calculateWiderHeight( height, state );

            if ( !( state.deltaLeft > 0 && ( state.boxWidth + state.deltaLeft ) >= width ) ) {
                $handles.css({
                    width: newWidth,
                    height: newHeight
                });

                $( state.activeImage ).css({
                    width: newWidth,
                    height: newHeight
                });
            }
        }

        function calculateTallerWidth( width, state ) {
            return Math.min( width, ( state.boxWidth * ( state.boxHeight + state.deltaTop ) / state.boxHeight ) );
        }

        function calculateTallerHeight( height, state ) {
            return Math.min( height, ( state.boxHeight + state.deltaTop ) );
        }

        function updateCssTaller( state ) {
            var width = document.body.offsetWidth - 20,
                height = document.body.offsetHeight - 20,
                newWidth = calculateTallerWidth( width, state ),
                newHeight = calculateTallerHeight( height, state );

            if ( !( state.deltaTop > 0 && ( state.boxHeight + state.deltaTop ) >= ( document.body.offsetHeight - 20 ) ) ) {
                $handles.css({
                    width: newWidth,
                    height: newHeight
                });

                $( state.activeImage ).css({
                    width: newWidth,
                    height: newHeight
                });
            }
        }

        function updateHandleCss( state ) {
            if ( state.boxWidth > state.boxHeight ) {
                updateCssWider( state );
            } else {
                updateCssTaller( state );
            }
        }

        function updateStateFromTouch( touch ) {
            var offset;

            if ( touch ) {
                state.initialTouchLeft = touch.pageX;
                state.initialTouchTop = touch.pageY;

                offset = $( state.activeImage ).offset();

                state.boxLeft = offset.left;
                state.boxTop = offset.top;
                state.boxWidth = offset.width;
                state.boxHeight = offset.height;
            }
        }

        function updateStateFromEventTarget( target ) {
            var offset;

            state.resizeHandlesVisible = true;
            state.activeImage = target;

            offset = $( target ).offset();

            state.boxLeft = offset.left;
            state.boxTop = offset.top;
            state.boxWidth = offset.width;
            state.boxHeight = offset.height;
        }

        function updateInitialTouch( touch ) {
            if ( touch ) {
                state.initialTouchLeft = touch.pageX;
                state.initialTouchTop = touch.pageY;
            }
        }

        function updateHandleVisibility( touch ) {
            if ( !touch ) {
                hideHandles();
            }
        }

        document.addEventListener( "touchstart", function( evt ) {
            var shallShowHandles = !state.resizeHandlesVisible && evt.target.nodeName.toLowerCase() === "img",
                touch = evt.touches && evt.touches[ 0 ];

            if ( state.resizeHandlesVisible ) {
                updateStateFromTouch( touch );
            }

            if ( shallShowHandles ) {
                showHandles( evt.target );
                updateStateFromEventTarget( evt.target );
                updateInitialTouch( touch );
                updateHandleVisibility( touch );
            }
        });

        document.addEventListener( "touchmove", function( evt ) {
            clearTimeout( clearMovingTimerId );

            var touch = evt.touches && evt.touches[0];

            if ( !touch ) {
                return;
            }

            // TODO: separation of concerns: clear ui/sizer and move relevant code in here to ui/sizer.

            state.deltaLeft = touch.pageX - state.initialTouchLeft;
            state.deltaTop = touch.pageY - state.initialTouchTop;

            $handles.css( $( state.activeImage ).offset() );

            updateHandleCss( state );

            state.isMoving = true;
        });

        document.addEventListener( "touchend", function( evt ) {

            // TODO: rename jv-image-sizer jv-handle etc.
            var shallHideHandles = ( !state.isMoving && ( evt.target.className.indexOf( "handle" ) > -1 || evt.target.className.indexOf( "image-sizer" ) > -1 ) ) ||
                (evt.target.nodeName.toLowerCase() !== "img" && evt.target.className.indexOf( "handle" ) === -1 && evt.target.className.indexOf( "image-sizer" ) === -1 );

            if ( shallHideHandles ) {
                hideHandles();

                state.resizeHandlesVisible = false;
                state.activeImage = null;
                state.isMoving = false;

                return;
            }

            clearTimeout( clearMovingTimerId );
            clearMovingTimerId = timer.setTimeout( function() {
                state.isMoving = false;
            }, 1000 );
        });
    }

    exports.initialize = function() {

        // ANDROID-425: image resize is not enabled in Android phones.
        if ( !validation.isAndroid() ) {
            bindTouchEvents();
        }

        this.initialize = function() {
        };
    };

    /**<testStub>**/
    /*jshint nomen:false*/
    exports._test_resetState = function() {
        exports.initialize = function() {
            if ( !validation.isAndroid() ) {
                bindTouchEvents();
            }

            this.initialize = function() {
            };
        };
    };
    /*jshint nomen:true*/
    /**</testStub>**/

    return exports;
});

define( 'proxy/event/main',[
    "../../config/rte/main",
    "../../event/content",
    "../../event/document",
    "../../event/mention",
    "../../event/gesture",
    "../../event/key",
    "../../event/sizer",
    "../../native/validation",
    "jquery"
], function( rteConfig, contentEvent, documentEvent, mentionEvent, gestureEvent, keyEvent, sizerEvent, validation, $ ) {
    

    var exports = {};

    exports.initialize = function() {
        var hasRte = $( rteConfig.selector.RTE ).length > 0;

        if ( !hasRte ) {
            return;
        }

        // Temporarily disable certain events until we do a merge.
        if ( validation.isAndroid() ) {
            sizerEvent.initialize();
            mentionEvent.initialize();
            keyEvent.initialize();

            contentEvent.initialize();
        }

        documentEvent.initialize();
        gestureEvent.initialize();
    };

    return exports;
});

define( 'rte',[
    "./io/bridge",
    "./proxy/io/main",
    "./util/log",
    "./proxy/command/main",
    "./proxy/ui/main",
    "./proxy/event/main",
    "./native/validation"
], function( bridge, io, log, command, ui, event, validation ) {
    

    var exports = {};

    function initializeIo() {
        bridge.initialize();
        command.initialize();
    }

    function initializeUi() {
        ui.initialize();
    }

    function initializeEvents() {
        event.initialize();
    }

    exports.initialize = function() {

        // CAVEAT: This guard clause SHOULD remain until iOS 3.3.0.
        if ( validation.isIos() ) {
            return false;
        }

        /*--(log.log( "rte/main: initialize(): begin." );)--*/

        initializeIo();
        initializeUi();
        initializeEvents();

        io.sendRteReadyStatusUpdate();

        /*--(log.log( "rte/main: initialize(): ready to go!" );)--*/

        return true;
    };

    return exports;
});

define( 'main',[
    "./rte",
    "./util/log"
], function( rte, log ) {
    

    /**
     * Application lifecycle starts here.
     */

    /*--(log.log( "Initializing RTE..." );)--*/

    rte.initialize();

    /*--(log.log( "Initialized RTE." );)--*/
});


require(["main"]);
}());